@@ -1,5 +1,5 @@
 /*
- * @(#)$Id: acconfig.h,v 1.7 2002/01/25 22:12:23 crazyed Exp $
+ * @(#)$Id: acconfig.h,v 1.35 2006/06/07 02:17:04 jnelson Exp $
  */
 
 /*
@@ -198,7 +198,7 @@
 /*
  *  Perl support.
  */
-#undef PERL
+#undef HAVE_PERL
 
 /* Define this is DIRSIZ takes no argument */
 #undef DIRSIZ_TAKES_NO_ARG
@@ -241,8 +241,142 @@
 
 /* Define this if you want OPENSSL support */
 #undef HAVE_SSL
+
 /* Define this if you want OPENSSL2 support */
 #undef WANT_SSL2
+
 /* Define this if you have sysctlbyname() */
 #undef HAVE_SYSCTLBYNAME
 
+/* Define this if you have SO_SNDLOWAT */
+#undef HAVE_SO_SNDLOWAT
+
+/* Define this if you have struct sockaddr_storage */
+#undef HAVE_STRUCT_SOCKADDR_STORAGE
+
+/* Define this if your system has SA_LEN in its sockaddrs */
+#undef HAVE_SA_LEN
+
+/* Define this if you have inet_ntop() */
+#undef HAVE_INET_NTOP
+
+/* Define this if you have inet_pton() */
+#undef HAVE_INET_PTON
+
+/* Define this if you have gethostbyname2() */
+#undef HAVE_GETHOSTBYNAME2
+
+/* Define this if you don't have siglen_t */
+#undef HAVE_SOCKLEN_T
+
+/* Define this if you have a getaddrinfo() with missing functionality */
+#undef GETADDRINFO_DOES_NOT_DO_AF_UNIX
+
+/* Define this if you do not want INET6 support */
+#undef DO_NOT_USE_IPV6
+
+/* Define this if you have struct addrinfo */
+#undef HAVE_STRUCT_ADDRINFO
+
+/* Define this if you have struct sockaddr_in6 */
+#undef HAVE_STRUCT_SOCKADDR_IN6
+
+/* Define this if you have getservbyport() */
+#undef HAVE_GETSERVBYPORT
+
+/* Define this if you have getaddrinfo */
+#undef HAVE_GETADDRINFO
+
+/* Define this if you have getnameinfo */
+#undef HAVE_GETNAMEINFO
+
+/* Define this if you have sys_siglist declared */
+#undef SYS_SIGLIST_DECLARED
+
+/* Define this if you have a broken realpath */
+#undef HAVE_BROKEN_REALPATH
+
+/* Define this if you have TCL */
+#undef HAVE_TCL
+
+/* Define this if you have RUBY */
+#undef HAVE_RUBY
+
+/* Define this if you have clock_gettime() */
+#undef HAVE_CLOCK_GETTIME
+
+/* Define this if you have <intypes.h> */
+#undef HAVE_INTTYPES_H
+
+/* Define this if you have intptr_t */
+#undef HAVE_INTPTR_T
+
+/* Define this if you have a (long long) */
+#undef HAVE_LONG_LONG
+
+/* Define this if you have atoll() */
+#undef HAVE_ATOLL
+
+/* Define this if you have strtoll() */
+#undef HAVE_STRTOLL
+
+/* Define this if you have atoq() */
+#undef HAVE_ATOQ
+
+/* Define this to use select() */
+#undef USE_SELECT
+
+/* Define this to use poll() */
+#undef USE_POLL
+
+/* Define this to use kqueue() */
+#undef USE_FREEBSD_KQUEUE
+
+/* Define this to use pthreads */
+#undef USE_PTHREAD
+
+/* Define this if you have arc4random() */
+#undef HAVE_ARC4RANDOM
+
+/* Define this if you have solaris ports */
+#undef USE_SOLARIS_PORTS
+
+/* Define this if your largest int is (long) */
+#undef HAVE_INTMAX_LONG
+
+/* Define this if your largest int is (quad_t) */
+#undef HAVE_INTMAX_QUADT
+
+/* Define this if your largest int is (long long) */
+#undef HAVE_INTMAX_LONG_LONG
+
+/* Define this if your largest int is (intmax_t) */
+#undef HAVE_INTMAX_NATIVE
+
+/* Define this if you have a strtoimax() */
+#undef HAVE_STRTOIMAX
+
+/* Define this if you have a strtoq() */
+#undef HAVE_STRTOQ
+
+/* Define this if you have <stddef.h> */
+#undef HAVE_STDDEF_H
+
+/* Define this if you have ruby and stuff */
+#undef RUBYCFLAGS
+
+/* Define this if you have ruby and stuff */
+#undef RUBYLDFLAGS
+
+/* Define this if you have tcl and stuff */
+#undef TCLCFLAGS
+
+/* Define this if you have tcl and stuff */
+#undef TCLLDFLAGS
+
+/* Define this if you have perl and stuff */
+#undef PERLCFLAGS
+
+/* Define this if you have perl and stuff */
+#undef PERLLDFLAGS
+
