@@ -10,38 +10,38 @@
 #ifndef __commands_h__
 #define __commands_h__
 
-/* flags used by e_away */
-#define AWAY_ONE 			0
-#define AWAY_ALL 			1
-
-/* flags used by parse_line */
-#define SECURITY_NO_VARIABLE_COMMAND	1
-#define SECURITY_NO_NONINTERACTIVE_EXEC 2
-#define SECURITY_NO_NONINTERACTIVE_SET	4
-
-extern	char	*sent_nick;
-extern	char	*sent_body;
-extern	char	*recv_nick;
-
 extern	int	will_catch_break_exceptions;
 extern	int	will_catch_continue_exceptions;
 extern	int	will_catch_return_exceptions;
 extern	int	break_exception;
 extern	int	continue_exception;
 extern	int	return_exception;
+extern	int	system_exception;
 
 extern	int	need_defered_commands;
 
-	int	check_wait_command 	(char *);
-	void	ExecuteTimers		(void);
-	void	parse_line 		(const char *, const char *, const char *, int, int);
+	void	init_commands		(void);
+
+        char *  call_lambda_function    (Char *, Char *, Char *);
+        void    call_lambda_command     (Char *, Char *, Char *);
+        char *  call_user_function      (Char *, Char *, char *, void *);
+        void    call_user_command       (Char *, Char *, char *, void *);
+	void	runcmds			(Char *, Char *);
+
+	int     parse_statement 	(const char *, int, const char *);
+
 	BUILT_IN_COMMAND(load);
-	void	send_text	 	(const char *, const char *, const char *, int);
+	void	send_text	 	(int, const char *, const char *, const char *, int);
 	int	redirect_text		(int, const char *, const char *, char *, int);
 	int	command_exist		(char *);
 	BUILT_IN_COMMAND(e_channel);
 	void	do_defered_commands	(void);
-	char	*get_all_commands	(void);
 	char	*get_command		(const char *);
 
+        void    dump_load_stack         (int);
+const   char *  current_filename        (void);
+const   char *  current_loader          (void);
+        int     current_line            (void);
+const   char *  current_package         (void);
+
 #endif /* __commands_h__ */
