@@ -9,8 +9,10 @@
 #ifndef __hook_h__
 #define __hook_h__
 
+#define FIRST_NAMED_HOOK 1000
+
 enum HOOK_TYPES {
-	ACTION_LIST,
+	ACTION_LIST = FIRST_NAMED_HOOK,
 	CHANNEL_NICK_LIST,
 	CHANNEL_SIGNOFF_LIST,
 	CHANNEL_SYNC_LIST,
@@ -18,6 +20,7 @@ enum HOOK_TYPES {
 	CTCP_LIST,
 	CTCP_REPLY_LIST,
 	CTCP_REQUEST_LIST,
+	DCC_ACTIVITY_LIST,
 	DCC_CHAT_LIST,
 	DCC_CONNECT_LIST,
 	DCC_LIST_LIST,
@@ -43,9 +46,9 @@ enum HOOK_TYPES {
 	INPUT_LIST,
 	INVITE_LIST,
 	JOIN_LIST,
+	KEYBINDING_LIST,
 	KICK_LIST,
 	KILL_LIST,
-	LEAVE_LIST,
 	LIST_LIST,
 	MAIL_LIST,
 	MODE_LIST,
@@ -53,13 +56,16 @@ enum HOOK_TYPES {
 	MSG_LIST,
 	MSG_GROUP_LIST,
 	NAMES_LIST,
+	NEW_NICKNAME_LIST,
 	NICKNAME_LIST,
 	NOTE_LIST,
 	NOTICE_LIST,
 	NOTIFY_SIGNOFF_LIST,
 	NOTIFY_SIGNON_LIST,
+	NUMERIC_LIST,
 	ODD_SERVER_STUFF_LIST,
 	OPER_NOTICE_LIST,
+	PART_LIST,
 	PONG_LIST,
 	PUBLIC_LIST,
 	PUBLIC_MSG_LIST,
@@ -74,8 +80,10 @@ enum HOOK_TYPES {
 	SEND_NOTICE_LIST,
 	SEND_PUBLIC_LIST,
 	SEND_TO_SERVER_LIST,
+	SERVER_ESTABLISHED_LIST,
 	SERVER_LOST_LIST,
 	SERVER_NOTICE_LIST,
+	SERVER_STATUS_LIST,
 	SET_LIST,
 	SIGNOFF_LIST,
 	SILENCE_LIST,
@@ -89,22 +97,33 @@ enum HOOK_TYPES {
 	WALL_LIST,
 	WALLOP_LIST,
 	WHO_LIST,
-	WIDELIST_LIST,
 	WINDOW_LIST,
+	WINDOW_COMMAND_LIST,
 	WINDOW_CREATE_LIST,
+	WINDOW_BEFOREKILL_LIST,
 	WINDOW_KILL_LIST,
 	YELL_LIST,
 	ZZZZ_THIS_ALWAYS_COMES_LAST_ZZZZ
 };
 
 #define NUMBER_OF_LISTS ZZZZ_THIS_ALWAYS_COMES_LAST_ZZZZ
+#define INVALID_HOOKNUM -1001
 
 	BUILT_IN_COMMAND(oncmd);
 	BUILT_IN_COMMAND(shookcmd);
-	int	do_hook 		(int, char *, ...) __A(2);
+
+	int	do_hook 		(int, const char *, ...) __A(2);
+	char 	* hookctl			(char *);
 	void	flush_on_hooks 		(void);
 	void	unload_on_hooks		(char *);
 	void	save_hooks 		(FILE *, int);
 	void	do_stack_on		(int, char *);
+	int	hook_find_free_serial	(int, int, int);
 
+	extern int deny_all_hooks;
+
+#define ADD_STR_TO_LIST(str, sep, item, len) if (len == 0) \
+	{str = malloc_strdup(item); len = strlen(str);} \
+	else malloc_strcat_wordlist_c(&str, sep, item, &len);
+	
 #endif /* __hook_h_ */
