@@ -22,11 +22,10 @@ typedef struct PromptStru
 	char	*data;
 	int	type;
 	int	echo;
-	void	(*func)();
+	void	(*func) (char *, char *);
 	struct	PromptStru	*next;
 }	WaitPrompt;
 
-
 typedef	struct	ScreenStru
 {
 	/* List stuff and overhead */
@@ -50,7 +49,7 @@ struct	ScreenStru *next;		/* Previous screen in list */
 	int	wserv_version;		/* The version of wserv talking to */
 
 	/* Input line and prompt stuff */
-	char	input_buffer[INPUT_BUFFER_SIZE+1];
+	unsigned char	input_buffer[INPUT_BUFFER_SIZE+1];
 					/* Current user input for us */
 	int	buffer_pos;		/* Where on input line cursor is */
 	int	buffer_min_pos;		/* First position after prompt */
@@ -59,34 +58,23 @@ struct	ScreenStru *next;		/* Previous screen in list */
 	int	input_zone_len;
 	int	input_start_zone;
 	int	input_end_zone;
-	char	*input_prompt;
+	unsigned char	*input_prompt;
 	int	input_prompt_len;
 	int	input_prompt_malloc;
 	int	input_line;
 
-	char	saved_input_buffer[INPUT_BUFFER_SIZE+1];
+	unsigned char	saved_input_buffer[INPUT_BUFFER_SIZE+1];
 	int	saved_buffer_pos;
 	int	saved_min_buffer_pos;
 
 	WaitPrompt	*promptlist;
 
 	/* Key qualifier stuff */
-	int	meta_hit;		/* The last meta key that was hit */
 	int	quote_hit;		/* True after QUOTE_CHARACTER hit */
-#if 0
-	int	meta_hit[10];		/* 
-					 * Used to build keybindings...
-					 * Only one of these can be set at
-					 * any given time, should probably
-					 * be an int and using bitflags.
-					 */
-#endif
-
-
-	/* Redirect stuff */
-	char	*redirect_name;
-	char	*redirect_token;
-	int	redirect_server;
+	Timeval last_press;		/* The last time a key was pressed.
+					   Used to determine
+					   key-independence. */
+	void *	last_key;		/* The last Key pressed. */
 
 	char	*tty_name;
 	int	co;
@@ -97,46 +85,30 @@ struct	ScreenStru *next;		/* Previous screen in list */
 	int	alive;
 }	Screen;
 
-/* Stuff for the screen/xterm junk */
-
-#define ST_NOTHING      -1
-#define ST_SCREEN       0
-#define ST_XTERM        1
-
-
-	void	add_wait_prompt 	(char *, void (*)(), char *, int, int);
-	void	set_current_screen 	(Screen *);
-	void	window_redirect		(char *, int);
-	int	check_screen_redirect	(char *);
+	void	add_wait_prompt 	(const char *, void (*)(char *, char *), const char *, int, int);
 	void	add_to_screen		(const unsigned char *);
-unsigned char**	split_up_line		(const unsigned char *, int);
-	int	output_line		(const unsigned char *);
-	void	cursor_not_in_display	(Screen *);
+	void	cursor_not_in_display	(struct ScreenStru *);
 	void	cursor_in_display	(Window *);
-	int	is_cursor_in_display	(Screen *);
-	void	repaint_window		(Window *, int, int);
-	Screen *create_new_screen	(void);
+	int	is_cursor_in_display	(struct ScreenStru *);
+	void	repaint_window_body	(Window *);
+	void	create_new_screen	(void);
 	Window	*create_additional_screen (void);
-	void	kill_screen		(Screen *);
-	void	close_all_screen	(void);
-	void	do_screens		(fd_set *);
+	void	kill_screen		(struct ScreenStru *);
 
-const	char *	all_off			(void);
-extern	int	normalize_never_xlate;
-extern	int	normalize_permit_all_attributes;
-	u_char *normalize_string	(const u_char *, int);
+const	u_char *all_off			(void);
+	u_char *new_normalize_string	(const u_char *, int, int);
 	u_char *denormalize_string	(const u_char *);
-	char   *normalize_color		(int, int, int, int);
-const	u_char *skip_ctl_c_seq		(const u_char *, int *, int *);
-	u_char **prepare_display	(const u_char *, int, int *, int);
-	int	output_with_count	(const unsigned char *, int, int);
+	u_char **prepare_display	(int, const u_char *, int, int *, int);
+	int	output_with_count	(const u_char *, int, int);
+	void    add_to_window_scrollback (Window *, const u_char *, intmax_t);
 
 /* Dont do any word-wrapping, just truncate each line at its place. */
 #define PREPARE_NOWRAP	0x01
 
-extern	Screen *main_screen;
-extern	Screen *last_input_screen;
-extern	Screen *screen_list;
-extern	Screen *output_screen;
+extern	struct ScreenStru *main_screen;
+extern	struct ScreenStru *last_input_screen;
+extern	struct ScreenStru *screen_list;
+extern	struct ScreenStru *output_screen;
+extern	int	display_line_mangler;
 
 #endif /* _SCREEN_H_ */
