@@ -9,76 +9,60 @@
 #define __vars_h__
 
 /* indexes for the irc_variable array */
+/*
 enum VAR_TYPES {
-	ALLOW_C1_CHARS_VAR,
-	ALT_CHARSET_VAR,
+*/
+
+#ifndef VARS_C
+extern
+#endif
+
+int	ALLOW_C1_CHARS_VAR,
 	ALWAYS_SPLIT_BIGGEST_VAR,
-	AUTO_NEW_NICK_VAR,
-	AUTO_RECONNECT_VAR,
-	AUTO_RECONNECT_DELAY_VAR,
-	AUTO_REJOIN_VAR,
-	AUTO_REJOIN_CONNECT_VAR,
-	AUTO_REJOIN_DELAY_VAR,
-	AUTO_UNMARK_AWAY_VAR,
-	AUTO_WHOWAS_VAR,
 	BAD_STYLE_VAR,
 	BANNER_VAR,
 	BANNER_EXPAND_VAR,
 	BEEP_VAR,
-	BEEP_MAX_VAR,
-	BEEP_ON_MSG_VAR,
-	BEEP_WHEN_AWAY_VAR,
-	BLINK_VIDEO_VAR,
-	BOLD_VIDEO_VAR,
 	CHANNEL_NAME_WIDTH_VAR,
-	CLIENTINFO_VAR,
+	CLIENT_INFORMATION_VAR,
 	CLOCK_VAR,
 	CLOCK_24HOUR_VAR,
 	CLOCK_FORMAT_VAR,
+	CLOCK_INTERVAL_VAR,
 	CMDCHARS_VAR,
-	COLOR_VAR,
-	COMMAND_MODE_VAR,
 	COMMENT_HACK_VAR,
-	CONNECT_TIMEOUT_VAR,
 	CONTINUED_LINE_VAR,
 	CPU_SAVER_AFTER_VAR,
 	CPU_SAVER_EVERY_VAR,
 	CURRENT_WINDOW_LEVEL_VAR,
 	DCC_AUTO_SEND_REJECTS_VAR,
+	DCC_DEQUOTE_FILENAMES_VAR,
 	DCC_LONG_PATHNAMES_VAR,
 	DCC_SLIDING_WINDOW_VAR,
 	DCC_STORE_PATH_VAR,
-	DCC_TIMEOUT_VAR,
 	DCC_USE_GATEWAY_ADDR_VAR,
 	DEBUG_VAR,
 	DISPATCH_UNKNOWN_COMMANDS_VAR,
 	DISPLAY_VAR,
-	DISPLAY_ANSI_VAR,
-	DISPLAY_PC_CHARACTERS_VAR,
 	DO_NOTIFY_IMMEDIATELY_VAR,
-	EIGHT_BIT_CHARACTERS_VAR,
 	FLOATING_POINT_MATH_VAR,
+	FLOATING_POINT_PRECISION_VAR,
 	FLOOD_AFTER_VAR,
+	FLOOD_IGNORE_VAR,
+	FLOOD_MASKUSER_VAR,
 	FLOOD_RATE_VAR,
+	FLOOD_RATE_PER_VAR,
 	FLOOD_USERS_VAR,
 	FLOOD_WARNING_VAR,
-	FULL_STATUS_LINE_VAR,
-	HELP_PAGER_VAR,
-	HELP_PATH_VAR,
-	HELP_PROMPT_VAR,
-	HELP_WINDOW_VAR,
 	HIDE_PRIVATE_CHANNELS_VAR,
 	HIGHLIGHT_CHAR_VAR,
 	HIGH_BIT_ESCAPE_VAR,
-	HISTORY_VAR,
-	HISTORY_CIRCLEQ_VAR,
-	HOLD_INTERVAL_VAR,
-	HOLD_MODE_VAR,
+	HOLD_SLIDER_VAR,
 	INDENT_VAR,
 	INPUT_ALIASES_VAR,
 	INPUT_PROMPT_VAR,
 	INSERT_MODE_VAR,
-	INVERSE_VIDEO_VAR,
+	KEY_INTERVAL_VAR,
 	LASTLOG_VAR,
 	LASTLOG_LEVEL_VAR,
 	LOAD_PATH_VAR,
@@ -86,15 +70,17 @@ enum VAR_TYPES {
 	LOGFILE_VAR,
 	LOG_REWRITE_VAR,
 	MAIL_VAR,
+	MAIL_INTERVAL_VAR,
+	MAIL_TYPE_VAR,
+	MANGLE_DISPLAY_VAR,
 	MANGLE_INBOUND_VAR,
 	MANGLE_LOGFILES_VAR,
 	MANGLE_OUTBOUND_VAR,
-	MAX_RECONNECTS_VAR,
-	META_STATES_VAR,
+	METRIC_TIME_VAR,
 	MIRC_BROKEN_DCC_RESUME_VAR,
 	MODE_STRIPPER_VAR,
-	ND_SPACE_MAX_VAR,
 	NEW_SERVER_LASTLOG_LEVEL_VAR,
+	NOTIFY_VAR,
 	NOTIFY_INTERVAL_VAR,
 	NOTIFY_LEVEL_VAR,
 	NOTIFY_ON_TERMINATION_VAR,
@@ -102,16 +88,14 @@ enum VAR_TYPES {
 	NO_CONTROL_LOG_VAR,
 	NO_CTCP_FLOOD_VAR,
 	NO_FAIL_DISCONNECT_VAR,
-	NUM_OF_WHOWAS_VAR,
+	OLD_MATH_PARSER_VAR,
+	OLD_SERVER_LASTLOG_LEVEL_VAR,
 	OUTPUT_REWRITE_VAR,
 	PAD_CHAR_VAR,
 	QUIT_MESSAGE_VAR,
-	RANDOM_LOCAL_PORTS_VAR,
 	RANDOM_SOURCE_VAR,
 	REALNAME_VAR,
-	REVERSE_STATUS_LINE_VAR,
 	SCREEN_OPTIONS_VAR,
-	SCROLL_VAR,
 	SCROLLBACK_VAR,
 	SCROLLBACK_RATIO_VAR,
 	SCROLL_LINES_VAR,
@@ -120,10 +104,8 @@ enum VAR_TYPES {
 	SHELL_FLAGS_VAR,
 	SHELL_LIMIT_VAR,
 	SHOW_CHANNEL_NAMES_VAR,
-	SHOW_END_OF_MSGS_VAR,
 	SHOW_NUMERICS_VAR,
 	SHOW_STATUS_ALL_VAR,
-	SHOW_WHO_HOPCOUNT_VAR,
 	SSL_CERTFILE_VAR,
 	SSL_KEYFILE_VAR,
 	SSL_PATH_VAR,
@@ -142,7 +124,8 @@ enum VAR_TYPES {
 	STATUS_INSERT_VAR,
 	STATUS_MAIL_VAR,
 	STATUS_MODE_VAR,
-	STATUS_NICK_VAR,
+	STATUS_NICKNAME_VAR,
+	STATUS_NOSWAP_VAR,
 	STATUS_NOTIFY_VAR,
 	STATUS_NO_REPEAT_VAR,
 	STATUS_OPER_VAR,
@@ -154,7 +137,7 @@ enum VAR_TYPES {
 	STATUS_SSL_ON_VAR,
 	STATUS_TRUNCATE_RHS_VAR,
 	STATUS_UMODE_VAR,
-	STATUS_USER0_VAR,
+	STATUS_USER_VAR,
 	STATUS_USER1_VAR,
 	STATUS_USER10_VAR,
 	STATUS_USER11_VAR,
@@ -196,24 +179,19 @@ enum VAR_TYPES {
 	STATUS_USER9_VAR,
 	STATUS_VOICE_VAR,
 	STATUS_WINDOW_VAR,
-	SUPPRESS_FROM_REMOTE_SERVER,
-	SUPPRESS_SERVER_MOTD_VAR,
-	SWITCH_CHANNEL_ON_PART_VAR,
-	TAB_VAR,
-	TAB_MAX_VAR,
+	SUPPRESS_FROM_REMOTE_SERVER_VAR,
+	SWITCH_CHANNELS_BETWEEN_WINDOWS_VAR,
 	TERM_DOES_BRIGHT_BLINK_VAR,
 	TRANSLATION_VAR,
 	TRANSLATION_PATH_VAR,
-	UNDERLINE_VIDEO_VAR,
-	USER_INFO_VAR,
-#define	USERINFO_VAR USER_INFO_VAR
-	VERBOSE_CTCP_VAR,
+	USER_INFORMATION_VAR,
 	WORD_BREAK_VAR,
 	WSERV_PATH_VAR,
+	WSERV_TYPE_VAR,
 	XTERM_VAR,
 	XTERM_OPTIONS_VAR,
 	NUMBER_OF_VARIABLES
-};
+/* } */ ;
 
 /* var_settings indexes ... also used in display.c for highlights */
 #define OFF 			0
@@ -224,20 +202,45 @@ enum VAR_TYPES {
 #define	DEBUG_EXPANSIONS	0x0002
 #define DEBUG_FUNCTIONS		0x0004
 
+typedef union builtin_variable {
+        char *  string;                 /* string value of variable */
+        int     integer;                /* int value of variable */
+} VARIABLE;
+
+/* the types of IrcVariables */
+#define STR_VAR 	0
+#define INT_VAR 	1
+#define CHAR_VAR 	2
+#define BOOL_VAR 	3
+
+/* IrcVariable: structure for each variable in the variable table */
+typedef struct
+{
+        int             type;           /* variable types, see below */
+        VARIABLE *      data;           /* The value of the variable */
+        void    (*func) (void *); 	/* func called when var is set */
+	char *		script;		/* script called when var is set */
+        u_short         flags;          /* flags for this variable */
+}       IrcVariable;
+
+
+
 	BUILT_IN_COMMAND(setcmd);
 
 	int	do_boolean 		(char *, int *);
-	int	get_int_var 		(enum VAR_TYPES);
-	char	*get_string_var 	(enum VAR_TYPES);
-	void	set_int_var		(enum VAR_TYPES, int);
-	void	set_string_var		(enum VAR_TYPES, const char *);
-	void	init_variables 		(void);
-	char	*make_string_var 	(const char *);
-	void	set_highlight_char	(char *);
+	int	do_short_boolean 	(char *, short *);
+	int	get_int_var 		(int);
+	char *	get_string_var 		(int);
+	void	set_var_value		(int, const char *, int);
+	void	init_variables_stage1	(void);
+	void	init_variables_stage2	(void);
+	char*	make_string_var 	(const char *);
 	int	charset_size 		(void);
 	void	save_variables 		(FILE *, int);
-	void	set_var_value 		(int, char *);
 	void	do_stack_set		(int, char *);
-	int	parse_mangle		(char *, int, char **);
+	int	parse_mangle		(const char *, int, char **);
+	char	*get_set		(const char *);
+	char *	make_string_var_bydata	(int, void *);
+	int	set_variable		(const char *, IrcVariable *, const char *, int);
 
 #endif /* _VARS_H_ */
