@@ -41,7 +41,7 @@
 #include "server.h"
 #include "status.h"
 #include "screen.h"
-#include "termx.h"
+#include "term.h"
 #include "timer.h"
 #include "input.h"
 #include "mail.h"
@@ -139,7 +139,7 @@ static	long	last_milliday;
 	}
 }
 
-void	reset_clock (const void *stuff)
+void	reset_clock (void *stuff)
 {
 	if (x_debug & DEBUG_BROKEN_CLOCK)
 		reset_broken_clock();
@@ -164,19 +164,23 @@ void	clock_systimer (void)
 	reset_clock(NULL);
 }
 
-void    set_clock_interval (const void *stuff)
+void    set_clock_interval (void *stuff)
 {
 	update_system_timer(clock_timeref);	/* XXX Oh heck, why not? */
 }
 
-void     set_clock_format (const void *stuff)
+void     set_clock_format (void *stuff)
 {
-	const char *value = (const char *)stuff;
+	const VARIABLE *v;
+	const char *value;
+
+	v = (const VARIABLE *)stuff;
+	value = v->string;
         malloc_strcpy(&time_format, value);
         reset_clock(NULL);
 }
 
-void	set_clock (const void *stuff)
+void	set_clock (void *stuff)
 {
 	update_system_timer(clock_timeref);
 	update_all_status();
@@ -220,13 +224,18 @@ int	cpu_saver_timer (void *schedule_only)
 		cpu_saver_on(0, NULL);
 	else
 		add_timer(1, cpu_saver_timeref, interval - been_idlin, 
-				1, cpu_saver_timer, NULL, NULL, -1);
+				1, cpu_saver_timer, NULL, NULL, 
+				GENERAL_TIMER, -1, 0);
 	return 0;
 }
 
-void    set_cpu_saver_after (const void *stuff)
+void    set_cpu_saver_after (void *stuff)
 {
-	int value = *(const int *)stuff;
+	VARIABLE *v;
+	int	value;
+
+	v = (VARIABLE *)stuff;
+	value = v->integer;
 
         if (value == 0)
 	{
@@ -238,14 +247,18 @@ void    set_cpu_saver_after (const void *stuff)
                 cpu_saver_timer(NULL);
 }
 
-void	set_cpu_saver_every (const void *stuff)
+void	set_cpu_saver_every (void *stuff)
 {
-	int value = *(const int *)stuff;
+	VARIABLE *v;
+	int	value;
+
+	v = (VARIABLE *)stuff;
+	value = v->integer;
 
 	if (value < 60)
 	{
 		say("/SET CPU_SAVER_EVERY must be set to at least 60");
-		set_int_var(CPU_SAVER_EVERY_VAR, 60);
+		v->integer = 60;
 	}
 }
 
@@ -255,21 +268,21 @@ void	set_cpu_saver_every (const void *stuff)
 struct system_timer {
 	char *	name;
 	int	honors_cpu_saver;
-	int	interval_variable;
-	int	toggle_variable;
+	int *	interval_variable;
+	int *	toggle_variable;
 	Timeval	last_event;
 	void	(*callback) (void);
 };
 
 struct system_timer system_timers[] = {
 	{ clock_timeref, 	1, 
-	  CLOCK_INTERVAL_VAR, 	CLOCK_VAR, 	{ 0, 0 },
+	  &CLOCK_INTERVAL_VAR, 	&CLOCK_VAR, 	{ 0, 0 },
 	  clock_systimer 	},
 	{ notify_timeref, 	1, 
-	  NOTIFY_INTERVAL_VAR,  NOTIFY_VAR, 	{ 0, 0 },
+	  &NOTIFY_INTERVAL_VAR,  &NOTIFY_VAR, 	{ 0, 0 },
 	  notify_systimer	},
 	{ mail_timeref, 	1, 
-	  MAIL_INTERVAL_VAR, 	MAIL_VAR, 	{ 0, 0 },
+	  &MAIL_INTERVAL_VAR, 	&MAIL_VAR, 	{ 0, 0 },
 	  mail_systimer 	},
 	{ NULL,			0, 
 	  0,			0,		{ 0, 0 },
@@ -291,15 +304,12 @@ static int	system_timer (void *entry)
 	}
 	else
 	{
-	    nominal_timeout = get_int_var(item->interval_variable);
+	    nominal_timeout = get_int_var(*item->interval_variable);
 	    timeout = time_to_next_interval(nominal_timeout);
 	}
 
-#if 0
-yell("Rescheduling system timer [%s] to go off in [%f] sec", 
-			item->name, timeout);
-#endif
-	add_timer(1, item->name, timeout, 1, system_timer, entry, NULL, -1);
+	add_timer(1, item->name, timeout, 1, system_timer, entry, NULL, 
+				GENERAL_TIMER, -1, 0);
 
 	item->callback();
 	item->last_event = now;
@@ -321,8 +331,8 @@ int	update_system_timer (const char *entry)
 		/* This needs to be set before calling 'system_timer' */
 		get_time(&now);
 
-		if (get_int_var(system_timers[i].toggle_variable) &&
-		    get_int_var(system_timers[i].interval_variable))
+		if (get_int_var(*system_timers[i].toggle_variable) &&
+		    get_int_var(*system_timers[i].interval_variable))
 			system_timer(&system_timers[i]);
 		else
 		{
