@@ -1,4 +1,4 @@
-/* $EPIC: ctcp.c,v 1.34 2004/11/10 03:20:35 jnelson Exp $ */
+/* $EPIC: ctcp.c,v 1.53 2006/07/05 23:40:14 jnelson Exp $ */
 /*
  * ctcp.c:handles the client-to-client protocol(ctcp). 
  *
@@ -55,7 +55,6 @@
 #include "window.h"
 #include "if.h"
 #include "flood.h"
-#include "notice.h"
 #include "words.h"
 
 #include <pwd.h>
@@ -92,30 +91,27 @@ typedef	struct _CtcpEntry
 	CTCP_Handler 	repl;	/* Function that is called for reply */
 }	CtcpEntry;
 
-/* forward declarations for the built in CTCP functions */
-static	char	*do_sed 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_version 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_clientinfo 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_ping 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_echo 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_userinfo 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_finger 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_time 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_atmosphere 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_dcc 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_utc 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_dcc_reply 	(CtcpEntry *, const char *, const char *, char *);
-static	char	*do_ping_reply 	(CtcpEntry *, const char *, const char *, char *);
+#define CTCP_HANDLER(x) \
+static char * x (CtcpEntry *ctcp, Char *from, Char *to, char *cmd)
 
+/* forward declarations for the built in CTCP functions */
+CTCP_HANDLER(do_crypto);
+CTCP_HANDLER(do_version);
+CTCP_HANDLER(do_clientinfo);
+CTCP_HANDLER(do_ping);
+CTCP_HANDLER(do_echo);
+CTCP_HANDLER(do_userinfo);
+CTCP_HANDLER(do_finger);
+CTCP_HANDLER(do_time);
+CTCP_HANDLER(do_atmosphere);
+CTCP_HANDLER(do_dcc);
+CTCP_HANDLER(do_utc);
+CTCP_HANDLER(do_dcc_reply);
+CTCP_HANDLER(do_ping_reply);
 
 static CtcpEntry ctcp_cmd[] =
 {
-	{ "SED",	CTCP_SED, 	CTCP_INLINE | CTCP_NOLIMIT,
-		"contains simple_encrypted_data",
-		do_sed, 	do_sed },
-	{ "UTC",	CTCP_UTC, 	CTCP_INLINE | CTCP_NOLIMIT,
-		"substitutes the local timezone",
-		do_utc, 	do_utc },
+	/* The most common ones */
 	{ "ACTION",	CTCP_ACTION, 	CTCP_SPECIAL | CTCP_NOLIMIT,
 		"contains action descriptions for atmosphere",
 		do_atmosphere, 	do_atmosphere },
@@ -125,6 +121,37 @@ static CtcpEntry ctcp_cmd[] =
 	{ "VERSION",	CTCP_VERSION,	CTCP_REPLY | CTCP_TELLUSER,
 		"shows client type, version and environment",
 		do_version, 	NULL },
+
+	/* Common ones to people using strong crypto */
+	{ "AESSHA256-CBC",	CTCP_AESSHA256,	CTCP_INLINE | CTCP_NOLIMIT,
+		"transmit aes256-cbc ciphertext using a sha256 key",
+		do_crypto, 	do_crypto },
+	{ "AES256-CBC",		CTCP_AES256,	CTCP_INLINE | CTCP_NOLIMIT,
+		"transmit aes256-cbc ciphertext",
+		do_crypto, 	do_crypto },
+	{ "CAST128ED-CBC",	CTCP_CAST5, 	CTCP_INLINE | CTCP_NOLIMIT,
+		"transmit cast5-cbc ciphertext",
+		do_crypto, 	do_crypto},
+	{ "BLOWFISH-CBC",	CTCP_BLOWFISH,	CTCP_INLINE | CTCP_NOLIMIT,
+		"transmit blowfish-cbc ciphertext",
+		do_crypto, 	do_crypto },
+	{ "SED",		CTCP_SED, 	CTCP_INLINE | CTCP_NOLIMIT,
+		"transmit simple_encrypted_data ciphertext",
+		do_crypto, 	do_crypto },
+	{ "SEDSHA",		CTCP_SEDSHA, 	CTCP_INLINE | CTCP_NOLIMIT,
+		"transmit simple_encrypted_data ciphertext using a sha256 key",
+		do_crypto, 	do_crypto },
+
+	/* The uncommon ones */
+	{ "PING", 	CTCP_PING, 	CTCP_REPLY | CTCP_TELLUSER,
+		"returns the arguments it receives",
+		do_ping, 	do_ping_reply },
+	{ "ECHO", 	CTCP_ECHO, 	CTCP_REPLY | CTCP_TELLUSER,
+		"returns the arguments it receives",
+		do_echo, 	NULL },
+	{ "UTC",	CTCP_UTC, 	CTCP_INLINE | CTCP_NOLIMIT,
+		"substitutes the local timezone",
+		do_utc, 	do_utc },
 	{ "CLIENTINFO",	CTCP_CLIENTINFO,CTCP_REPLY | CTCP_TELLUSER,
 		"gives information about available CTCP commands",
 		do_clientinfo, 	NULL },
@@ -140,12 +167,6 @@ static CtcpEntry ctcp_cmd[] =
 	{ "TIME",	CTCP_TIME, 	CTCP_REPLY | CTCP_TELLUSER,
 		"tells you the time on the user's host",
 		do_time, 	NULL },
-	{ "PING", 	CTCP_PING, 	CTCP_REPLY | CTCP_TELLUSER,
-		"returns the arguments it receives",
-		do_ping, 	do_ping_reply },
-	{ "ECHO", 	CTCP_ECHO, 	CTCP_REPLY | CTCP_TELLUSER,
-		"returns the arguments it receives",
-		do_echo, 	NULL },
 	{ NULL,		CTCP_CUSTOM,	CTCP_REPLY | CTCP_TELLUSER,
 		NULL,
 		NULL, NULL }
@@ -167,25 +188,22 @@ int     sed = 0;
  */
 int	in_ctcp_flag = 0;
 
-#define CTCP_HANDLER(x) \
-	static char * x (CtcpEntry *ctcp, const char *from, const char *to, char *cmd)
-
-
 
 /**************************** CTCP PARSERS ****************************/
 
 /********** INLINE EXPANSION CTCPS ***************/
 /*
- * do_sed: Performs the Simple Encrypted Data trasfer for ctcp.  Returns in a
- * malloc string the decryped message (if a key is set for that user) or the
- * text "[ENCRYPTED MESSAGE]" 
+ * do_crypto: Performs strong decryption for ctcp.  Returns in a malloc 
+ * string the decryped message (if a key is set for that user) or the text 
+ * "[ENCRYPTED MESSAGE]" 
  */
-CTCP_HANDLER(do_sed)
+CTCP_HANDLER(do_crypto)
 {
 	Crypt	*key = NULL;
 	const char	*crypt_who;
 	char 	*tofrom;
 	char	*ret = NULL, *ret2 = NULL;
+	int	type;
 
 	if (*from == '=')		/* DCC CHAT message */
 		crypt_who = from;
@@ -197,8 +215,8 @@ CTCP_HANDLER(do_sed)
 	tofrom = malloc_strdup3(to, ",", from);
 	malloc_strcat2_c(&tofrom, "!", FromUserHost, NULL);
 
-	if ((key = is_crypted(tofrom)) ||
-	    (key = is_crypted(crypt_who)))
+	if ((key = is_crypted(tofrom, from_server, ctcp->id)) ||
+	    (key = is_crypted(crypt_who, from_server, ctcp->id)))
 		ret = decrypt_msg(cmd, key);
 
 	new_free(&tofrom);
@@ -242,16 +260,14 @@ CTCP_HANDLER(do_utc)
  */
 CTCP_HANDLER(do_atmosphere)
 {
-	const char	*old_message_from;
-	int	old_message_level;
+	int	l;
 
 	if (!cmd || !*cmd)
 		return NULL;
 
-	save_message_from(&old_message_from, &old_message_level);
 	if (is_channel(to))
 	{
-		message_from(to, LOG_ACTION);
+		l = message_from(to, LEVEL_ACTION);
 		if (do_hook(ACTION_LIST, "%s %s %s", from, to, cmd))
 		{
 			if (is_current_channel(to, from_server))
@@ -262,12 +278,12 @@ CTCP_HANDLER(do_atmosphere)
 	}
 	else
 	{
-		message_from(from, LOG_ACTION);
+		l = message_from(from, LEVEL_ACTION);
 		if (do_hook(ACTION_LIST, "%s %s %s", from, to, cmd))
 			put_it("*> %s %s", from, cmd);
 	}
 
-	restore_message_from(old_message_from, old_message_level);
+	pop_message_from(l);
 	return NULL;
 }
 
@@ -289,8 +305,8 @@ CTCP_HANDLER(do_dcc)
 
 	if     (!(type = next_arg(cmd, &cmd)) ||
 		!(description = (get_int_var(DCC_DEQUOTE_FILENAMES_VAR)
-			 ? new_next_arg(cmd, &cmd)
-			 : next_arg(cmd, &cmd))) ||
+				? new_next_arg(cmd, &cmd)
+				: next_arg(cmd, &cmd))) ||
 		!(inetaddr = next_arg(cmd, &cmd)) ||
 		!(port = next_arg(cmd, &cmd)))
 			return NULL;
@@ -380,13 +396,13 @@ CTCP_HANDLER(do_version)
 	send_ctcp(CTCP_NOTICE, from, CTCP_VERSION, 
 			"ircII %s %s %s - %s", 
 			irc_version, the_unix, the_version,
-			(tmp = get_string_var(CLIENTINFO_VAR)) ? 
+			(tmp = get_string_var(CLIENT_INFORMATION_VAR)) ? 
 				tmp : IRCII_COMMENT);
 #else
 	send_ctcp(CTCP_NOTICE, from, CTCP_VERSION, 
 			"ircII %s *IX - %s", 
 			irc_version,
-			(tmp = get_string_var(CLIENTINFO_VAR)) ? 
+			(tmp = get_string_var(CLIENT_INFORMATION_VAR)) ? 
 				tmp : IRCII_COMMENT);
 #endif
 	return NULL;
@@ -406,7 +422,7 @@ CTCP_HANDLER(do_userinfo)
 	char *tmp;
 
 	send_ctcp(CTCP_NOTICE, from, CTCP_USERINFO, "%s", 
-		(tmp = get_string_var(USERINFO_VAR)) ? tmp : "<No User Information>");
+		(tmp = get_string_var(USER_INFORMATION_VAR)) ? tmp : "<No User Information>");
 	return NULL;
 }
 
@@ -457,7 +473,7 @@ CTCP_HANDLER(do_finger)
 #define GECOS_DELIMITER ','
 #endif
 
-#if !defined(I_DONT_TRUST_MY_USERS)
+#if defined(ALLOW_USER_SPECIFIED_LOGIN)
 	if ((ctcpuser = getenv("IRCUSER"))) 
 		strlcpy(userbuff, ctcpuser, sizeof userbuff);
 	else
@@ -469,7 +485,7 @@ CTCP_HANDLER(do_finger)
 			strlcpy(userbuff, "epic-user", sizeof userbuff);
 	}
 
-#if !defined(I_DONT_TRUST_MY_USERS)
+#if defined(ALLOW_USER_SPECIFIED_LOGIN)
 	if ((ctcpfinger = getenv("IRCFINGER"))) 
 		strlcpy(gecosbuff, ctcpfinger, sizeof gecosbuff);
 	else
@@ -587,7 +603,6 @@ char *	do_ctcp (const char *from, const char *to, char *str)
 {
 	int 	flag;
 	int	fflag;
-	int	lastlog_level;
 	char 	local_ctcp_buffer [BIG_BUFFER_SIZE + 1],
 		the_ctcp          [IRCD_BUFFER_SIZE + 1],
 		last              [IRCD_BUFFER_SIZE + 1];
@@ -597,6 +612,7 @@ char *	do_ctcp (const char *from, const char *to, char *str)
 	char	*ptr = NULL;
 	int	allow_ctcp_reply = 1;
 static	time_t	last_ctcp_parsed = 0;
+	int	l;
 
 	int delim_char = charcount(str, CTCP_DELIM_CHAR);
 
@@ -605,14 +621,13 @@ static	time_t	last_ctcp_parsed = 0;
 	if (delim_char > 8)
 		allow_ctcp_reply = 0;	/* Historical limit of 4 CTCPs */
 
-	flag = check_ignore_channel(from, FromUserHost, to, IGNORE_CTCPS);
+	flag = check_ignore_channel(from, FromUserHost, to, LEVEL_CTCP);
 	fflag = new_check_flooding(from, FromUserHost, is_channel(to) ? to : NULL,
-						str, CTCP_FLOOD); 
+						str, LEVEL_CTCP);
 
 	in_ctcp_flag++;
 	strlcpy(local_ctcp_buffer, str, sizeof(local_ctcp_buffer) - 2);
 
-	lastlog_level = set_lastlog_msg_level(LOG_CTCP);
 	for (;;strlcat(local_ctcp_buffer, last, sizeof(local_ctcp_buffer) - 2))
 	{
 		if (split_CTCP(local_ctcp_buffer, the_ctcp, last))
@@ -693,13 +708,13 @@ static	time_t	last_ctcp_parsed = 0;
 		if (ctcp_argument)
 			*ctcp_argument++ = 0;
 		else
-			ctcp_argument = empty_string;
+			ctcp_argument = endstr(the_ctcp);
 
 		/* Set up the window level/logging */
 		if (im_on_channel(to, from_server))
-			message_from(to, LOG_CTCP);
+			l = message_from(to, LEVEL_CTCP);
 		else
-			message_from(from, LOG_CTCP);
+			l = message_from(from, LEVEL_CTCP);
 
 		/*
 		 * Then we look for the correct CTCP.
@@ -717,24 +732,26 @@ static	time_t	last_ctcp_parsed = 0;
 			 * Offer it to the user.
 			 * Maybe they know what to do with it.
 			 */
-			if (do_hook(CTCP_LIST, "%s %s %s %s", from, to, 
+			if (do_hook(CTCP_REQUEST_LIST, "%s %s %s %s",
+				from, to, ctcp_command, ctcp_argument))
+			    if (do_hook(CTCP_LIST, "%s %s %s %s", from, to, 
 						ctcp_command, ctcp_argument))
-			{
-				if (get_int_var(VERBOSE_CTCP_VAR))
+			    {
 				    say("Unknown CTCP %s from %s to %s: %s%s",
 					ctcp_command, from, to, 
 					*ctcp_argument ? ": " : empty_string, 
 					ctcp_argument);
-			}
+			    }
 			time(&last_ctcp_parsed);
 			allow_ctcp_reply = 0;
+			pop_message_from(l);
 			continue;
 		}
 
 		/* 
 		 * We did find it.  Acknowledge it.
 		 */
-		ptr = empty_string;
+		ptr = NULL;
 		if (do_hook(CTCP_REQUEST_LIST, "%s %s %s %s",
 				from, to, ctcp_command, ctcp_argument))
 		{
@@ -765,7 +782,7 @@ static	time_t	last_ctcp_parsed = 0;
 		 * If its an ``INLINE'' CTCP, we paste it back in.
 		 */
 		if (ctcp_cmd[i].flag & CTCP_INLINE)
-			strlcat(local_ctcp_buffer, ptr, sizeof local_ctcp_buffer);
+			strlcat(local_ctcp_buffer, ptr ? ptr : empty_string, sizeof local_ctcp_buffer);
 
 		/* 
 		 * If its ``INTERESTING'', tell the user.
@@ -777,8 +794,6 @@ static	time_t	last_ctcp_parsed = 0;
 		    if (do_hook(CTCP_LIST, "%s %s %s %s", 
 				from, to, ctcp_command, ctcp_argument))
 		    {
-			if (get_int_var(VERBOSE_CTCP_VAR)) 
-			{
 			    if (is_me(from_server, to))
 				say("CTCP %s from %s%s%s", 
 					ctcp_command, from, 
@@ -789,17 +804,12 @@ static	time_t	last_ctcp_parsed = 0;
 					ctcp_command, from, to, 
 					*ctcp_argument ? ": " : empty_string, 
 					ctcp_argument);
-			}
 		    }
 		}
-		if (ptr != empty_string)
-			new_free(&ptr);
+		new_free(&ptr);
+		pop_message_from(l);
 	}
 
-	/* Reset the window level/logging */
-	set_lastlog_msg_level(lastlog_level);
-	message_from(NULL, LOG_CRAP);
-
 	in_ctcp_flag--;
 
 	/* 
@@ -818,7 +828,6 @@ static	time_t	last_ctcp_parsed = 0;
 char *	do_notice_ctcp (const char *from, const char *to, char *str)
 {
 	int 	flag;
-	int	lastlog_level;
 	char 	local_ctcp_buffer [BIG_BUFFER_SIZE + 1],
 		the_ctcp          [IRCD_BUFFER_SIZE + 1],
 		last              [IRCD_BUFFER_SIZE + 1];
@@ -827,6 +836,7 @@ char *	do_notice_ctcp (const char *from, const char *to, char *str)
 	int	i;
 	char	*ptr;
 	int	allow_ctcp_reply = 1;
+	int	l;
 
 	int delim_char = charcount(str, CTCP_DELIM_CHAR);
 
@@ -835,12 +845,12 @@ char *	do_notice_ctcp (const char *from, const char *to, char *str)
 	if (delim_char > 8)
 		allow_ctcp_reply = 0;	/* Ignore all the CTCPs. */
 
-	flag = check_ignore_channel(from, FromUserHost, to, IGNORE_CTCPS);
-	if (!in_ctcp_flag)
-		in_ctcp_flag = -1;
-	strlcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE - 2);
+	/* We handle ignore, but not flooding (obviously) */
+	flag = check_ignore_channel(from, FromUserHost, to, LEVEL_CTCP);
+	in_ctcp_flag++;
+	strlcpy(local_ctcp_buffer, str, sizeof(local_ctcp_buffer) - 2);
 
-	for (;;strlcat(local_ctcp_buffer, last, sizeof local_ctcp_buffer))
+	for (;;strlcat(local_ctcp_buffer, last, sizeof(local_ctcp_buffer) - 2))
 	{
 		if (split_CTCP(local_ctcp_buffer, the_ctcp, last))
 			break;		/* All done! */
@@ -849,7 +859,8 @@ char *	do_notice_ctcp (const char *from, const char *to, char *str)
 			continue;	/* Empty requests are ignored */
 
 		/*
-		 * Apply sanity rules
+		 * The logic of all this is essentially the same as 
+		 * do_ctcp
 		 */
 
 		if (!allow_ctcp_reply)
@@ -863,8 +874,11 @@ char *	do_notice_ctcp (const char *from, const char *to, char *str)
 			continue;
 		}
 
+		/* But we don't check ctcp flooding (obviously) */
+
 		/* Global messages -- just drop the CTCP */
-		if (*to == '$' || (*to == '#' && !im_on_channel(to, from_server)))
+		if (*to == '$' || (is_channel(to) && 
+					!im_on_channel(to, from_server)))
 		{
 			allow_ctcp_reply = 0;
 			continue;
@@ -873,20 +887,24 @@ char *	do_notice_ctcp (const char *from, const char *to, char *str)
 
 		/*
 		 * Parse CTCP message
-		 * CTCP spec says word delim MUST be space.
+		 * CTCP spec says word delim MUST be space
 		 */
 		ctcp_command = the_ctcp;
 		ctcp_argument = strchr(the_ctcp, ' ');
 		if (ctcp_argument)
 			*ctcp_argument++ = 0;
 		else
-			ctcp_argument = empty_string;
+			ctcp_argument = endstr(the_ctcp);
 
+		/* Set up the window level/logging */
+		if (is_channel(to))
+			l = message_from(to, LEVEL_CTCP);
+		else
+			l = message_from(from, LEVEL_CTCP);
 
 		/* 
 		 * Find the correct CTCP and run it.
 		 */
-
 		for (i = 0; i < NUMBER_OF_CTCPS; i++)
 			if (!strcmp(ctcp_command, ctcp_cmd[i].name))
 				break;
@@ -897,31 +915,32 @@ char *	do_notice_ctcp (const char *from, const char *to, char *str)
 		 */
 		if (i < NUMBER_OF_CTCPS && ctcp_cmd[i].repl)
 		{
-			if ((ptr = ctcp_cmd[i].repl(ctcp_cmd + i, from, to, ctcp_argument)))
-			{
-				strlcat(local_ctcp_buffer, ptr, sizeof local_ctcp_buffer);
-				new_free(&ptr);
-				continue;
-			}
+		    if ((ptr = ctcp_cmd[i].repl(ctcp_cmd + i, from, to, 
+						ctcp_argument)))
+		    {
+			strlcat(local_ctcp_buffer, ptr, 
+					sizeof local_ctcp_buffer);
+			new_free(&ptr);
+			pop_message_from(l);
+			continue;
+		    }
 		}
 
-		/* Set up the window level/logging */
-		lastlog_level = set_lastlog_msg_level(LOG_CTCP);
-		message_from(NULL, LOG_CTCP);
-
 		/* Toss it at the user.  */
-		if (do_hook(CTCP_REPLY_LIST, "%s %s %s %s", from, to, ctcp_command, ctcp_argument))
-			say("CTCP %s reply from %s: %s", ctcp_command, from, ctcp_argument);
-
-		/* Reset the window level/logging */
-		set_lastlog_msg_level(lastlog_level);
-
+		if (ctcp_cmd[i].flag & CTCP_TELLUSER)
+		{
+		    if (do_hook(CTCP_REPLY_LIST, "%s %s %s %s", 
+					from, to, ctcp_command, ctcp_argument))
+			say("CTCP %s reply from %s: %s", 
+					ctcp_command, from, ctcp_argument);
+		}
 		if (!(ctcp_cmd[i].flag & CTCP_NOLIMIT))
 			allow_ctcp_reply = 0;
+
+		pop_message_from(l);
 	}
 
-	if (in_ctcp_flag == -1)
-		in_ctcp_flag = 0;
+	in_ctcp_flag--;
 
 	/* 
 	 * local_ctcp_buffer is derived from 'str', so its always
@@ -953,12 +972,14 @@ void	send_ctcp (int type, const char *to, int datatag, const char *format, ...)
 	char 	putbuf [BIG_BUFFER_SIZE + 1],
 		*putbuf2;
 	int	len;
+	int	l;
 
 	/* Make sure that the final \001 doesnt get truncated */
 	if ((len = IRCD_BUFFER_SIZE - (12 + strlen(to))) <= 0)
 		return;				/* Whatever. */
 	putbuf2 = alloca(len);
 
+	l = message_from(to, LEVEL_CTCP);
 	if (format)
 	{
 		va_list args;
@@ -989,7 +1010,8 @@ void	send_ctcp (int type, const char *to, int datatag, const char *format, ...)
 	putbuf2[len - 2] = CTCP_DELIM_CHAR;
 	putbuf2[len - 1] = 0;
 
-	send_text(to, putbuf2, ctcp_type[type], 0);
+	send_text(from_server, to, putbuf2, ctcp_type[type], 0);
+	pop_message_from(l);
 }
 
 
@@ -1041,3 +1063,4 @@ static int split_CTCP (char *raw_message, char *ctcp_dest, char *after_ctcp)
 	strlcpy(after_ctcp, ctcp_end, IRCD_BUFFER_SIZE - 1);
 	return 0;		/* All done! */
 }
+
