@@ -1,9 +1,37 @@
+/* $EPIC: exec.c,v 1.39 2005/10/30 22:41:19 jnelson Exp $ */
 /*
  * exec.c: handles exec'd process for IRCII 
  *
- * Copyright 1990 Michael Sandrof
- * Copyright 1997 EPIC Software Labs
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1997, 2003 EPIC Software Labs
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #include "irc.h"
@@ -64,12 +92,14 @@ static	int	process_list_size = 0;
 
 static 	void 	handle_filedesc 	(Process *, int *, int, int);
 static 	void 	cleanup_dead_processes 	(void);
-static 	void 	ignore_process 		(int index);
+static 	void 	ignore_process 		(int);
+static 	void 	close_in		(int);
 static 	void 	kill_process 		(int, int);
 static 	void 	kill_all_processes 	(int signo);
 static 	int 	valid_process_index 	(int proccess);
 static 	int 	is_logical_unique 	(char *logical);
 static	int 	logical_to_index 	(const char *logical);
+static	void 	do_exec (int fd);
 
 /*
  * A nice array of the possible signals.  Used by the coredump trapping
@@ -88,13 +118,13 @@ static	int 	logical_to_index 	(const char *logical);
  */
 BUILT_IN_COMMAND(execcmd)
 {
-	char		*who = NULL,
-			*logical = NULL,
-			*redirect = NULL,
-			*flag;
+	const char	*who = NULL;
+	char		*logical = NULL;
+	const char	*flag;
+	const char 	*redirect = NULL;
 	unsigned 	refnum = 0;
+	size_t		len;
 	int		sig,
-			len,
 			i,
 			refnum_flag = 0,
 			logical_flag = 0;
@@ -111,8 +141,6 @@ BUILT_IN_COMMAND(execcmd)
 	 */
 	if (!*args)
 	{
-		int	i;
-
 		if (!process_list)
 		{
 			say("No processes are running");
@@ -122,8 +150,6 @@ BUILT_IN_COMMAND(execcmd)
 		say("Process List:");
 		for (i = 0; i < process_list_size; i++)
 		{
-			Process	*proc;
-
 			if (((proc = process_list[i])) == NULL)
 				continue;
 
@@ -151,14 +177,14 @@ BUILT_IN_COMMAND(execcmd)
 		 */
 		if (my_strnicmp(flag, "OUT", len) == 0)
 		{
-			if (doing_privmsg)
+			if (get_server_doing_privmsg(from_server))
 				redirect = "NOTICE";
 			else
 				redirect = "PRIVMSG";
 
-			if (!(who = get_channel_by_refnum(0)))
+			if (!(who = get_target_by_refnum(0)))
 			{
-			     say("No current channel in this window for -OUT");
+			     say("No query or channel in this window for -OUT");
 			     return;
 			}
 		}
@@ -193,7 +219,7 @@ BUILT_IN_COMMAND(execcmd)
 		 */
 		else if (my_strnicmp(flag, "MSG", len) == 0)
 		{
-			if (doing_privmsg)
+			if (get_server_doing_privmsg(from_server))
 				redirect = "NOTICE";
 			else
 				redirect = "PRIVMSG";
@@ -263,6 +289,19 @@ BUILT_IN_COMMAND(execcmd)
 			return;
 		}
 
+		/*
+		 * /EXEC -CLOSEIN close the processes STDIN,
+		 * in the hope that it will take the hint.
+		 */
+		else if (my_strnicmp(flag, "CLOSEIN", len) == 0)
+		{
+			if ((i = get_process_index(&args)) == -1)
+				return;
+
+			close_in(i);
+			return;
+		}
+
 		/*
 		 * /EXEC -NOTICE <target> redirects the output of an /exec 
 		 * to a specified target
@@ -356,7 +395,7 @@ BUILT_IN_COMMAND(execcmd)
 	 */
 	if (*args == '%')
 	{
-		int i;
+		int	l;
 
 		/*
 		 * Make sure the process is actually running
@@ -365,7 +404,7 @@ BUILT_IN_COMMAND(execcmd)
 			return;
 
 		proc = process_list[i];
-		message_to(refnum);
+		l = message_setall(refnum, NULL, LEVEL_OTHER);
 
 		/*
 		 * Check to see if the user wants to change windows
@@ -404,7 +443,20 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			else 
 				say("The name %s is not unique!", logical);
 		}
-		message_to(0);
+
+		/*
+		 * Change the stdout and stderr lines _if_ they are given.
+		 */
+		if (stdoutc)
+			malloc_strcpy(&proc->stdoutc, stdoutc);
+		if (stdoutpc)
+			malloc_strcpy(&proc->stdoutpc, stdoutpc);
+		if (stderrc)
+			malloc_strcpy(&proc->stderrc, stderrc);
+		if (stderrpc)
+			malloc_strcpy(&proc->stderrpc, stderrpc);
+
+		pop_message_from(l);
 	}
 
 	/*
@@ -415,9 +467,10 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		int	p0[2], p1[2], p2[2],
 			pid, cnt;
 		char	*shell,
-			*flag,
 			*arg;
-		char	*name = args;
+		char	*name;
+
+		name = LOCAL_COPY(args);
 
 		if (!is_logical_unique(logical))
 		{
@@ -434,12 +487,12 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		if (pipe(p0) || pipe(p1) || pipe(p2))
 		{
 			say("Unable to start new process: %s", strerror(errno));
-			new_close(p0[0]);
-			new_close(p0[1]);
-			new_close(p1[0]);
-			new_close(p1[1]);
-			new_close(p2[0]);
-			new_close(p2[1]);
+			close(p0[0]);
+			close(p0[1]);
+			close(p1[0]);
+			close(p1[1]);
+			close(p2[0]);
+			close(p2[1]);
 			return;
 		}
 
@@ -454,8 +507,6 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		 */
 		case 0:
 		{
-			int i;
-
 			/*
 			 * Fire up a new job control session,
 			 * Sever all ties we had with the parent ircII process
@@ -490,24 +541,27 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			 */
 			if (direct || !shell)
 			{
-				char	**args;
 				int	max;
+				char **my_args;
 
 				cnt = 0;
 				max = 5;
-				args = new_malloc(sizeof(char *) * max);
+				my_args = new_malloc(sizeof(char *) * max);
 				while ((arg = new_next_arg(name, &name)))
 				{
-					if (cnt == max)
+					my_args[cnt] = arg;
+
+					if (++cnt >= max)
 					{
 						max += 5;
-						RESIZE(args, char *, max);
+						RESIZE(my_args, char *, max);
 					}
-
-					args[cnt++] = arg;
 				}
-				args[cnt] = NULL;
-				execvp(args[0], args);
+				my_args[cnt] = NULL;
+				execvp(my_args[0], my_args);
+
+				printf("*** Error running program \"%s\": %s\n",
+						args, strerror(errno));
 			}
 
 			/*
@@ -519,13 +573,14 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 					flag = empty_string;
 
 				execl(shell, shell, flag, name, NULL);
+
+				printf("*** Error running program \"%s %s\": %s\n", 
+						shell, args, strerror(errno));
 			}
 
 			/*
 			 * Something really died if we got here
 			 */
-			printf("*** Error starting shell \"%s\": %s\n", 
-					shell, strerror(errno));
 			_exit(-1);
 			break;
 		}
@@ -535,12 +590,10 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		 */
 		default:
 		{
-			int	i;
-			Process	*proc = new_malloc(sizeof(Process));
-
-			new_close(p0[0]);
-			new_close(p1[1]);
-			new_close(p2[1]);
+			proc = new_malloc(sizeof(Process));
+			close(p0[0]);
+			close(p1[1]);
+			close(p2[1]);
 
 			/*
 			 * Init the proc list if neccesary
@@ -577,9 +630,9 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			/*
 			 * Fill in the new proc entry
 			 */
-			proc->name = m_strdup(name);
+			proc->name = malloc_strdup(name);
 			if (logical)
-				proc->logical = m_strdup(logical);
+				proc->logical = malloc_strdup(logical);
 			else
 				proc->logical = NULL;
 			proc->index = i;
@@ -590,7 +643,7 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			proc->refnum = refnum;
 			proc->redirect = NULL;
 			if (redirect)
-				proc->redirect = m_strdup(redirect);
+				proc->redirect = malloc_strdup(redirect);
 			proc->server = from_server;
 			proc->counter = 0;
 			proc->exited = 0;
@@ -600,34 +653,34 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			proc->who = NULL;
 			proc->disowned = 0;
 			if (who)
-				proc->who = m_strdup(who);
+				proc->who = malloc_strdup(who);
 			proc->dumb = 0;
 
 			if (stdoutc)
-				proc->stdoutc = m_strdup(stdoutc);
+				proc->stdoutc = malloc_strdup(stdoutc);
 			else
 				proc->stdoutc = NULL;
 
 			if (stdoutpc)
-				proc->stdoutpc = m_strdup(stdoutpc);
+				proc->stdoutpc = malloc_strdup(stdoutpc);
 			else
 				proc->stdoutpc = NULL;
 
 			if (stderrc)
-				proc->stderrc = m_strdup(stderrc);
+				proc->stderrc = malloc_strdup(stderrc);
 			else
 				proc->stderrc = NULL;
 
 			if (stderrpc)
-				proc->stderrpc = m_strdup(stderrpc);
+				proc->stderrpc = malloc_strdup(stderrpc);
 			else
 				proc->stderrpc = NULL;
 
 			if (endc)
 				add_process_wait(proc->index, endc);
 
-			new_open(proc->p_stdout);
-			new_open(proc->p_stderr);
+			new_open(proc->p_stdout, do_exec, NEWIO_READ, 1);
+			new_open(proc->p_stderr, do_exec, NEWIO_READ, 1);
 			break;
 		}
 		}
@@ -643,7 +696,7 @@ say("Output from process %d (%s) now going to you", i, proc->name);
  * are closed.  If EOF has been asserted on both, then  we mark the process
  * as being "dumb".  Once it is reaped (exited), it is expunged.
  */
-void 		do_processes (fd_set *rd)
+void 		do_exec (int fd)
 {
 	int	i;
 	int	limit;
@@ -659,16 +712,14 @@ void 		do_processes (fd_set *rd)
 		if (!proc)
 			continue;
 
-		if (proc->p_stdout != -1 && FD_ISSET(proc->p_stdout, rd))
+		if (proc->p_stdout != -1 && proc->p_stdout == fd)
 		{
-			FD_CLR(proc->p_stdout, rd);
 			handle_filedesc(proc, &proc->p_stdout, 
 					EXEC_PROMPT_LIST, EXEC_LIST);
 		}
 
-		if (proc->p_stderr != -1 && FD_ISSET(proc->p_stderr, rd))
+		if (proc->p_stderr != -1 && proc->p_stderr == fd)
 		{
-			FD_CLR(proc->p_stderr, rd);
 			handle_filedesc(proc, &proc->p_stderr,
 					EXEC_PROMPT_LIST, EXEC_ERRORS_LIST);
 		}
@@ -687,95 +738,110 @@ void 		do_processes (fd_set *rd)
 static void 	handle_filedesc (Process *proc, int *fd, int hook_nonl, int hook_nl)
 {
 	char 	exec_buffer[IO_BUFFER_SIZE + 1];
-	int	len;
+	ssize_t	len;
+	int	ofs;
+	const char *callback = NULL;
+	int	hook = -1;
+	int	l;
 
-	switch ((len = dgets(exec_buffer, *fd, 0)))	/* No buffering! */
+	/* No buffering! */
+	switch ((len = dgets(*fd, exec_buffer, IO_BUFFER_SIZE, 0))) 
 	{
-		case -1:	/* Something died */
+	    case -1:		/* Something died */
+	    {
+		*fd = new_close(*fd);
+		if (proc->p_stdout == -1 && proc->p_stderr == -1)
+			proc->dumb = 1;
+		return;				/* PUNT! */
+	    }
+
+	    case 0:		/* We didnt get a full line */
+	    {
+		/* 
+		 * XXX This is a hack.  dgets() can return 0 for a line
+		 * containing solely a newline, as well as a line that didn't
+		 * have a newline.  So we have to check to see if the line 
+		 * contains only a newline!
+		 */
+		if (exec_buffer[0] != '\n')
 		{
-			*fd = new_close(*fd);
-			if (proc->p_stdout == -1 && proc->p_stderr == -1)
-				proc->dumb = 1;
-			break;
+		    if (hook_nl == EXEC_LIST)
+		    {
+			if (proc->stdoutpc && *proc->stdoutpc)
+			    callback = proc->stdoutpc;
+		    }
+		    else if (hook_nl == EXEC_ERRORS_LIST)
+		    {
+			if (proc->stderrpc && *proc->stderrpc)
+			    callback = proc->stderrpc;
+		    }
+		    hook = hook_nonl;
+		    break;
 		}
-		case 0:		/* We didnt get a full line */
+
+		/* XXX HACK -- Line contains only a newline.  */
+		*exec_buffer = 0;
+		/* FALLTHROUGH */
+	    }
+
+	    default:		/* We got a full line */
+	    {
+		if (hook_nl == EXEC_LIST)
 		{
-			/* 
-			 * XXX XXX Major, world class hack here XXX XXX 
-			 * Since this problem has been addressed already
-			 * by newio2.c, it doesn't matter if we just fix the
-			 * symptom here since this code is all obsolete.
-			 */
-			if (exec_buffer[0] == '\n')
-			{
-				exec_buffer[0] = 0;
-				goto this_sucks;
-			}
-
-			     if (hook_nl == EXEC_LIST && proc->stdoutpc)
-				parse_line("EXEC", proc->stdoutpc, 
-						exec_buffer, 0, 0);
-			else if (hook_nl == EXEC_ERRORS_LIST && proc->stderrpc)
-				parse_line("EXEC", proc->stderrpc, 
-						exec_buffer, 0, 0);
-			else if (proc->logical)
-				do_hook(hook_nonl, "%s %s", 
-					proc->logical, exec_buffer);
-			else
-				do_hook(hook_nonl, "%d %s", 
-					proc->index, exec_buffer);
-
-			set_prompt_by_refnum(proc->refnum, exec_buffer);
-			update_input(UPDATE_ALL);
-			break;
+		    if (proc->stdoutc && *proc->stdoutc)
+			callback = proc->stdoutc;
 		}
-		default:	/* We got a full line */
+		else if (hook_nl == EXEC_ERRORS_LIST)
 		{
-this_sucks:
-			message_to(proc->refnum);
-			proc->counter++;
-
-			while (len > 0 && (exec_buffer[len] == '\n' ||
-					   exec_buffer[len] == '\r'))
-			{
-				exec_buffer[len--] = 0;
-			}
-
-			if (proc->redirect) 
-				redirect_text(proc->server, proc->who, 
-					exec_buffer, proc->redirect, 1);
-
-			     if (hook_nl == EXEC_LIST && proc->stdoutc)
-				parse_line("EXEC", proc->stdoutc, 
-						exec_buffer, 0, 0);
-			else if (hook_nl == EXEC_ERRORS_LIST && proc->stderrc)
-				parse_line("EXEC", proc->stderrc, 
-						exec_buffer, 0, 0);
-			else if (proc->logical)
-			{
-				if ((do_hook(hook_nl, "%s %s", 
-						proc->logical, exec_buffer)))
-					if (!proc->redirect)
-						put_it("%s", exec_buffer);
-			}
-			else
-			{
-				if ((do_hook(hook_nl, "%d %s", 
-						proc->index, exec_buffer)))
-					if (!proc->redirect)
-						put_it("%s", exec_buffer);
-			}
-
-			message_to(0);
+		    if (proc->stderrc && *proc->stderrc)
+			callback = proc->stderrc;
 		}
+		hook = hook_nl;
+		break;
+	    }
 	}
+
+	ofs = from_server;
+	from_server = proc->server;
+	if (proc->refnum)
+		l = message_setall(proc->refnum, NULL, LEVEL_OTHER);
+	else
+		l = message_from(NULL, LEVEL_OTHER);
+
+	proc->counter++;
+
+	while (len > 0 && (exec_buffer[len - 1] == '\n' ||
+			   exec_buffer[len - 1] == '\r'))
+	     exec_buffer[--len] = 0;
+
+	if (proc->redirect) 
+	     redirect_text(proc->server, proc->who, 
+				exec_buffer, proc->redirect, 1);
+
+	if (callback)
+	    call_lambda_command("EXEC", callback, exec_buffer);
+	else if (proc->logical)
+	{
+	     if ((do_hook(hook, "%s %s", proc->logical, exec_buffer)))
+		if (!proc->redirect)
+		    put_it("%s", exec_buffer);
+	}
+	else
+	{
+	    if ((do_hook(hook, "%d %s", proc->index, exec_buffer)))
+		if (!proc->redirect)
+		    put_it("%s", exec_buffer);
+	}
+
+	pop_message_from(l);
+	from_server = ofs;
 }
 
 /*
  * get_child_exit: This looks for dead child processes of the client.
  * There are two main sources of dead children:  Either an /exec'd process
  * has exited, or the client has attempted to fork() off a helper process
- * (such as wserv or gzip) and that process has choked on itself.  
+ * (such as wserv or gzip) and that process has choked on itself.
  *
  * When SIGCHLD is recieved, the global variable 'dead_children_processes'
  * is incremented.  When this function is called, we go through and call
@@ -808,9 +874,9 @@ int 		get_child_exit (pid_t wanted)
 		 */
 		if (wanted != -1 && pid == wanted)
 		{
-			if (WIFEXITED(status))   
+			if (WIFEXITED(status))
 				return WEXITSTATUS(status);
-			if (WIFSTOPPED(status))  
+			if (WIFSTOPPED(status))
 				return -(WSTOPSIG(status));
 			if (WIFSIGNALED(status)) 
 				return -(WTERMSIG(status));
@@ -869,6 +935,7 @@ void 		clean_up_processes (void)
 		say("Closing all left over exec's");
 		kill_all_processes(SIGTERM);
 		sleep(2);
+		get_child_exit(-1);
 		kill_all_processes(SIGKILL);
 	}
 }
@@ -884,40 +951,29 @@ int 		text_to_process (int proc_index, const char *text, int show)
 {
 	Process	*	proc;
 	char	*	my_buffer;
+	size_t	size;
 
 	if (valid_process_index(proc_index) == 0)
 		return 1;
 
 	proc = process_list[proc_index];
 
-	message_to(proc->refnum);
 	if (show)
+	{
+		int	l = message_setall(proc->refnum, NULL, LEVEL_OTHER);
 		put_it("%s%s", get_prompt_by_refnum(proc->refnum), text);
-	message_to(0);
+		pop_message_from(l);
+	}
 
-	my_buffer = alloca(strlen(text) + 2);
-	/* Obviously this is safe. */
-	strcpy(my_buffer, text);
-	strcat(my_buffer, "\n");
+	size = strlen(text) + 2;
+	my_buffer = alloca(size);
+	snprintf(my_buffer, size, "%s\n", text);
 	write(proc->p_stdin, my_buffer, strlen(my_buffer));
 	set_prompt_by_refnum(proc->refnum, empty_string);
 
 	return (0);
 }
 
-/*
- * When a server goes away, re-assign all of the /exec's that are
- * current bound to that server.
- */
-void 		exec_server_delete (int i)
-{
-	int	j;
-
-	for (j = 0; j < process_list_size; j++)
-		if (process_list[j] && process_list[j]->server >= i)
-			process_list[j]->server--;
-}
-
 /*
  * This adds a new /wait %proc -cmd   entry to a running process.
  */
@@ -928,7 +984,7 @@ void 		add_process_wait (int proc_index, const char *cmd)
 
 	new_ewl = new_malloc(sizeof(List));
 	new_ewl->next = NULL;
-	new_ewl->name = m_strdup(cmd);
+	new_ewl->name = malloc_strdup(cmd);
 
 	if ((posn = proc->waitcmds))
 	{
@@ -977,12 +1033,14 @@ static void 	cleanup_dead_processes (void)
 	int	i;
 	List	*cmd,
 		*next;
-	Process *dead, *proc;
+	Process *deadproc, *proc;
 	char	*exit_info;
+	int	old_from_server, l;
 
 	if (!process_list)
 		return;		/* Nothing to do */
 
+	old_from_server = from_server;
 	for (i = 0; i < process_list_size; i++)
 	{
 		if (!(proc = process_list[i]))
@@ -996,38 +1054,40 @@ static void 	cleanup_dead_processes (void)
 		if ((!proc->exited || !proc->dumb) && !proc->disowned)
 			continue;		/* Not really dead yet */
 
-		dead = process_list[i];
+		deadproc = process_list[i];
 		process_list[i] = NULL;
 
 		/*
 		 * First thing to do is fill out the exit information
 		 */
-		if (dead->logical)
+		if (deadproc->logical)
 		{
-			size_t	len = strlen(dead->logical) + 25;
+			size_t	len = strlen(deadproc->logical) + 25;
 
 			exit_info = alloca(len);
 			snprintf(exit_info, len, "%s %d %d", 
-					dead->logical, dead->termsig,
-					dead->retcode);
+					deadproc->logical, deadproc->termsig,
+					deadproc->retcode);
 		}
 		else
 		{
 			exit_info = alloca(40);
 			snprintf(exit_info, 32, "%d %d %d",
-				dead->index, dead->termsig, dead->retcode);
+				deadproc->index, deadproc->termsig, deadproc->retcode);
 		}
 
+		from_server = deadproc->server;
+		l = message_from(NULL, LEVEL_OTHER);
 
 		/*
 		 * First thing we do is run any /wait %proc -cmd commands
 		 */
-		next = dead->waitcmds;
-		dead->waitcmds = NULL;
+		next = deadproc->waitcmds;
+		deadproc->waitcmds = NULL;
 		while ((cmd = next))
 		{
 			next = cmd->next;
-			parse_line(NULL, cmd->name, exit_info, 0, 0);
+			call_lambda_command("WAITPROC", cmd->name, exit_info);
 			new_free(&cmd->name);
 			new_free((char **)&cmd);
 		}
@@ -1037,28 +1097,44 @@ static void 	cleanup_dead_processes (void)
 		 */
 		if (do_hook(EXEC_EXIT_LIST, "%s", exit_info))
 		{
-			if (get_int_var(NOTIFY_ON_TERMINATION_VAR))
+		    if (get_int_var(NOTIFY_ON_TERMINATION_VAR))
+		    {
+			if (deadproc->termsig > 0 && deadproc->termsig < NSIG)
 			{
-				if (dead->termsig > 0 && dead->termsig < NSIG)
-	say("Process %d (%s) terminated with signal %s (%d)", 
-	   dead->index, dead->name, sys_siglist[dead->termsig], dead->termsig);
-				else if (dead->disowned)
-	say("Process %d (%s) disowned", dead->index, dead->name);
-				else
-	say("Process %d (%s) terminated with return code %d", 
-		dead->index, dead->name, dead->retcode);
+				say("Process %d (%s) terminated "
+					"with signal %s (%d)", 
+				   deadproc->index, deadproc->name, 
+				   sys_siglist[deadproc->termsig], 
+				   deadproc->termsig);
 			}
+			else if (deadproc->disowned)
+			{
+				say("Process %d (%s) disowned", 
+				   deadproc->index, deadproc->name);
+			}
+			else
+			{
+				say("Process %d (%s) terminated "
+					"with return code %d", 
+				   deadproc->index, deadproc->name, 
+				   deadproc->retcode);
+			}
+		    }
 		}
+		pop_message_from(l);
 
-		if (dead->p_stdin != -1)
-			close(dead->p_stdin);
-		new_close(dead->p_stdout);
-		new_close(dead->p_stderr);
-		new_free(&dead->name);
-		new_free(&dead->logical);
-		new_free(&dead->who);
-		new_free(&dead->redirect);
-		new_free((char **)&dead);
+		deadproc->p_stdin = new_close(deadproc->p_stdin);
+		deadproc->p_stdout = new_close(deadproc->p_stdout);
+		deadproc->p_stderr = new_close(deadproc->p_stderr);
+		new_free(&deadproc->name);
+		new_free(&deadproc->logical);
+		new_free(&deadproc->who);
+		new_free(&deadproc->redirect);
+		new_free(&deadproc->stdoutc);
+		new_free(&deadproc->stdoutpc);
+		new_free(&deadproc->stderrc);
+		new_free(&deadproc->stderrpc);
+		new_free((char **)&deadproc);
 	}
 
 	/*
@@ -1075,6 +1151,8 @@ static void 	cleanup_dead_processes (void)
 		process_list_size = i + 1;
 		RESIZE(process_list, Process, process_list_size);
 	}
+
+	from_server = old_from_server;
 }
 
 
@@ -1086,14 +1164,14 @@ static void 	cleanup_dead_processes (void)
  * condition on its output fd, so it will probably either take the hint, or
  * its output will go the bit bucket (which we want to happen)
  */
-static void 	ignore_process (int index)
+static void 	ignore_process (int idx)
 {
 	Process *proc;
 
-	if (valid_process_index(index) == 0)
+	if (valid_process_index(idx) == 0)
 		return;
 
-	proc = process_list[index];
+	proc = process_list[idx];
 	if (proc->p_stdin != -1)
 		proc->p_stdin = new_close(proc->p_stdin);
 	if (proc->p_stdout != -1)
@@ -1104,6 +1182,22 @@ static void 	ignore_process (int index)
 	proc->dumb = 1;
 }
 
+/*
+ * close_in:  When we are finished with the process but still want the
+ * rest of its output, we close its input, and hopefully it will get the
+ * message and close up shop.
+ */
+static void 	close_in (int idx)
+{
+	Process *proc;
+
+	if (valid_process_index(idx) == 0)
+		return;
+
+	proc = process_list[idx];
+	proc->p_stdin = new_close(proc->p_stdin);
+}
+
 
 
 
@@ -1116,7 +1210,8 @@ static void 	ignore_process (int index)
  */
 static void 	kill_process (int kill_index, int sig)
 {
-	pid_t pgid;
+	pid_t	pgid;
+	int	old_from_server, l;
 
 	if (!process_list || kill_index > process_list_size || 
 			!process_list[kill_index])
@@ -1125,10 +1220,17 @@ static void 	kill_process (int kill_index, int sig)
 		return;
 	}
 
+	old_from_server = from_server;
+	from_server = process_list[kill_index]->server;
+	l = message_from(NULL, LEVEL_OTHER);
+
 	say("Sending signal %s (%d) to process %d: %s", 
 		sys_siglist[sig], sig, kill_index, 
 		process_list[kill_index]->name);
 
+	pop_message_from(l);
+	from_server = old_from_server;
+
 #ifdef HAVE_GETPGID
 	pgid = getpgid(process_list[kill_index]->pid);
 #else
