@@ -1,4 +1,4 @@
-/* $EPIC:$ */
+/* $EPIC: expr2.c,v 1.33 2006/06/06 05:08:48 jnelson Exp $ */
 /*
  * Zsh: math.c,v 3.1.2.1 1997/06/01 06:13:15 hzoli Exp 
  * math.c - mathematical expression evaluation
@@ -33,12 +33,12 @@
  *
  */
 /*
- * Substantial modifications by Jeremy Nelson which are
- * Coypright 1998, 2000 EPIC Software Labs, All rights reserved.
+ * Substantial modifications by EPIC Software Labs which are
+ * Coypright 1998, 2003 EPIC Software Labs, All rights reserved.
  *
  * You may distribute this file under the same terms as the above, by 
- * including the parties "Jeremy Nelson" and "EPIC Software Labs" to the 
- * limitations of liability and the express disclaimer of all warranties.  
+ * including the parties "EPIC Software Labs" to the limitations of 
+ * liability and the express disclaimer of all warranties.  
  * This software is provided "AS IS".
  */
 #include <math.h>
@@ -83,7 +83,7 @@
 /*
  * Question:  Why is everything stored in a struct?
  * 
- * Answer: All the information for each expression is stored in a struct.  
+ * Answer: All the information for each expression is stored in a struct.
  * This is done so that there are no global variables in use (they're all 
  * collected making them easier to handle), and makes re-entrancy possible
  * since you don't have to worry about whether or not all of the state
@@ -97,7 +97,7 @@
  * a clean way to clean up after an expression (no memory leaks), and
  * permit a reference to any operator to persist for the entire lifetime
  * of the expression, without having to worry about who might be holding
- * a reference to an operand (tokens never change over their lifetime).  
+ * a reference to an operand (tokens never change over their lifetime).
  * By refering to each token through an integer, rather than a pointer, 
  * we can also prevent stale pointers, which can cause crashes.
  * 
@@ -145,7 +145,7 @@ typedef struct TOKEN_type
 	char *	lval;			/* Cached unexpanded variable name */
 	char *	raw_value;		/* Cached unexpected string */
 	char *	expanded_value;		/* Cached full expanded string */
-	long	integer_value;		/* Cached integer value */
+	intmax_t integer_value;		/* Cached integer value */
 	double	float_value;		/* Cached floating point value */
 	short	boolean_value;		/* Cached boolean value */
 } SYMBOL;
@@ -219,23 +219,20 @@ typedef struct
 
 	TOKEN	last_token;
 
-	char	*args;
-	int	*args_flag;
+	const char	*args;
 } expr_info;
 
 /* 
- * Useful macro to get at a specific token.  
+ * Useful macro to get at a specific token.
  * 'c' is the expression context, 'v' is the token handle.
  */
 #define TOK(c, v) 	c->tokens[v]
 
 /* Forward function references */
-__inline static	TOKEN	tokenize_raw (expr_info *c, char *t);
+__inline static	TOKEN	tokenize_raw (expr_info *c, const char *t);
 	static	char *	after_expando_special (expr_info *c);
-	static	char *	canon_number (char *input);
 	static	char *	alias_special_char (char **buffer, char *ptr, 
-					const char *args, char *quote_em, 
-					int *args_flag);
+					const char *args, char *quote_em);
 
 
 /******************** EXPRESSION CONSTRUCTOR AND DESTRUCTOR ****************/
@@ -243,7 +240,7 @@ __inline static	TOKEN	tokenize_raw (expr_info *c, char *t);
  * Clean the expression context pointed to by 'c'.
  * This function must be called before you call mathparse().
  */
-void	setup_expr_info (expr_info *c)
+static void	setup_expr_info (expr_info *c)
 {
 	int	i;
 
@@ -251,6 +248,7 @@ void	setup_expr_info (expr_info *c)
 	c->noeval = 0;
 	c->operand = 1;
 	c->token = 0;
+#ifdef NO_CHEATING
 	for (i = 0; i <= TOKENCOUNT; i++)
 	{
 		TOK(c, i).used = USED_NONE;
@@ -263,6 +261,10 @@ void	setup_expr_info (expr_info *c)
 	}
 	for (i = 0; i <= STACKSZ; i++)
 		c->stack[i] = 0;
+#else
+	memset(c->tokens, 0, sizeof(c->tokens));
+	memset(c->stack, 0, sizeof(c->stack));
+#endif
 	c->sp = -1;
 	c->mtok = 0;
 	c->errflag = 0;
@@ -274,7 +276,7 @@ void	setup_expr_info (expr_info *c)
  * Clean up the expression context pointed to by 'c'.
  * This function must be called after you call mathparse().
  */
-void 	destroy_expr_info (expr_info *c)
+static void 	destroy_expr_info (expr_info *c)
 {
 	int	i;
 
@@ -314,7 +316,7 @@ void 	destroy_expr_info (expr_info *c)
  */
 
 enum LEX {
-	M_INPAR,
+	M_FUNCTION,	M_INPAR,
 	NOT, 		COMP, 		PREMINUS,	PREPLUS,
 			UPLUS,		UMINUS,		STRLEN,
 			WORDC,		DEREF,
@@ -324,7 +326,7 @@ enum LEX {
 	SHLEFT,		SHRIGHT,
 	LES,		LEQ,		GRE,		GEQ,
 	MATCH,		NOMATCH,
-	DEQ,		NEQ,
+	DEQ,		NEQ,		LITEQ,		NOTLITEQ,
 	AND,
 	XOR,
 	OR,
@@ -363,7 +365,7 @@ enum LEX {
  */
 static	int	prec[TOKCOUNT] = 
 {
-	1,
+	1,		1,
 	2,		2,		2,		2,
 			2,		2,		2,
 			2,		2,
@@ -373,7 +375,7 @@ static	int	prec[TOKCOUNT] =
 	6,		6,
 	7,		7,		7,		7,
 	8,		8,
-	9,		9,
+	9,		9,		9,		9,
 	10,
 	11,
 	12,
@@ -417,7 +419,7 @@ static	int	prec[TOKCOUNT] =
  */
 static 	int 	assoc[TOKCOUNT] =
 {
-	LR,
+	LR,		LR,
 	RL,		RL,		RL,		RL,
 			RL,		RL,		RL,
 			RL,		RL,
@@ -427,7 +429,7 @@ static 	int 	assoc[TOKCOUNT] =
 	LR,		LR,
 	LR,		LR,		LR,		LR,
 	LR,		LR,
-	LR,		LR,
+	LR,		LR,		LR,		LR,
 	LR,
 	LR,
 	LR,
@@ -451,7 +453,7 @@ static 	int 	assoc[TOKCOUNT] =
 
 /* ********************* CREATE NEW SYMBOLS AND TOKENS ********************/
 /* Lvalues are stored via this routine */
-__inline static	TOKEN		tokenize_lval (expr_info *c, char *t)
+__inline static	TOKEN		tokenize_lval (expr_info *c, const char *t)
 {
 	if (c->token >= TOKENCOUNT)
 	{
@@ -472,7 +474,7 @@ __inline static	TOKEN		tokenize_lval (expr_info *c, char *t)
  * result of operators) directly create "expanded" tokens.
  */
 /* THIS FUNCTION MAKES A NEW COPY OF 'T'.  YOU MUST DISPOSE OF 'T' YOURSELF */
-__inline static	TOKEN		tokenize_raw (expr_info *c, char *t)
+__inline static	TOKEN		tokenize_raw (expr_info *c, const char *t)
 {
 	if (c->token >= TOKENCOUNT)
 	{
@@ -508,7 +510,7 @@ __inline static	TOKEN		tokenize_expanded (expr_info *c, char *t)
  * is both expensive, and unnecesary.  This type of token is created by
  * integer-only operations like &, |, ^, <<, >>.
  */
-__inline static	TOKEN		tokenize_integer (expr_info *c, long t)
+__inline static	TOKEN		tokenize_integer (expr_info *c, intmax_t t)
 {
 	if (c->token >= TOKENCOUNT)
 	{
@@ -582,7 +584,7 @@ __inline static	TOKEN		tokenize_bool (expr_info *c, BooL t)
  * the most information, starting with expanded string, to the boolean
  * value. 
  */ 
-__inline static	char *	get_token_raw (expr_info *c, TOKEN v)
+__inline static	const char *	get_token_raw (expr_info *c, TOKEN v)
 {
 	if (v == MAGIC_TOKEN)	/* Magic token */
 		return c->args;			/* XXX Probably wrong */
@@ -598,10 +600,6 @@ __inline static	char *	get_token_raw (expr_info *c, TOKEN v)
 
 	if (c->noeval)
 		return get_token_raw(c, 0);
-#if 0
-		panic("c->noeval is not valid here. [1]");
-#endif
-
 
 	if ((TOK(c, v).used & USED_RAW) == 0)
 	{
@@ -612,15 +610,17 @@ __inline static	char *	get_token_raw (expr_info *c, TOKEN v)
 		else if (TOK(c, v).used & USED_FLOAT)
 		{
 			TOK(c, v).raw_value = 
-				m_sprintf("%f", TOK(c, v).float_value);
+				malloc_sprintf(NULL, "%f", TOK(c, v).float_value);
 			canon_number(TOK(c, v).raw_value);
 		}
 		else if (TOK(c, v).used & USED_INTEGER)
 			TOK(c, v).raw_value = 
-				m_sprintf("%ld", TOK(c, v).integer_value);
+				malloc_sprintf(NULL, INTMAX_FORMAT, 
+						TOK(c, v).integer_value);
 		else if (TOK(c, v).used & USED_BOOLEAN)
 			TOK(c, v).raw_value = 
-				m_sprintf("%d", TOK(c, v).boolean_value);
+				malloc_sprintf(NULL, "%d",
+						TOK(c, v).boolean_value);
 		else if (TOK(c, v).used & USED_LVAL)
 		{
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -628,7 +628,7 @@ __inline static	char *	get_token_raw (expr_info *c, TOKEN v)
 					v, TOK(c, v).lval);
 
 			TOK(c, v).raw_value = expand_alias(TOK(c, v).lval, 
-					c->args, c->args_flag, NULL);
+								c->args);
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell(">>> Expanded var name [%d]: [%s] to [%s]",
@@ -648,7 +648,7 @@ __inline static	char *	get_token_raw (expr_info *c, TOKEN v)
  * and only if (TOK(c, v).used & USED_LVAL)!  Otherwise, this returns
  * NULL and emits an error.
  */
-__inline static char *	get_token_lval (expr_info *c, TOKEN v)
+__inline static const char *	get_token_lval (expr_info *c, TOKEN v)
 {
 	if (v == MAGIC_TOKEN)	/* Magic token */
 		return c->args;			/* XXX Probably wrong */
@@ -662,13 +662,7 @@ __inline static char *	get_token_lval (expr_info *c, TOKEN v)
 	if (v == 0)
 		return NULL;		/* Suppress the operation entirely */
 	else if (c->noeval)
-	{
 		return NULL;		/* Suppress the operation entirely */
-#if 0
-		return get_token_raw(c, 0);
-		panic("c->noeval is not valid here. [0]");
-#endif
-	}
 	else if (((TOK(c, v).used & USED_LVAL) == 0))
 	{
 		error("Token [%d] is not an lvalue", v);
@@ -678,14 +672,13 @@ __inline static char *	get_token_lval (expr_info *c, TOKEN v)
 		return get_token_raw(c, v);
 }
 
-
 /*
  * Get the "expanded" representation of the token.  The first time you
  * call this function, it gets the "raw" value, if there is one, and then
  * call expand_alias() to get the value.  Of course, there is always a
  * "raw" value, as get_token_raw() can convert from anything.
  */
-__inline static	char *	get_token_expanded (expr_info *c, TOKEN v)
+__inline static	const char *	get_token_expanded (expr_info *c, TOKEN v)
 {
 	if (v == MAGIC_TOKEN)	/* Magic token */
 		return c->args;
@@ -701,9 +694,6 @@ __inline static	char *	get_token_expanded (expr_info *c, TOKEN v)
 
 	if (c->noeval)
 		return get_token_raw(c, 0);
-#if 0
-		panic("c->noeval is not valid here. [2]");
-#endif
 
 	if (x_debug & DEBUG_NEW_MATH_DEBUG)
 		yell(">>> Getting token [%d] now.", v);
@@ -729,9 +719,9 @@ __inline static	char *	get_token_expanded (expr_info *c, TOKEN v)
 					v, myval);
 
 			alias_special_char(&buffer, myval, c->args, 
-					NULL, c->args_flag);
+					NULL);
 			if (!buffer)
-				buffer = m_strdup(empty_string);
+				buffer = malloc_strdup(empty_string);
 			TOK(c, v).expanded_value = buffer;
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -751,9 +741,7 @@ __inline static	char *	get_token_expanded (expr_info *c, TOKEN v)
 					v, myval);
 
 			TOK(c, v).used |= USED_EXPANDED;
-			TOK(c, v).expanded_value = 
-				expand_alias(myval, c->args, 
-						c->args_flag, NULL);
+			TOK(c, v).expanded_value = expand_alias(myval, c->args);
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("<<< Expanded token [%d]: [%s] to: [%s]", 
@@ -768,15 +756,41 @@ __inline static	char *	get_token_expanded (expr_info *c, TOKEN v)
 	return TOK(c, v).expanded_value;
 }
 
+/*
+ * This is seriously funky.  In the old math parser, function calls could
+ * be made upon any rvalue (ick), and CrazyEddy asked me nicely to implement
+ * this backwards compatability, and how could I ever say no to him? :-D
+ */
+__inline static const char *	get_token_fname (expr_info *c, TOKEN v)
+{
+	if (v == MAGIC_TOKEN)	/* Magic token */
+		return c->args;			/* XXX Probably wrong */
+
+	if (v < 0 || v >= c->token)
+	{
+		error("Token index [%d] is out of range", v);
+		return NULL;			/* No lvalue here! */
+	}
+
+	if (v == 0)
+		return NULL;		/* Suppress the operation entirely */
+	else if (c->noeval)
+		return NULL;		/* Suppress the operation entirely */
+	else if (TOK(c, v).used & USED_LVAL)
+		return get_token_raw(c, v);
+	else
+		return get_token_expanded(c, v);
+}
+
 /*
  * Get the integer representation of the token.  The first time you call
  * this function, it calls atof() on the "raw" value, if there is one, 
  * to get the result.
  */
-__inline static	long	get_token_integer (expr_info *c, TOKEN v)
+__inline static	intmax_t	get_token_integer (expr_info *c, TOKEN v)
 {
 	if (v == MAGIC_TOKEN)	/* Magic token */
-		return atol(c->args);		/* XXX Probably wrong */
+		return STR2INT(c->args);		/* XXX Probably wrong */
 
 	if (v < 0 || v >= c->token)
 	{
@@ -786,10 +800,10 @@ __inline static	long	get_token_integer (expr_info *c, TOKEN v)
 	
 	if ((TOK(c, v).used & USED_INTEGER) == 0)
 	{
-		char *	myval = get_token_expanded(c, v);
+		const char *	myval = get_token_expanded(c, v);
 
 		TOK(c, v).used |= USED_INTEGER;
-		TOK(c, v).integer_value = atol(myval);
+		TOK(c, v).integer_value = STR2INT(myval);
 	}
 	return TOK(c, v).integer_value;
 }
@@ -810,7 +824,7 @@ __inline static	double	get_token_float (expr_info *c, TOKEN v)
 	
 	if ((TOK(c, v).used & USED_FLOAT) == 0)
 	{
-		char *	myval = get_token_expanded(c, v);
+		const char *	myval = get_token_expanded(c, v);
 
 		TOK(c, v).used |= USED_FLOAT;
 		TOK(c, v).float_value = atof(myval);
@@ -834,7 +848,7 @@ __inline static	BooL	get_token_boolean (expr_info *c, TOKEN v)
 	
 	if ((TOK(c, v).used & USED_BOOLEAN) == 0)
 	{
-		char *	myval = get_token_expanded(c, v);
+		const char *	myval = get_token_expanded(c, v);
 
 		TOK(c, v).used |= USED_BOOLEAN;
 		TOK(c, v).boolean_value = check_val(myval);
@@ -877,7 +891,7 @@ __inline static	TOKEN	push_float (expr_info *c, double val)
 	return push_token(c, tokenize_float(c, val));
 }
 
-__inline static	TOKEN	push_integer (expr_info *c, long val)
+__inline static	TOKEN	push_integer (expr_info *c, intmax_t val)
 { 
 	return push_token(c, tokenize_integer(c, val));
 }
@@ -887,7 +901,7 @@ __inline static TOKEN	push_boolean (expr_info *c, BooL val)
 	return push_token(c, tokenize_bool(c, val));
 }
 
-__inline static TOKEN	push_lval (expr_info *c, char *val)
+__inline static TOKEN	push_lval (expr_info *c, const char *val)
 {
 	return push_token(c, tokenize_lval(c, val));
 }
@@ -931,7 +945,7 @@ __inline static	double		pop_float (expr_info *c)
 	return get_token_float(c, pop_token(c));
 }
 
-__inline static	char *	pop_expanded (expr_info *c)
+__inline static	const char *	pop_expanded (expr_info *c)
 {
 	return get_token_expanded(c, pop_token(c));
 }
@@ -953,13 +967,13 @@ __inline static	void	pop_2_floats (expr_info *c, double *a, double *b)
 	*a = pop_float(c);
 }
 
-__inline static	void	pop_2_integers (expr_info *c, long *a, long *b)
+__inline static	void	pop_2_integers (expr_info *c, intmax_t *a, intmax_t *b)
 {
 	*b = pop_integer(c);
 	*a = pop_integer(c);
 }
 
-__inline static void	pop_2_strings (expr_info *c, char **s, char **t)
+__inline static void	pop_2_strings (expr_info *c, const char **s, const char **t)
 {
 	*t = pop_expanded(c);
 	*s = pop_expanded(c);
@@ -989,12 +1003,12 @@ __inline static	void	pop_3_tokens (expr_info *c, BooL *a, TOKEN *v, TOKEN *w)
  * This is the reducer.  It takes the relevant arguments off the argument
  * stack and then performs the neccesary operation on them.
  */
-static void	op (expr_info *cx, int what)
+static void	reduce (expr_info *cx, int what)
 {
 	double	a, b;
 	BooL	c, d;
-	long	i, j;
-	char *s, *t;
+	intmax_t i, j;
+	const char *s, *t;
 	TOKEN	v, w;
 
 	if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -1009,10 +1023,21 @@ static void	op (expr_info *cx, int what)
 	if (cx->errflag)
 		return;		/* Dont parse on an error */
 
+/* Check to see if we are evaluating the expression at this point. */
+#define CHECK_NOEVAL							\
+	if (cx->noeval)							\
+	{								\
+		if (x_debug & DEBUG_NEW_MATH_DEBUG)			\
+			yell("O: Operation short-circuited");		\
+		push_token(cx, 0);					\
+		break;							\
+	}
+
 /* Perform an ordinary garden variety floating point binary operation. */
 #define BINARY_FLOAT(floatop)						\
 	{ 								\
 		pop_2_floats(cx, &a, &b); 				\
+		CHECK_NOEVAL						\
 		push_float(cx, (floatop));				\
 		if (x_debug & DEBUG_NEW_MATH_DEBUG) 			\
 			yell("O: %s (%f %f) -> %f", 			\
@@ -1024,9 +1049,11 @@ static void	op (expr_info *cx, int what)
 #define BINARY_INTEGER(intop)						\
 	{ 								\
 		pop_2_integers(cx, &i, &j); 				\
+		CHECK_NOEVAL						\
 		push_integer(cx, (intop)); 				\
 		if (x_debug & DEBUG_NEW_MATH_DEBUG) 			\
-			yell("O: %s (%ld %ld) -> %ld", 			\
+			yell("O: %s (" INTMAX_FORMAT " " INTMAX_FORMAT	\
+					 ") -> " INTMAX_FORMAT, 	\
 				#intop, i, j, intop); 			\
 		break; 							\
 	}
@@ -1035,6 +1062,7 @@ static void	op (expr_info *cx, int what)
 #define BINARY_BOOLEAN(boolop) 						\
 	{ 								\
 		pop_2_booleans(cx, &c, &d); 				\
+		CHECK_NOEVAL						\
 		push_boolean(cx, (boolop)); 				\
 		if (x_debug & DEBUG_NEW_MATH_DEBUG) 			\
 			yell("O: %s (%d %d) -> %d", 			\
@@ -1046,6 +1074,7 @@ static void	op (expr_info *cx, int what)
 #define BINARY_FLOAT_NOZERO(floatop)					\
 	{ 								\
 		pop_2_floats(cx, &a, &b); 				\
+		CHECK_NOEVAL						\
 		if (b == 0.0) 						\
 		{ 							\
 			if (x_debug & DEBUG_NEW_MATH_DEBUG) 		\
@@ -1068,10 +1097,12 @@ static void	op (expr_info *cx, int what)
 #define BINARY_INTEGER_NOZERO(intop)					\
 	{ 								\
 		pop_2_integers(cx, &i, &j); 				\
+		CHECK_NOEVAL						\
 		if (j == 0) 						\
 		{ 							\
 			if (x_debug & DEBUG_NEW_MATH_DEBUG) 		\
-				yell("O: %s (%ld %ld) -> []", 		\
+				yell("O: %s (" INTMAX_FORMAT " " INTMAX_FORMAT \
+						") -> []", 		\
 					#intop, i, j); 			\
 			error("Division by zero"); 			\
 			push_token(cx, 0);				\
@@ -1079,7 +1110,8 @@ static void	op (expr_info *cx, int what)
 		else 							\
 		{ 							\
 			if (x_debug & DEBUG_NEW_MATH_DEBUG) 		\
-			    yell("O: %s (%ld %ld) -> %ld", 		\
+			    yell("O: %s (" INTMAX_FORMAT " " INTMAX_FORMAT \
+						") -> " INTMAX_FORMAT,  \
 					#intop, i, j, intop); 		\
 			push_integer(cx, (intop)); 			\
 		} 							\
@@ -1100,11 +1132,13 @@ static void	op (expr_info *cx, int what)
 #define IMPLIED_INTEGER(intop) 						\
 	{ 								\
 		GET_LVAL_RVAL						\
+		CHECK_NOEVAL						\
 		i = get_token_integer(cx, v);				\
 		j = get_token_integer(cx, w);				\
 									\
 		if (x_debug & DEBUG_NEW_MATH_DEBUG) 			\
-			yell("O: %s = %s (%ld %ld) -> %ld",  		\
+			yell("O: %s = %s (" INTMAX_FORMAT " " INTMAX_FORMAT \
+					") -> " INTMAX_FORMAT, 		\
 				s, #intop, i, j, intop); 		\
 									\
 		w = tokenize_integer(cx, (intop));			\
@@ -1118,6 +1152,7 @@ static void	op (expr_info *cx, int what)
 #define IMPLIED_FLOAT(floatop) 						\
 	{ 								\
 		GET_LVAL_RVAL						\
+		CHECK_NOEVAL						\
 		a = get_token_float(cx, v);				\
 		b = get_token_float(cx, w);				\
 									\
@@ -1136,6 +1171,7 @@ static void	op (expr_info *cx, int what)
 #define IMPLIED_BOOLEAN(boolop) 					\
 	{ 								\
 		GET_LVAL_RVAL						\
+		CHECK_NOEVAL						\
 		c = get_token_boolean(cx, v);				\
 		d = get_token_boolean(cx, w);				\
 									\
@@ -1158,6 +1194,7 @@ static void	op (expr_info *cx, int what)
 #define IMPLIED_FLOAT_NOZERO(floatop) 					\
 	{ 								\
 		GET_LVAL_RVAL						\
+		CHECK_NOEVAL						\
 		a = get_token_float(cx, v);				\
 		b = get_token_float(cx, w);				\
 									\
@@ -1190,13 +1227,15 @@ static void	op (expr_info *cx, int what)
 #define IMPLIED_INTEGER_NOZERO(intop) 					\
 	{ 								\
 		GET_LVAL_RVAL						\
+		CHECK_NOEVAL						\
 		i = get_token_float(cx, v);				\
 		j = get_token_float(cx, w);				\
 									\
 		if (j == 0) 						\
 		{ 							\
 			if (x_debug & DEBUG_NEW_MATH_DEBUG) 		\
-				yell("O: %s = %s (%ld %ld) -> 0",  	\
+				yell("O: %s = %s (" INTMAX_FORMAT " " 	\
+						INTMAX_FORMAT ") -> 0",	\
 					s, #intop, i, j); 		\
 			error("Division by zero"); 			\
 			add_var_alias(s, empty_string, 0);		\
@@ -1205,7 +1244,9 @@ static void	op (expr_info *cx, int what)
 		} 							\
 									\
 		if (x_debug & DEBUG_NEW_MATH_DEBUG) 			\
-			yell("O: %s =  %s (%ld %ld) -> %ld",  		\
+			yell("O: %s =  %s (" INTMAX_FORMAT " "  	\
+						INTMAX_FORMAT ") -> " 	\
+						INTMAX_FORMAT,		\
 				s, #intop, i, j, intop); 		\
 									\
 		w = tokenize_float(cx, (intop));			\
@@ -1228,10 +1269,12 @@ static void	op (expr_info *cx, int what)
 			push_token(cx, 0);				\
 			break;						\
 		}							\
+		CHECK_NOEVAL						\
 									\
 		j = get_token_integer(cx, v);				\
 		if (x_debug & DEBUG_NEW_MATH_DEBUG) 			\
-			yell("O: %s (%s %ld) -> %ld", 			\
+			yell("O: %s (%s " INTMAX_FORMAT ") -> " 	\
+					INTMAX_FORMAT, 			\
 				#intop_result, s, j, (intop_result));	\
 									\
 		w = tokenize_integer(cx, (intop_assign));		\
@@ -1256,6 +1299,7 @@ static void	op (expr_info *cx, int what)
 #define COMPARE(x, y) 							\
 	{ 								\
 		pop_2_strings(cx, &s, &t);				\
+		CHECK_NOEVAL						\
 		if (is_real_number(s) && is_real_number(t))		\
 		{							\
 			a = atof(s), b = atof(t);			\
@@ -1280,40 +1324,47 @@ static void	op (expr_info *cx, int what)
 		/* Simple unary prefix operators */
 		case NOT:
 			c = pop_boolean(cx);
+			CHECK_NOEVAL
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("O: !%d -> %d", c, !c);
 			push_boolean(cx, !c);
 			break;
 		case COMP:
 			i = pop_integer(cx);
+			CHECK_NOEVAL
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
-				yell(": ~%ld -> %ld", i, ~i);
+				yell(": ~" INTMAX_FORMAT " -> "
+					   INTMAX_FORMAT, i, ~i);
 			push_integer(cx, ~i);
 			break;
 		case UPLUS:
 			a = pop_float(cx);
+			CHECK_NOEVAL
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("O: +%f -> %f", a, a);
 			push_float(cx, a);
 			break;
 		case UMINUS:
 			a = pop_float(cx);
+			CHECK_NOEVAL
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("O: -%f -> %f", a, -a);
 			push_float(cx, -a);
 			break;
 		case STRLEN:
 			s = pop_expanded(cx);
+			CHECK_NOEVAL
 			i = strlen(s);
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
-				yell("O: @(%s) -> %ld", s, i);
+				yell("O: @(%s) -> " INTMAX_FORMAT, s, i);
 			push_integer(cx, i);
 			break;
 		case WORDC:
 			s = pop_expanded(cx);
-			i = word_count(s);
+			CHECK_NOEVAL
+			i = count_words(s, DWORD_YES, "\"");
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
-				yell("O: #(%s) -> %ld", s, i);
+				yell("O: #(%s) -> " INTMAX_FORMAT, s, i);
 			push_integer(cx, i);
 			break;
 		case DEREF:
@@ -1321,8 +1372,15 @@ static void	op (expr_info *cx, int what)
 			if (top(cx) == MAGIC_TOKEN)
 				break;		/* Dont do anything */
 
-			*cx->args_flag = 1;
-			push_lval(cx, pop_expanded(cx));
+			/*
+			 * We need to consume the operand, even if
+			 * we don't intend to use it; plus we need
+			 * to ensure this defeats auto-append.  Ick.
+			 */
+			s = pop_expanded(cx);
+
+			CHECK_NOEVAL
+			push_lval(cx, s);
 			break;
 		}
 
@@ -1352,10 +1410,11 @@ static void	op (expr_info *cx, int what)
 			char *	myval;
 
 			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("O: (%s) ## (%s) -> %s%s", s, t, s, t);
 
-			myval = m_2dup(s, t);
+			myval = malloc_strdup2(s, t);
 			push_string(cx, myval);
 			new_free(&myval);
 			break;
@@ -1380,7 +1439,8 @@ static void	op (expr_info *cx, int what)
 			char *	myval;
 
 			GET_LVAL_RVAL
-			myval = m_strdup(get_token_expanded(cx, v));
+			CHECK_NOEVAL
+			myval = malloc_strdup(get_token_expanded(cx, v));
 			t = get_token_expanded(cx, w);
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG) 
@@ -1398,7 +1458,8 @@ static void	op (expr_info *cx, int what)
 			char *	myval;
 
 			GET_LVAL_RVAL
-			myval = m_strdup(get_token_expanded(cx, w));
+			CHECK_NOEVAL
+			myval = malloc_strdup(get_token_expanded(cx, w));
 			t = get_token_expanded(cx, v);
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG) 
@@ -1414,6 +1475,7 @@ static void	op (expr_info *cx, int what)
 		case EQ:
 		{
 			GET_LVAL_RVAL
+			CHECK_NOEVAL
 			t = get_token_expanded(cx, w);
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -1425,9 +1487,10 @@ static void	op (expr_info *cx, int what)
 		}
 		case SWAP:
 		{
-			char *sval, *tval;
+			const char *sval, *tval;
 
 			pop_2_tokens(cx, &v, &w);
+			CHECK_NOEVAL
 			if (!(s = get_token_lval(cx, v)))
 			{
 				push_token(cx, 0);
@@ -1454,6 +1517,7 @@ static void	op (expr_info *cx, int what)
 		case DEQ:
 		{
 			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
 			c = my_stricmp(s, t) ? 0 : 1;
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -1465,6 +1529,7 @@ static void	op (expr_info *cx, int what)
 		case NEQ:
 		{
 			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
 			c = my_stricmp(s, t) ? 1 : 0;
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -1473,9 +1538,35 @@ static void	op (expr_info *cx, int what)
 			push_boolean(cx, c);
 			break;
 		}
+		case LITEQ:
+		{
+			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
+			c = strcmp(s, t) ? 0 : 1;
+
+			if (x_debug & DEBUG_NEW_MATH_DEBUG)
+				yell("O: %s === %s -> %d", s, t, c);
+
+			push_boolean(cx, c);
+			break;
+		}
+		case NOTLITEQ:
+		{
+			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
+			c = strcmp(s, t) ? 1 : 0;
+
+			if (x_debug & DEBUG_NEW_MATH_DEBUG)
+				yell("O: %s !== %s -> %d", s, t, c);
+
+			push_boolean(cx, c);
+			break;
+		}
+
 		case MATCH:
 		{
 			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
 			c = wild_match(t, s) ? 1 : 0;
 			
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -1487,6 +1578,7 @@ static void	op (expr_info *cx, int what)
 		case NOMATCH:
 		{
 			pop_2_strings(cx, &s, &t);
+			CHECK_NOEVAL
 			c = wild_match(t, s) ? 0 : 1;
 			
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -1505,6 +1597,7 @@ static void	op (expr_info *cx, int what)
 		case QUEST:
 		{
 			pop_3_tokens(cx, &c, &v, &w);
+			CHECK_NOEVAL
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 			{
 				yell("O: %d ? %s : %s -> %s", c,
@@ -1522,12 +1615,25 @@ static void	op (expr_info *cx, int what)
 		case COMMA:
 		{
 			pop_2_tokens(cx, &v, &w);
+			CHECK_NOEVAL
+
+			/* 
+			 * XXX We need to EXPAND the left value, even
+			 * if we don't need it, because it might be a 
+			 * function call!
+			 */
+			get_token_expanded(cx, v);
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("O: %s , %s -> %s", 
 					get_token_expanded(cx, v),
 					get_token_expanded(cx, w),
 					get_token_expanded(cx, w));
+
+			/*
+			 * But we DONT want to expand the right value, 
+			 * until we use it LATER.
+			 */
 			push_token(cx, w);
 			break;
 		}
@@ -1542,7 +1648,7 @@ static void	op (expr_info *cx, int what)
 /**************************** EXPRESSION LEXER ******************************/
 static	int	dummy = 1;
 
-int	lexerr (expr_info *c, char *format, ...)
+static int	lexerr (expr_info *c, const char *format, ...)
 {
 	char 	buffer[BIG_BUFFER_SIZE + 1];
 	va_list	a;
@@ -1563,20 +1669,19 @@ int	lexerr (expr_info *c, char *format, ...)
  * case 'operand' is set to 1.  When an operand is lexed, then the next token
  * is expected to be a binary operator, so 'operand' is set to 0. 
  */
-__inline int	check_implied_arg (expr_info *c)
+static __inline int	check_implied_arg (expr_info *c)
 {
 	if (c->operand == 2)
 	{
 		push_token(c, MAGIC_TOKEN);	/* XXXX Bleh */
 		c->operand = 0;
-		*c->args_flag = 1;
 		return 0;
 	}
 
 	return c->operand;
 }
 
-__inline TOKEN 	operator (expr_info *c, char *x, int y, TOKEN z)
+static __inline TOKEN 	operator (expr_info *c, const char *x, int y, TOKEN z)
 {
 	check_implied_arg(c);
 	if (c->operand)
@@ -1587,7 +1692,7 @@ __inline TOKEN 	operator (expr_info *c, char *x, int y, TOKEN z)
 	return z;
 }
 
-__inline TOKEN 	unary (expr_info *c, char *x, int y, TOKEN z)
+static __inline TOKEN 	unary (expr_info *c, const char *x, int y, TOKEN z)
 {
 	if (!c->operand)
 		return lexerr(c, "A unary operator (%s) was found where "
@@ -1617,8 +1722,43 @@ static int	zzlex (expr_info *c)
 	    switch (*(c->ptr++)) 
 	    {
 		case '(':
-			c->operand = 1;
-			return M_INPAR;
+		{
+			char *p = c->ptr;
+			char oc = 0;
+			ssize_t	span;
+
+			/*
+			 * If we see an open paren, and we're expecting an
+			 * operator, then this is a function call.  If we 
+			 * are expecting an operand, then it is a paren-set.
+			 */
+			if (c->operand == 0)
+			{
+			    if ((span = MatchingBracket(p, '(', ')')) >= 0)
+			    {
+				c->ptr = p + span;
+				oc = *c->ptr;
+				*c->ptr = 0;
+			    }
+			    else
+				c->ptr = endstr(c->ptr);
+
+			    if (c->noeval)
+				c->last_token = 0;
+			    else
+				c->last_token = tokenize_raw(c, p);
+
+			    if (oc)
+				*c->ptr++ = oc;
+			    c->operand = 0;
+			    return M_FUNCTION;
+			}
+			else
+			{
+				c->operand = 1;
+				return M_INPAR;
+			}
+		}
 		case ')':
 			/*
 			 * If we get a close paren and the lexer is expecting
@@ -1707,8 +1847,14 @@ static int	zzlex (expr_info *c)
 
 		case '!':
 		{
-			if (*c->ptr == '=')
-				OPERATOR("!=", 1, NEQ)
+			if (*c->ptr == '=') 
+			{
+				c->ptr++;
+				if (*c->ptr == '=')
+					OPERATOR("!==", 1, NOTLITEQ)
+				else
+					OPERATOR("!=", 0, NEQ)
+			}
 			else if (*c->ptr == '~')
 				OPERATOR("!~", 1, NOMATCH)
 			else
@@ -1829,7 +1975,13 @@ static int	zzlex (expr_info *c)
 
 		case '=':
 			if (*c->ptr == '=') 
-				OPERATOR("==", 1, DEQ)
+			{
+				c->ptr++;
+				if (*c->ptr == '=')
+					OPERATOR("===", 1, LITEQ)
+				else
+					OPERATOR("==", 0, DEQ)
+			}
 			else if (*c->ptr == '~')
 				OPERATOR("=~", 1, MATCH)
 			else
@@ -1877,25 +2029,26 @@ static int	zzlex (expr_info *c)
 		{
 			char *p = c->ptr;
 			char oc = 0;
+			ssize_t	span;
 
 			if (!c->operand)
 				return lexerr(c, "Misplaced { token");
 
-			if ((c->ptr = MatchingBracket(p, '{', '}')))
+			if ((span = MatchingBracket(p, '{', '}')) >= 0)
 			{
+				c->ptr = p + span;
 				oc = *c->ptr;
 				*c->ptr = 0;
 			}
 			else
-				c->ptr = empty_string;
+				c->ptr = endstr(c->ptr);
 
 			c->last_token = 0;
 			if (!c->noeval)
 			{
-				char *result;
+				char *	result;
 
-				result = parse_line_with_return(NULL, p, 
-							(char *)c->args, 0, 0);
+				result = call_lambda_function(NULL, p, c->args);
 				c->last_token = tokenize_expanded(c, result);
 				new_free(&result);
 			}
@@ -1911,7 +2064,7 @@ static int	zzlex (expr_info *c)
 		 * This is an UNEXPANDED-STRING operand type.
 		 * Extract everything inside the [...]'s, and then
 		 * tokenize that as a "raw" token.  It will be expanded
-		 * on an as-needed basis.  
+		 * on an as-needed basis.
 		 *
 		 * If we are in the no-eval section of a short-circuit, 
 		 * then we throw away this token entirely.
@@ -1920,17 +2073,19 @@ static int	zzlex (expr_info *c)
 		{
 			char *p = c->ptr;
 			char oc = 0;
+			ssize_t	span;
 
 			if (!c->operand)
 				return lexerr(c, "Misplaced [ token");
 
-			if ((c->ptr = MatchingBracket(p, '[', ']')))
+			if ((span = MatchingBracket(p, '[', ']')) >= 0)
 			{
+				c->ptr = p + span;
 				oc = *c->ptr;
 				*c->ptr = 0;
 			}
 			else
-				c->ptr = empty_string;
+				c->ptr = endstr(c->ptr);
 
 			if (c->noeval)
 				c->last_token = 0;
@@ -1942,9 +2097,82 @@ static int	zzlex (expr_info *c)
 			c->operand = 0;
 			return ID;
 		}
-		case ' ':
-		case '\t':
-		case '\n':
+
+		/*
+		 * This is a DOUBLE QUOTED (unexpanded-string) operand type.
+		 * Extract the inside of "..." and tokenize that as a "raw"
+		 * value.  It will be expanded later as needed.
+		 */
+		case '"':
+		{
+			char *p = c->ptr;
+			char oc = 0;
+			ssize_t	span;
+
+			if (!c->operand)
+				return lexerr(c, "Misplaced \" token");
+
+			if ((span = MatchingBracket(p, '\0', '"')) >= 0)
+			{
+				c->ptr = p + span;
+				oc = *c->ptr;
+				*c->ptr = 0;
+			}
+			else
+				c->ptr = endstr(c->ptr);
+
+			if (c->noeval)
+				c->last_token = 0;
+			else
+				c->last_token = tokenize_raw(c, p);
+
+			if (oc)
+				*c->ptr++ = oc;
+			c->operand = 0;
+			return ID;
+		}
+
+		/*
+		 * This is a SINGLE QUOTED (expanded-string) operand type.
+		 * Extract the inside of '...' and tokenize that as an
+		 * "expanded" value.  It will be expanded later as needed.
+		 */
+		case '\'':
+		{
+			char *p = c->ptr;
+			char oc = 0;
+			ssize_t	span;
+
+			if (!c->operand)
+				return lexerr(c, "Misplaced ' token");
+
+			if ((span = MatchingBracket(p, '\0', '\'')) >= 0)
+			{
+				c->ptr = p + span;
+				oc = *c->ptr;
+				*c->ptr = 0;
+			}
+			else
+				c->ptr = endstr(c->ptr);
+
+			if (c->noeval)
+				c->last_token = 0;
+			else
+			{
+			    char *ick = NULL;
+			    malloc_strcat_ues(&ick, p, "'");
+			    c->last_token = tokenize_expanded(c, ick);
+			    new_free(&ick);
+			}
+
+			if (oc)
+				*c->ptr++ = oc;
+			c->operand = 0;
+			return ID;
+		}
+
+		/* The "Space" characters. */
+		case 9: case 10: case 11: case 12: case 13: case ' ':
 			start++;
 			break;
 
@@ -2039,7 +2267,7 @@ handle_expando:
 			else
 			{
 				c->last_token = 0; /* Empty token */
-				c->ptr = empty_string;
+				c->ptr = endstr(c->ptr);
 			}
 
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
@@ -2100,6 +2328,7 @@ static void	mathparse (expr_info *c, int pc)
 		 * Getting the value is done on an as-needed basis.
 		 */
 		case ID:
+		{
 			if (x_debug & DEBUG_NEW_MATH_DEBUG)
 				yell("Parsed identifier token [%s]", 
 					get_token_expanded(c, c->last_token));
@@ -2112,6 +2341,46 @@ static void	mathparse (expr_info *c, int pc)
 			 */
 			push_token(c, c->last_token);
 			break;
+		}
+
+		/*
+		 * For a function call, the argument list is in c->last_token
+		 * and c->mtok is FUNCTION.  The function's name is on the top
+		 * of the queue.  XXX This is a hack because this is treated 
+		 * as a special case, and not as an operator.  But I don't
+		 * really care.  Maybe I'll "fix" this later. XXX
+		 */
+		case M_FUNCTION:
+		{
+			const char *funcname, *args;
+			char *hack = NULL;
+			size_t clue = 0;
+			char *retval;
+			int	func_token;
+
+			func_token = pop_token(c);
+			funcname = get_token_fname(c, func_token);
+			args = get_token_raw(c, c->last_token);
+
+			if (c->noeval)
+				push_token(c, 0);
+			else
+			{
+			    malloc_strcpy_c(&hack, funcname, &clue);
+			    malloc_strcat_c(&hack, "(", &clue);
+			    malloc_strcat_c(&hack, args, &clue);
+			    malloc_strcat_c(&hack, ")", &clue);
+
+			    if (x_debug & DEBUG_NEW_MATH_DEBUG)
+				yell("Parsed function call %s", hack);
+
+			    retval = call_function(hack, c->args);
+			    c->last_token = tokenize_expanded(c, retval);
+
+			    push_token(c, c->last_token);
+			}
+			break;
+		}
 
 		/*
 		 * An open-parenthesis indicates that we should
@@ -2158,7 +2427,7 @@ static void	mathparse (expr_info *c, int pc)
 			mathparse(c, prec[QUEST]);
 			if (u)
 				c->noeval--;
-			op(c, QUEST);
+			reduce(c, QUEST);
 
 			continue;
 		}
@@ -2216,7 +2485,7 @@ static void	mathparse (expr_info *c, int pc)
 			 * Then reduce this operation.
 			 */
 			c->noeval = onoeval;
-			op(c, otok);
+			reduce(c, otok);
 			continue;
 		}
 	    }
@@ -2237,20 +2506,19 @@ static void	mathparse (expr_info *c, int pc)
  * of state information stored in global variables.  Therefore, this math
  * parser is re-entrant safe.
  */
-char *	matheval (char *s, const char *args, int *args_flag)
+static char *	matheval (char *s, const char *args)
 {
 	expr_info	context;
 	char *		ret = NULL;
 
 	/* Sanity check */
 	if (!s || !*s)
-		return m_strdup(empty_string);
+		return malloc_strdup(empty_string);
 
 	/* Create new state */
 	setup_expr_info(&context);
 	context.ptr = s;
-	context.args = (char *)args;
-	context.args_flag = args_flag;
+	context.args = args;
 
 	/* Actually do the parsing */
 	mathparse(&context, TOPPREC);
@@ -2258,7 +2526,7 @@ char *	matheval (char *s, const char *args, int *args_flag)
 	/* Check for error */
 	if (context.errflag)
 	{
-		ret = m_strdup(empty_string);
+		ret = malloc_strdup(empty_string);
 		goto cleanup;
 	}
 
@@ -2276,7 +2544,7 @@ char *	matheval (char *s, const char *args, int *args_flag)
 	}
 
 	/* Get the return value, if requested */
-	ret = m_strdup(get_token_expanded(&context, pop_token(&context)));
+	ret = malloc_strdup(get_token_expanded(&context, pop_token(&context)));
 
 cleanup:
 	/* Clean up and restore order */
