@@ -1,4 +1,4 @@
-/* $EPIC: flood.c,v 1.15 2003/09/23 21:49:47 jnelson Exp $ */
+/* $EPIC: flood.c,v 1.26 2005/10/30 22:41:19 jnelson Exp $ */
 /*
  * flood.c: handle channel flooding.
  *
@@ -50,22 +50,7 @@
 #include "functions.h"
 #include "lastlog.h"
 #include "window.h"
-
-static	const char *ignore_types[NUMBER_OF_FLOODS] =
-{
-	"CRAP",
-	"CTCPS",
-	"INVITES",
-	"JOINS",
-	"MSGS",
-	"NICKS",
-	"NOTES",
-	"NOTICES",
-	"PUBLIC",
-	"TOPICS",
-	"WALLOPS",
-	"WALLS"
-};
+#include "reg.h"
 
 typedef struct flood_stru
 {
@@ -73,7 +58,7 @@ typedef struct flood_stru
 	char		*channel;
 	int		server;
 
-	FloodType	type;
+	int		level;
 	long		cnt;
 	Timeval		start;
 	int		floods;
@@ -97,8 +82,14 @@ static const char *	normalize_nuh (const char *nuh)
 	} else if (!nuh) {
 	} else if (maskuser == 1 && isalnum(*nuh)) {
 	} else {
+		char prefix = *nuh;
 		char *nnuh = strrchr(nuh, '@');
-		nuh = nnuh ? nnuh : nuh;
+		if (nnuh - nuh >= 2) {
+			*--nnuh = *star;
+			if (maskuser == 1)
+				*--nnuh = prefix;
+			nuh = nnuh;
+		}
 	}
 
 	return nuh;
@@ -111,7 +102,7 @@ static const char *	normalize_nuh (const char *nuh)
  * person.  It will return 1 if flooding is being check for someone and an ON
  * FLOOD is activated.
  */
-int	new_check_flooding (const char *nick, const char *nuh, const char *chan, const char *line, FloodType type)
+int	new_check_flooding (const char *nick, const char *nuh, const char *chan, const char *line, int level)
 {
 static	int	 pos = 0;
 	int	 i,
@@ -121,7 +112,11 @@ static	int	 pos = 0;
 	Timeval	 right_now;
 	double	 diff;
 	Flooding *tmp;
+	int	l;
+	char *	freeit;
 
+	freeit = malloc_strdup(nuh);
+	nuh = freeit;
 
 	/*
 	 * Figure out how many people we want to track
@@ -145,7 +140,7 @@ static	int	 pos = 0;
 			flood[i].nuh = NULL;
 			flood[i].channel = NULL;
 			flood[i].server = NOSERV;
-			flood[i].type = -1;
+			flood[i].level = LEVEL_NONE;
 			flood[i].cnt = 0;
 			get_time(&(flood[i].start));
 			flood[i].floods = 0;
@@ -163,13 +158,16 @@ static	int	 pos = 0;
 		if (flood)
 			new_free((char **)&flood);
 		users = 0;
+		new_free(&freeit);
 		return 0;
 	}
 
 	if (nuh && *nuh)
 		nuh = normalize_nuh(nuh);
-	else
+	else {
+		new_free(&freeit);
 		return 0;
+	}
 
 	/*
 	 * What server are we using?
@@ -191,7 +189,7 @@ static	int	 pos = 0;
 		/*
 		 * Do some inexpensive tests first
 		 */
-		if (type != flood[i].type)
+		if (level != flood[i].level)
 			continue;
 		if (server != flood[i].server)
 			continue;
@@ -245,11 +243,12 @@ static	int	 pos = 0;
 			new_free(&tmp->channel);
 
 		tmp->server = server;
-		tmp->type = type;
+		tmp->level = level;
 		tmp->cnt = 0;
 		tmp->start = right_now;
 
 		pos = (0 < old_pos ? old_pos : users) - 1;
+		new_free(&freeit);
 		return 0;
 	}
 	else
@@ -265,16 +264,16 @@ static	int	 pos = 0;
 		diff = time_diff(tmp->start, right_now);
 
 		if ((diff == 0.0 || tmp->cnt / diff >= rate) &&
-				(retval = do_hook(FLOOD_LIST, "%s %s %s %s",
-				nick, ignore_types[type],
-				chan ? chan : "*", line)))
+				(retval = do_hook(FLOOD_LIST, "%s %s %s %ld %s",
+				nick, level_to_str(tmp->level),
+				chan ? chan : "*", tmp->cnt, line)))
 		{
 			tmp->floods++;
-			message_from(chan, LOG_CRAP);
+			l = message_from(chan, LEVEL_OTHER);
 			if (get_int_var(FLOOD_WARNING_VAR))
 				say("FLOOD: %ld %s detected from %s in %f seconds",
-					tmp->cnt+1, ignore_types[type], nick, diff);
-			message_from(NULL, LOG_CRAP);
+					tmp->cnt+1, level_to_str(tmp->level), nick, diff);
+			pop_message_from(l);
 		}
 		else
 		{
@@ -286,15 +285,17 @@ static	int	 pos = 0;
 		}
 	}
 
+	new_free(&freeit);
+
 	if (get_int_var(FLOOD_IGNORE_VAR))
 		return retval;
 	else
 		return 0;
 }
 
-int	check_flooding (const char *nick, const char *nuh, FloodType type, const char *line)
+int	check_flooding (const char *nick, const char *nuh, int mask, const char *line)
 {
-	return new_check_flooding(nick, nuh, NULL, line, type);
+	return new_check_flooding(nick, nuh, NULL, line, mask);
 }
 
 /*
@@ -303,23 +304,66 @@ int	check_flooding (const char *nick, const char *nuh, FloodType type, const cha
  */
 char *	function_floodinfo (char *args)
 {
-	const char	*arg;
+	char	*arg;
 	char *ret = NULL;
 	size_t	clue = 0;
 	Timeval right_now;
 	int	i;
+	double	idiff;
 
 	get_time(&right_now);
 
 	while ((arg = new_next_arg(args, &args))) {
-		if (!strspn(arg, "*?"))
-			arg = normalize_nuh(arg);
+	const	char	*nuh = star;
+	const	char	*chan = star;
+	const	char	*level = star;
+		int	server = -1;
+		int	count = 0;
+		double	diff = 0;
+		double	rate = 0;
+		int	cless = 0, dless = 0, rless = 0;
+		size_t	len = strlen(arg);
+
+		dequote_buffer(arg, &len);
+		if (arg && *arg)
+			GET_STR_ARG(nuh, arg);
+		if (arg && *arg)
+			GET_STR_ARG(chan, arg);
+		if (arg && *arg)
+			GET_STR_ARG(level, arg);
+		if (arg && *arg)
+			GET_INT_ARG(server, arg);
+		if (arg && *arg)
+			GET_INT_ARG(count, arg);
+		if (arg && *arg)
+			GET_FLOAT_ARG(diff, arg);
+		if (arg && *arg)
+			GET_FLOAT_ARG(rate, arg);
+		if (count < 0)
+			count= -count, cless++;
+		if (diff < 0)
+			diff = -diff, dless++;
+		if (rate < 0)
+			rate = -rate, rless++;
+
 		for (i = 0; i < users; i++) {
-			if (flood[i].nuh && wild_match(arg, flood[i].nuh)) {
+			if (server >= 0 && flood[i].server != server) {
+			} else if (!flood[i].nuh) {
+			} else if (!wild_match(nuh, flood[i].nuh) && !wild_match(flood[i].nuh, nuh)) {
+			} else if (!wild_match(chan, flood[i].channel ? flood[i].channel : star)) {
+			} else if (!wild_match(level, level_to_str(flood[i].level))) {
+			} else if (!cless && flood[i].cnt < count) {
+			} else if ( cless && flood[i].cnt > count) {
+			} else if (!(idiff = time_diff(flood[i].start, right_now))) {
+			} else if (!dless && idiff < diff) {
+			} else if ( dless && idiff > diff) {
+			} else if (!rless && flood[i].cnt / idiff < rate) {
+			} else if ( rless && flood[i].cnt / idiff > rate) {
+			} else {
 				malloc_strcat_wordlist_c(&ret, space, "\"", &clue);
 				malloc_strcat_wordlist_c(&ret, empty_string, flood[i].nuh, &clue);
 				malloc_strcat_wordlist_c(&ret, space, flood[i].channel ? flood[i].channel : star, &clue);
-				malloc_strcat_wordlist_c(&ret, space, ignore_types[flood[i].type], &clue);
+				malloc_strcat_wordlist_c(&ret, space, level_to_str(flood[i].level), &clue);
 				malloc_strcat_wordlist_c(&ret, space, ltoa(flood[i].server), &clue);
 				malloc_strcat_wordlist_c(&ret, space, ltoa(flood[i].cnt), &clue);
 				malloc_strcat_wordlist_c(&ret, space, ftoa(time_diff(flood[i].start, right_now)), &clue);
