@@ -1,5 +1,6 @@
+/* $EPIC: glob.c,v 1.11 2005/03/15 05:36:20 jnelson Exp $ */
 #include "config.h"
-#if defined(NEED_GLOB)
+#ifdef NEED_GLOB
 
 /*
  * Copyright (c) 1989, 1993
@@ -34,7 +35,6 @@
  */
 
 /* from: static char sccsid[] = "@(#)glob.c	8.3 (Berkeley) 10/13/93"; */
-/* $Id: glob.c,v 1.2 2001/09/26 16:37:10 jnelson Exp $ */
 
 /*
  * glob(3) -- a superset of the one defined in POSIX 1003.2.
@@ -108,7 +108,10 @@
 #define	M_MASK		0xffff
 #define	M_ASCII		0x00ff
 
-typedef u_short Char;
+#ifdef Char
+#undef Char
+#endif
+typedef unsigned short Char;
 
 #define	CHAR(c)		((Char)((c)&M_ASCII))
 #define	META(c)		((Char)((c)|M_QUOTE))
@@ -123,10 +126,10 @@ typedef u_short Char;
 
 static int	 	compare (const void *, const void *);
 static void	 	g_Ctoc (const Char *, char *);
-static int	 	g_lstat (Char *, struct stat *, glob_t *);
+static int	 	g_lstat (Char *, Stat *, glob_t *);
 static DIR	*	g_opendir (Char *, glob_t *);
-static Char	*	g_strchr (Char *, int);
-static int	 	g_stat (Char *, struct stat *, glob_t *);
+static ssize_t		g_strchr (const Char *, int);
+static int	 	g_stat (Char *, Stat *, glob_t *);
 static int	 	glob0 (const Char *, glob_t *);
 static int	 	glob1 (Char *, glob_t *);
 static int	 	glob2 (Char *, Char *, Char *, glob_t *);
@@ -146,7 +149,7 @@ int bsd_glob		(	const char *pattern,
 	int c;
 	Char *bufnext, *bufend, patbuf[MAXPATHLEN+1];
 
-	patnext = (u_char *) pattern;
+	patnext = (const u_char *) pattern;
 	if (!(flags & GLOB_APPEND)) 
 	{
 		pglob->gl_pathc = 0;
@@ -200,14 +203,18 @@ static int globexp1	(	const Char *pattern,
 {
 	const Char* ptr = pattern;
 	int rv;
+	ssize_t	span;
 
 	/* Protect a single {}, for find(1), like csh */
 	if (pattern[0] == LBRACE && pattern[1] == RBRACE && pattern[2] == EOS)
 		return glob0(pattern, pglob);
 
-	while ((ptr = (const Char *) g_strchr((Char *) ptr, LBRACE)) != NULL)
+	while ((span = g_strchr(ptr, LBRACE)) >= 0)
+	{
+		ptr += span;
 		if (!globexp2(ptr, pattern, pglob, &rv))
 			return rv;
+	}
 
 	return glob0(pattern, pglob);
 }
@@ -417,7 +424,7 @@ static int glob0		(	const Char *pattern,
 			if (c == NOT)
 				++qpatnext;
 			if (*qpatnext == EOS || 
-			    g_strchr((Char *) qpatnext+1, RBRACKET) == NULL) 
+			    g_strchr(qpatnext+1, RBRACKET) < 0)
 			{
 				*bufnext++ = LBRACKET;
 				if (c == NOT)
@@ -490,7 +497,7 @@ static int glob0		(	const Char *pattern,
 static int compare		(	const void *p,
 					const void *q		)
 {
-	return (strcmp(*(char **)p, *(char **)q));
+	return (strcmp(*(const char * const *)p, *(const char * const *)q));
 }
 
 static int glob1		(	Char *pattern,
@@ -514,7 +521,7 @@ static int glob2		(	Char *pathbuf,
 					Char *pattern,
 					glob_t *pglob		)
 {
-	struct stat sb;
+	Stat sb;
 	Char *p, *q;
 	int anymeta;
 
@@ -585,12 +592,12 @@ static int glob3		(	Char *pathbuf,
 	 * and dirent.h as taking pointers to differently typed opaque
 	 * structures.
 	 */
-	struct dirent *(*readdirfunc)();
+	struct dirent *(*readdirfunc)(DIR *);
 
 	*pathend = EOS;
 	errno = 0;
-	    
-	if ((dirp = g_opendir(pathbuf, pglob)) == NULL) 
+
+	if ((dirp = g_opendir(pathbuf, pglob)) == NULL)
 	{
 		/* TODO: don't call for ENOENT or ENOTDIR? */
 		if (pglob->gl_errfunc) 
@@ -607,7 +614,7 @@ static int glob3		(	Char *pathbuf,
 
 	/* Search directory for matching names. */
 	if (pglob->gl_flags & GLOB_ALTDIRFUNC)
-		readdirfunc = pglob->gl_readdir;
+		readdirfunc = (struct dirent *(*)(DIR *))pglob->gl_readdir;
 	else
 		readdirfunc = readdir;
 
@@ -791,7 +798,7 @@ static DIR *g_opendir		(	register Char *str,
 	char buf[MAXPATHLEN];
 
 	if (!*str)
-		strcpy(buf, ".");
+		strlcpy(buf, ".", sizeof buf);
 	else
 		g_Ctoc(str, buf);
 
@@ -802,7 +809,7 @@ static DIR *g_opendir		(	register Char *str,
 }
 
 static int g_lstat		(	register Char *fn,
-					struct stat *sb,
+					Stat *sb,
 					glob_t *pglob			)
 {
 	char buf[MAXPATHLEN];
@@ -814,7 +821,7 @@ static int g_lstat		(	register Char *fn,
 }
 
 static int g_stat		(	register Char *fn,
-					struct stat *sb,
+					Stat *sb,
 					glob_t *pglob			)
 {
 	char buf[MAXPATHLEN];
@@ -825,14 +832,15 @@ static int g_stat		(	register Char *fn,
 	return(stat(buf, sb));
 }
 
-static Char *g_strchr		(	Char *str, 
+static ssize_t	g_strchr	(	const Char *str, 
 					int ch				)
 {
+	const Char *start = str;
 	do {
 		if (*str == ch)
-			return (str);
+			return str - start;
 	} while (*str++);
-	return (NULL);
+	return -1;
 }
 
 static void g_Ctoc		(	register const Char *str,
