@@ -1,4 +1,4 @@
-/* $EPIC: irc.c,v 1.778 2014/03/31 13:57:22 jnelson Exp $ */
+/* $EPIC: irc.c,v 1.1053 2006/07/15 15:26:17 jnelson Exp $ */
 /*
  * ircII: a new irc client.  I like it.  I hope you will too!
  *
@@ -6,7 +6,7 @@
  * Copyright (c) 1991, 1992 Troy Rollo.
  * Copyright (c) 1992-1996 Matthew Green.
  * Copyright  1994 Jake Khuon.
- * Copyright  1993, 2003 EPIC Software Labs.
+ * Copyright  1993, 2006 EPIC Software Labs.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,19 +40,19 @@
 /*
  * irc_version is what $J returns, its the common-name for the version.
  */
-const char irc_version[] = "EPIC4-2.10.5";
-const char useful_info[] = "epic4 2 10 5";
+const char irc_version[] = "EPIC5-0.3.1";
+const char useful_info[] = "epic5 0 3 1";
 
 /*
  * internal_version is what $V returns, its the integer-id for the
  * version, and corresponds to the date of release, YYYYMMDD.
  */ 
-const char internal_version[] = "20080904";
+const char internal_version[] = "20060715";
 
 /*
  * In theory, this number is incremented for every commit.
  */
-const unsigned long	commit_id = 779;
+const unsigned long	commit_id = 1381;
 
 /*
  * As a way to poke fun at the current rage of naming releases after
@@ -60,7 +60,7 @@ const unsigned long	commit_id = 779;
  * reality, I have decided to start doing that with EPIC.  These names
  * are intentionally and maliciously silly.  Complaints will be ignored.
  */
-const char ridiculous_version_name[] = "Induration";
+const char ridiculous_version_name[] = "Sycophant";
 
 #define __need_putchar_x__
 #include "status.h"
@@ -71,7 +71,7 @@ const char ridiculous_version_name[] = "Induration";
 #include "input.h"
 #include "alias.h"
 #include "output.h"
-#include "termx.h"
+#include "term.h"
 #include "exec.h"
 #include "screen.h"
 #include "log.h"
@@ -81,14 +81,14 @@ const char ridiculous_version_name[] = "Induration";
 #include "ircaux.h"
 #include "commands.h"
 #include "window.h"
-#include "history.h"
 #include "exec.h"
 #include "notify.h"
 #include "mail.h"
 #include "timer.h"
 #include "newio.h"
 #include "parse.h"
-#include "notice.h"
+#include "levels.h"
+#include "extlang.h"
 #include <pwd.h>
 
 
@@ -110,7 +110,7 @@ int		use_iexten = -1;
  * of the numeric.  Its negative so if we pass it to do_hook, it can
  * tell its a numeric and not a named ON event.
  */
-int		current_numeric;
+int		current_numeric = -1;
 
 /* Set if the client is not using a termios interface */
 int		dumb_mode = 0;
@@ -121,12 +121,6 @@ int		background = 0;
 /* Set if the client is checking fd 0 for input. (usu. op of "background") */
 int		use_input = 1;
 
-/* The number of WAIT tokens sent out so far */
-int		waiting_out = 0;
-
-/* The number of WAIT tokens returned so far */
-int		waiting_in = 0;
-
 /*
  * Set when an OPER command is sent out, reset when umode +o or 464 reply
  * comes back.  This is *seriously* bogus.
@@ -145,12 +139,6 @@ Timeval		idle_time = { 0, 0 };
 /* Set to the time the client booted up */
 Timeval		start_time;
 
-/* The number of child processes still unreaped. */
-int		child_dead = 0;
-
-/* Set if the current output is from a trusted source */
-int		trusted_output = 1;
-
 /* Set to 0 when you want to suppress all beeps (such as window repaints) */
 int		global_beep_ok = 1;
 
@@ -166,27 +154,18 @@ volatile int	cntl_c_hit = 0;
 /* This is 1 if we are in the foreground process group */
 int		foreground = 1;
 
-/* This is 0 until your ~/.ircrc is loaded */
-int		ircrc_loaded = 0;
-
-/* This is 0 unless you specify the -B command line flag */
-int		load_ircrc_right_away = 0;
-
 /* This is 1 if you want all logging to be inhibited. Dont leave this on! */
 int		inhibit_logging = 0;
 
-/* This is 1 if we're currently loading the "global" script, 0 afterwards */
-int		loading_global = 0;
-
 /* This is reset every time io() is called.  Use this to save calls to time */
 Timeval		now = {0, 0};
 
-/* When you "highlight ignore", this is what is used to embellish */
-char *		highlight_char;
-
 /* Output which is displayed without modification by the user */
 int		privileged_output = 0;
 
+/* Output which should not trigger %F in hidden windows */
+int		do_window_notifies = 1;
+
 /*
  * If set, outbound connections will be bind()ed to the address
  * specified.  if unset, the default address for your host will
@@ -198,29 +177,26 @@ int		privileged_output = 0;
  * set to addresses that do not belong to the current hostname.
  * If that happens, outbound connections will fail, and its not my fault.
  */
+char *		LocalIPv4HostName = NULL;
+char *		LocalIPv6HostName = NULL;
 char *		LocalHostName = NULL;
-ISA *		LocalIPv4Addr = NULL;
-#ifdef INET6
-ISA6 *		LocalIPv6Addr = NULL;
-#endif
 
 int		inbound_line_mangler = 0,
 		outbound_line_mangler = 0;
 
-char		*startup_file = NULL,		/* full path .epicrc file */
-		*epicrc_file = NULL,		/* full path .epicrc file */
-		*ircrc_file = NULL,		/* full path .ircrc file */
+static char	*epicrc_file = NULL,		/* full path .epicrc file */
+		*ircrc_file = NULL;		/* full path .ircrc file */
+char		*startup_file = NULL,		/* Set when epicrc loaded */
 		*my_path = (char *) 0,		/* path to users home dir */
 		*irc_lib = (char *) 0,		/* path to the ircII library */
 		*default_channel = NULL,	/* Channel to join on connect */
 		nickname[NICKNAME_LEN + 1],	/* users nickname */
 		hostname[NAME_LEN + 1],		/* name of current host */
-		realname[REALNAME_LEN + 1],	/* real name of user */
 		username[NAME_LEN + 1],		/* usernameof user */
 		userhost[NAME_LEN + 1],		/* userhost of user */
 		*send_umode = NULL,		/* sent umode */
-		*last_notify_nick = (char *) 0,	/* last detected nickname */
-		empty_string[] = "",		/* just an empty string */
+		*last_notify_nick = (char *) 0;	/* last detected nickname */
+const char	empty_string[] = "",		/* just an empty string */
 		space[] = " ",			/* just a lonely space */
 		on[] = "ON",
 		off[] = "OFF",
@@ -228,13 +204,8 @@ char		*startup_file = NULL,		/* full path .epicrc file */
 		one[] = "1",
 		star[] = "*",
 		dot[] = ".",
-		comma[] = ",",
-		*cut_buffer = (char *) 0;	/* global cut_buffer */
-
-fd_set		readables, held_readables;
-fd_set		writables, held_writables;
-int		global_max_fd = -1;
-
+		comma[] = ",";
+unsigned char	*cut_buffer = (char *) 0;	/* global cut_buffer */
 
 static		char	switch_help[] =
 "Usage: epic [switches] [nickname] [server list]                      \n\
@@ -254,6 +225,7 @@ static		char	switch_help[] =
       -h\tPrint this help message                                     \n\
       -q\tThe program will not load your .ircrc file                  \n\
       -s\tThe program will not connect to a server upon startup       \n\
+      -S\tEach argument will be tokenised by whitechar                \n\
       -v\tPrint the version of this irc client and exit               \n\
       -x\tRun the client in full X_DEBUG mode                         \n\
       -c <chan>\tJoin <chan> after first connection to a server       \n\
@@ -275,15 +247,9 @@ static SIGNAL_HANDLER(sig_irc_exit)
 void	irc_exit (int really_quit, const char *format, ...)
 {
 	char 	buffer[BIG_BUFFER_SIZE];
-	char *	sub_format;
+	char *	quit_message = NULL;
 	int	old_window_display = window_display;
 	int	value;
-#ifdef PERL
-	extern void perlstartstop(int);
-#endif
-#ifdef TCL
-	extern void tclstartstop(int);
-#endif
 
 	/*
 	 * If we get called recursively, something is hosed.
@@ -308,29 +274,29 @@ void	irc_exit (int really_quit, const char *format, ...)
 	{
 		va_list arglist;
 		va_start(arglist, format);
-		vsnprintf(buffer, BIG_BUFFER_SIZE - 1, format, arglist);
+		vsnprintf(buffer, sizeof(buffer), format, arglist);
 		va_end(arglist);
+		quit_message = buffer;
 	}
 	else
 	{
-		if (!(format = get_string_var(QUIT_MESSAGE_VAR)))
-			format = "%s";
-
-		sub_format = convert_sub_format(format, 's');
-		snprintf(buffer, BIG_BUFFER_SIZE - 1, sub_format, irc_version);
-		new_free(&sub_format);
+		strlcpy(buffer, "Default", sizeof(buffer));
+		quit_message = NULL;
 	}
 
-
 	/* Do some clean up */
 	do_hook(EXIT_LIST, "%s", buffer);
-#ifdef TCL
+#ifdef HAVE_TCL
 	tclstartstop(0);
 #endif
-#ifdef PERL
+#ifdef HAVE_PERL
 	perlstartstop(0);  /* In case there's perl code in the exit hook. */
 #endif
-	close_all_servers(buffer);
+#ifdef RUBY
+	rubystartstop(0);
+#endif
+
+	close_all_servers(quit_message);
 	value = 0;
 	logger(&value);
 	get_child_exit(-1);  /* In case some children died in the exit hook. */
@@ -347,9 +313,10 @@ void	irc_exit (int really_quit, const char *format, ...)
 	
 	/* Try to free as much memory as possible */
 	window_display = 0;
-	dumpcmd(NULL, NULL, NULL);
+/*	dumpcmd(NULL, NULL, NULL); */
+	flush_all_symbols();
+	flush_on_hooks();
 	set_lastlog_size(&value);
-	set_history_size(&value);
 	remove_channel(NULL, 0);
 	destroy_call_stack();
 	remove_bindings();
@@ -441,7 +408,7 @@ static void quit_response (char *dummy, char *ptr)
 }
 
 /* irc_quit: prompts the user if they wish to exit, then does the right thing */
-void irc_quit (char unused, char *not_used)
+BUILT_IN_KEYBINDING(irc_quit)
 {
 	static	int in_it = 0;
 
@@ -480,9 +447,15 @@ static SIGNAL_HANDLER(sig_user1)
 	clean_up_processes();
 }
 
+static SIGNAL_HANDLER(sig_user2)
+{
+	system_exception++;
+}
+
 static	void	show_version (void)
 {
-	printf("ircII %s (Commit id: %lu) (Date of release: %s)\n\r", irc_version, commit_id, internal_version);
+	printf("ircII %s (Commit id: %lu) (Date of release: %s)\n\r", 
+			irc_version, commit_id, internal_version);
 	exit (0);
 }
 
@@ -506,18 +479,20 @@ static	void	show_version (void)
 static	void	parse_args (int argc, char **argv)
 {
 	int ch;
-	int add_servers = 0;
+	int append_servers = 0;
 	struct passwd *entry;
 	char *ptr = (char *) 0;
 	char *tmp_hostname = NULL;
 	char *the_path = NULL;
 	char *translation_path = NULL;
+	
+	int altargc = 0;
+	char **altargv;
 
 	extern char *optarg;
 	extern int optind;
 
 	*nickname = 0;
-	*realname = 0;
 	*username = 0;
 
 	/* 
@@ -531,7 +506,7 @@ static	void	parse_args (int argc, char **argv)
 		{
 			if ((ptr = strchr(entry->pw_gecos, ',')))
 				*ptr = 0;
-			strlcpy(realname, entry->pw_gecos, sizeof realname);
+			set_var_value(REALNAME_VAR, entry->pw_gecos, 0);
 		}
 
 		if (entry->pw_name && *(entry->pw_name))
@@ -587,11 +562,15 @@ static	void	parse_args (int argc, char **argv)
 		}
 
 	if ((ptr = getenv("IRCNAME")))
-		strlcpy(realname, ptr, sizeof realname);
+		set_var_value(REALNAME_VAR, ptr, 0);
 	else if ((ptr = getenv("NAME")))
-		strlcpy(realname, ptr, sizeof realname);
-	else if (!*realname)
-		strlcpy(realname, "*Unknown*", sizeof realname);
+		set_var_value(REALNAME_VAR, ptr, 0);
+	else
+	{
+		ptr = get_string_var(REALNAME_VAR);
+		if (!ptr || !*ptr)
+			set_var_value(REALNAME_VAR, "*Unknown*", 0);
+	}
 
 	if ((ptr = getenv("HOME")))
 		malloc_strcpy(&my_path, ptr);
@@ -626,7 +605,7 @@ static	void	parse_args (int argc, char **argv)
 	else
 		the_path = malloc_sprintf(NULL, DEFAULT_IRCPATH, irc_lib);
 
-	set_string_var(LOAD_PATH_VAR, the_path);
+	set_var_value(LOAD_PATH_VAR, the_path, 0);
 	new_free(&the_path);
 
 	if ((ptr = getenv("IRCHOST")) && *ptr)
@@ -637,13 +616,62 @@ static	void	parse_args (int argc, char **argv)
 	else
 		translation_path = malloc_strdup2(IRCLIB, "/translation/");
 
-	set_string_var(TRANSLATION_PATH_VAR, translation_path);
+	set_var_value(TRANSLATION_PATH_VAR, translation_path, 0);
 	new_free(&translation_path);
+	
+
+	/* The -S-option  / shebang tokeniser */
+	if (argc > 1 && ((argv[1][0] == '-' && argv[1][1] == 'S')
+		|| strchr(argv[1], ' ') != NULL 
+		|| strchr(argv[1], '\t') != NULL)
+	)
+	{
+		int argn, argpos, bufpos, c = 0, n;
+		altargv = new_malloc(sizeof(char *));
+		altargv[0] = new_malloc(strlen(argv[0]) +1);
+		strcpy(altargv[0], argv[0]);
+		altargc = 1;
+		
+		for (argn = 1; argn < argc; argn++)
+		{
+		    int arglen = strlen(argv[argn]);
+		    char buffer[arglen +1];
+		    argpos = -1;
+		    do
+		    {
+		 	for (n = 0; n <= arglen; ) buffer[n++] = '\0';
+			bufpos = 0;
+			for (argpos++; argpos <= arglen; argpos++)
+			{
+				c = argv[argn][argpos];
+				if (c == '\0' || c == ' ' || c == '\t')
+					break;
+				buffer[bufpos++] = c;
+			}
+			if (bufpos == 0)
+			{
+				if (c == '\0')
+					break;
+				continue;
+			}
+			RESIZE(altargv, char *, altargc +1);
+			altargv[altargc] = new_malloc(bufpos +1);
+			strcpy(altargv[altargc++], buffer);
+			if (c != '\0')
+				continue;
+			break;   	
+		    } while(1);
+		}
+		RESIZE(altargv, char *, altargc +1);
+		altargv[altargc] = NULL;
+		argc = altargc;
+		argv = altargv;
+	}
 
 	/*
 	 * Parse the command line arguments.
 	 */
-	while ((ch = getopt(argc, argv, "aBbc:dfFhH:l:L:n:oOp:qsvxz:")) != EOF)
+	while ((ch = getopt(argc, argv, "aBbc:dfFhH:l:L:n:oOp:qsSvxz:")) != EOF)
 	{
 		switch (ch)
 		{
@@ -680,8 +708,8 @@ static	void	parse_args (int argc, char **argv)
 				malloc_strcpy(&epicrc_file, optarg);
 				break;
 
-			case 'a': /* add server, not replace */
-				add_servers = 1;
+			case 'a': /* append server, not replace */
+				append_servers = 1;
 				break;
 
 			case 'q': /* quick startup -- no .ircrc */
@@ -691,6 +719,9 @@ static	void	parse_args (int argc, char **argv)
 			case 's': /* dont connect - let user choose server */
 				dont_connect = 1;
 				break;
+			
+			case 'S': /* dummy option - to not choke on -S */
+				break;
 
 			case 'b':
 /* siiiiiiiigh */
@@ -718,7 +749,7 @@ static	void	parse_args (int argc, char **argv)
 				break;
 
 			case 'B':
-				load_ircrc_right_away = 1;
+				/* Historical option */
 				break;
 
 			case 'c':
@@ -751,7 +782,7 @@ static	void	parse_args (int argc, char **argv)
 
 	for (; *argv; argc--, argv++)
 		if (**argv)
-			build_server_list(*argv, NULL);
+			add_servers(*argv, NULL);
 
 	if (!use_input && quick_startup)
 	{
@@ -764,7 +795,7 @@ static	void	parse_args (int argc, char **argv)
 		exit(1);
 	}
 
-	if (strcmp(nickname, "0") && !check_nickname(nickname, 1))
+	if (!check_nickname(nickname, 1))
 	{
 		fprintf(stderr, "Invalid nickname: [%s]\n", nickname);
 		fprintf(stderr, "Please restart IRC II with a valid nickname\n");
@@ -775,15 +806,15 @@ static	void	parse_args (int argc, char **argv)
 	 * Find and build the server lists...
 	 */
 	if ((ptr = getenv("IRCSERVER")))
-		build_server_list(ptr, NULL);
+		add_servers(ptr, NULL);
 
-	if (!server_list_size() || add_servers)
+	if (!server_list_size() || append_servers)
 	{
 		read_server_file();
 		if (!server_list_size())
 		{
 			ptr = malloc_strdup(DEFAULT_SERVER);
-			build_server_list(ptr, NULL);
+			add_servers(ptr, NULL);
 			new_free(&ptr);
 		}
 	}
@@ -811,21 +842,20 @@ static	void	parse_args (int argc, char **argv)
 		}
 	}
 
+	if (altargc > 0)
+	{
+		int argn;
+		for (argn = 0; argn < altargc; argn++)
+		{
+			new_free(&(altargv[altargc]));
+		}
+		new_free(&altargv);
+		altargc = 0;
+	}
 	
 	return;
 }
 
-/* fire scripted signal events -pegasus */
-void do_signals(void)
-{
-	int sig_no;
-
-	signals_caught[0] = 0;
-	for (sig_no = 0; sig_no < NSIG; sig_no++)
-		while (signals_caught[sig_no])
-			do_hook(SIGNAL_LIST, "%d %d", sig_no, signals_caught[sig_no]--);
-}
-
 /* 
  * io() is a ONE TIME THROUGH loop!  It simply does ONE check on the
  * file descriptors, and if there is nothing waiting, it will time
@@ -847,11 +877,7 @@ static	const	char	*caller[51] = { NULL }; /* XXXX */
 static	int		level = 0,
 			old_level = 0,
 			last_warn = 0;
-static 	const Timeval	right_away = { 0, 0 };
-
 	Timeval		timer;
-	fd_set		rd, wd;
-
 
 	level++;
 	get_time(&now);
@@ -861,59 +887,41 @@ static 	const Timeval	right_away = { 0, 0 };
 
 	if (x_debug & DEBUG_WAITS)
 	{
-		if (level != old_level)
-		{
-			yell("Moving from io level [%d] to level [%d] from [%s]", old_level, level, what);
-			old_level = level;
-		}
+	    if (level != old_level)
+	    {
+		yell("Moving from io level [%d] to level [%d] from [%s]", 
+				old_level, level, what);
+		old_level = level;
+	    }
 	}
 
+	/* Try to avoid letting recursion of io() to get out of control */
 	if (level && (level - last_warn == 5))
 	{
-		last_warn = level;
-		yell("io's recursion level is [%d],  [%s]<-[%s]<-[%s]<-[%s]<-[%s]", level, what, caller[level-1], caller[level-2], caller[level-3], caller[level-4]);
-		if (level % 50 == 0)
-			panic("Ahoy there matey!  Abandon ship!");
+	    last_warn = level;
+	    yell("io's recursion level is [%d],  [%s]<-[%s]<-[%s]<-[%s]<-[%s]",
+			level, what, caller[level-1], caller[level-2], 
+					caller[level-3], caller[level-4]);
+	    if (level % 50 == 0)
+		panic("Ahoy there matey!  Abandon ship!");
 	}
 	else if (level && (last_warn - level == 5))
-		last_warn -= 5;
-
+	    last_warn -= 5;
 
 	caller[level] = what;
 
-#if 1
-	/* 
-	 * XXXXXXX! It pains me greatly to call this here.
-	 * This is an experimental test to see if i can get
-	 * away with not having to call update_all_windows()
-	 * in about a zillion other places by doing it here.
-	 */
-	update_all_windows();
-#endif
-
-	/* SET UP FD SETS */
-	rd = readables;
-	wd = writables;
-
-	/* If there is a timer that expires sooner, wait for that */
-	/* There is now a timer at all times, so this is our baseline */
+	/* Calculate the time to the next timer timeout */
 	timer = TimerTimeout();
 
-	/* If for any reason the timeout is negative, do a poll */
-	if (time_diff(right_away, timer) < 0)
-		timer = right_away;
-
 	/* GO AHEAD AND WAIT FOR SOME DATA TO COME IN */
-	switch (new_select(&rd, &wd, &timer))
+	make_window_current(NULL);
+	switch (do_wait(&timer))
 	{
-		/* Timeout -- nothing interesting. */
+		/* Timeout -- Need to do timers */
 		case 0:
 		{
 			get_time(&now);
-#ifdef HAVE_SSL
-			/* Yes, this is slow, but we have to check for this */
-			do_server(&rd, &wd);
-#endif
+			ExecuteTimers();
 			break;
 		}
 
@@ -935,35 +943,47 @@ static 	const Timeval	right_away = { 0, 0 };
 		default:
 		{
 			get_time(&now);
-			make_window_current(NULL);
-			dcc_check(&rd, &wd);
-			do_server(&rd, &wd);
-			do_processes(&rd, &wd);
-			do_screens(&rd, &wd);
+			do_filedesc();
 			break;
 		} 
 	}
 
-	if (signals_caught[0] != 0)
-		do_signals();
-	ExecuteTimers();
+	/* 
+	 * Various things that need to be done synchronously...
+	 */
+
+	/* Account for dead child processes */
 	get_child_exit(-1);
+
+	/* Run /DEFERed commands */
 	if (level == 1 && need_defered_commands)
 		do_defered_commands();
 
+	/* Make sure all the servers are connected that ought to be */
+	window_check_servers();
+
+	/* Make sure all the channels are joined that ought to be */
+	window_check_channels();
+
+	/* Redraw the screen after a SIGCONT */
+	if (need_redraw)
+		redraw_all_screens();
+
+	/* Make sure all the windows and status bars are made current */
+	update_all_windows();
+
+	/* Move the cursor back to the input line */
 	cursor_to_input();
 
-	/* (set in term.c) -- we need to redraw the screen */
-	if (need_redraw)
-		refresh_a_screen(main_screen);
-
-	window_check_channels();
-	update_all_windows();
+	/* Reclaim any alloca()ed space */
 	alloca(0);
+
+	/* Release this io() accounting level */
 	caller[level] = NULL;
 	level--;
 
 #ifdef DELAYED_FREES
+	/* Reclaim any malloc()ed space */
 	if (level == 0 && need_delayed_free)
 		do_delayed_frees();
 #endif
@@ -1118,6 +1138,22 @@ static void check_invalid_host (void)
 	return;
 }
 
+/*************************************************************************/
+void    load_ircrc (void)
+{
+	if (startup_file || quick_startup)
+		return;
+
+	if (access(epicrc_file, R_OK) == 0)
+		startup_file = malloc_strdup(epicrc_file);
+	else if (access(ircrc_file, R_OK) == 0)
+		startup_file = malloc_strdup(ircrc_file);
+	else
+		startup_file = malloc_strdup("global");
+
+	load("LOAD", startup_file, empty_string);
+}
+
 /*************************************************************************/
 int 	main (int argc, char *argv[])
 {
@@ -1128,11 +1164,18 @@ int 	main (int argc, char *argv[])
 	check_password();
 	check_valid_user();
 	check_invalid_host();
+
+	init_levels();
+	init_variables_stage1();
 	parse_args(argc, argv);
 	init_binds();
 	init_keys();
+	init_commands();
+	init_functions();
+	init_expandos();
+	init_newio();
 
-	fprintf(stderr, "EPIC Version 4 -- %s\n", ridiculous_version_name);
+	fprintf(stderr, "EPIC Version 5 -- %s\n", ridiculous_version_name);
 	fprintf(stderr, "EPIC Software Labs (2004)\n");
 	fprintf(stderr, "Version (%s), Commit Id (%lu) -- Date (%s)\n", irc_version, commit_id, internal_version);
 	fprintf(stderr, "%s\n", compile_info);
@@ -1143,23 +1186,10 @@ int 	main (int argc, char *argv[])
 		dumb_mode = 1;
 	fprintf(stderr, "\n");
 
-	FD_ZERO(&readables);
-	FD_ZERO(&writables);
-	FD_ZERO(&held_readables);
-	FD_ZERO(&held_writables);
-
 	/* If we're a bot, do the bot thing. */
 	if (!use_input && fork())
 		_exit(0);
 
-	/* make sure we don't start with spurious signals events firing */
-	memset(&signals_caught, 0, NSIG * sizeof(int));
-	/* hook all signals! */
-	init_signals();
-	/* we *might* want to check for SIG_ERR from the above function.
-	 * i leave it to hop to decide what to do on SIG_ERR. -pegasus 
-	 */
-
 	/* these should be taken by both dumb and smart displays */
 	my_signal(SIGSEGV, coredump);
 	my_signal(SIGBUS, coredump);
@@ -1171,13 +1201,17 @@ int 	main (int argc, char *argv[])
 	my_signal(SIGINT, cntl_c);
 	my_signal(SIGALRM, nothing);
 	my_signal(SIGUSR1, sig_user1);
+	my_signal(SIGUSR2, sig_user2);
+
+	message_from(NULL, LEVEL_OTHER);
 
 	if ((dumb_mode == 0) && (init_screen() == 0))
 	{
 		my_signal(SIGCONT, term_cont);
 		my_signal(SIGWINCH, sig_refresh_screen);
-		init_variables();
+		init_variables_stage2();
 		build_status(NULL);
+		permit_status_update(1);
 		update_input(UPDATE_ALL);
 	}
 	else
@@ -1190,7 +1224,7 @@ int 	main (int argc, char *argv[])
 		dumb_mode = 1;		/* Just in case */
 		create_new_screen();
 		new_window(main_screen);
-		init_variables();
+		init_variables_stage2();
 		build_status(NULL);
 	}
 
@@ -1200,17 +1234,17 @@ int 	main (int argc, char *argv[])
 	/* The all-collecting stack frame */
 	make_local_stack("TOP");
 
-	/* XXXX Move this somewhere else eventually XXXX */
-	if (load_ircrc_right_away)
-		load_ircrc();
+	load_ircrc();
 
 	set_input(empty_string);
-	set_input_prompt(get_string_var(INPUT_PROMPT_VAR));
 
 	if (dont_connect)
 		display_server_list();		/* Let user choose server */
 	else
-		reconnect(NOSERV, 0);		/* Connect to default server */
+	{
+		current_window->server = 0;	/* Connect to default server */
+		window_check_servers();
+	}
 
 	get_time(&idle_time);
 	reset_system_timers();
