@@ -1,18 +1,40 @@
+/* $EPIC: lastlog.c,v 1.63 2006/06/29 01:13:53 jnelson Exp $ */
 /*
  * lastlog.c: handles the lastlog features of irc. 
  *
- * Written By Michael Sandrof
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1993, 2003 EPIC Software Labs.
+ * All rights reserved.
  *
- * Copyright(c) 1990 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
  *
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
-
-#if 0
-static	char	rcsid[] = "@(#)$Id: lastlog.c,v 1.12 2002/02/14 01:59:55 jnelson Exp $";
-#endif
-
 #include "irc.h"
+#include "levels.h"
 #include "lastlog.h"
 #include "window.h"
 #include "screen.h"
@@ -20,175 +42,222 @@ static	char	rcsid[] = "@(#)$Id: lastlog.c,v 1.12 2002/02/14 01:59:55 jnelson Exp
 #include "ircaux.h"
 #include "output.h"
 #include "numbers.h"
-#include <regex.h>
+#include "functions.h"
+#include "reg.h"
 
-static int	show_lastlog (Lastlog **l, int *skip, int *number, int, char *match, regex_t *reg, int *max);
+typedef struct	lastlog_stru
+{
+	int	level;
+	char	*target;
+	char	*msg;
+	struct	lastlog_stru	*older;
+	struct	lastlog_stru	*newer;
+	time_t	when;
+	int	visible;
+	intmax_t refnum;
+	int	winref;
+}	Lastlog;
 
+static	intmax_t global_lastlog_refnum = 0;
+
+static int	show_lastlog (Lastlog **l, int *skip, int *number, Mask *level_mask, char *match, regex_t *reg, int *max, const char *target, int mangler);
+
+Lastlog *	lastlog_oldest = NULL;
+Lastlog *	lastlog_newest = NULL;
+
+/**********************************************************************/
 /*
  * lastlog_level: current bitmap setting of which things should be stored in
  * the lastlog.  The LOG_MSG, LOG_NOTICE, etc., defines tell more about this 
  */
-static	int	lastlog_level;
-static	int	notify_level;
-	int	new_server_lastlog_level;
-	int	beep_on_level;
-	int 	current_window_level = 0;
-
-/*
- * msg_level: the mask for the current message level.  What?  Did he really
- * say that?  This is set in the set_lastlog_msg_level() routine as it
- * compared to the lastlog_level variable to see if what ever is being added
- * should actually be added 
- */
-static	int	msg_level = LOG_CRAP;
-
-#define NUMBER_OF_LEVELS 16
-static	char	*levels[] =
-{
-	"CRAP",		"PUBLIC",	"MSGS",		"NOTICES",
-	"WALLS",	"WALLOPS",	"NOTES",	"OPNOTES",
-	"SNOTES",	"ACTIONS",	"DCC",		"CTCP",
-	"USERLOG1",	"USERLOG2",	"USERLOG3",	"USERLOG4"
-};
+static	Mask	lastlog_mask;
+static	Mask	notify_mask;
+	Mask *	new_server_lastlog_mask = NULL;
+	Mask *	old_server_lastlog_mask = NULL;
+	Mask 	current_window_mask;
 
 /*
  * warn_lastlog_level: tells the user what levels are available.
  */
 static void	warn_lastlog_levels (void)
 {
-	char buffer[BIG_BUFFER_SIZE + 1];
-	int i;
+	char *buffer;
 
-	strmcpy(buffer, "Valid levels: ", BIG_BUFFER_SIZE);
-	for (i = 0; i < NUMBER_OF_LEVELS; i++)
-	{
-		strmcat(buffer, levels[i], BIG_BUFFER_SIZE);
-		strmcat(buffer, " ", BIG_BUFFER_SIZE);
-	}
+	strlcpy(buffer, "Valid levels: ", sizeof buffer);
+	buffer = get_all_levels();
 	say("%s", buffer);
+	new_free(&buffer);	
 }
 
 /*
- * bits_to_lastlog_level: converts the bitmap of lastlog levels into a nice
- * string format.  Note that this uses the global buffer, so watch out 
- */
-char	*bits_to_lastlog_level (int level)
-{
-	static	char	buffer[256]; /* this *should* be enough for this */
-	int	i,
-		p;
-
-	if (level == LOG_ALL)
-		strcpy(buffer, "ALL");
-	else if (level == 0)
-		strcpy(buffer, "NONE");
-	else
-	{
-		*buffer = '\0';
-		for (i = 0, p = 1; i < NUMBER_OF_LEVELS; i++, p *= 2)
-		{
-			if (level & p)
-			{
-				if (*buffer)
-					strmcat(buffer, " ", 255);
-				strmcat(buffer, levels[i], 255);
-			}
-		}
-	}
-	return (buffer);
-}
-
-int	parse_lastlog_level (char *str)
-{
-	char	*ptr,
-		*rest;
-	int	len,
-		i,
-		p,
-		level,
-		neg;
-	int	warn = 0;
-
-	level = 0;
-	while ((str = next_arg(str, &rest)) != NULL)
-	{
-	    while (str)
-	    {
-		if ((ptr = strchr(str, ',')) != NULL)
-			*ptr++ = '\0';
-		if ((len = strlen(str)) != 0)
-		{
-			if (my_strnicmp(str, "ALL", len) == 0)
-				level = LOG_ALL;
-			else if (my_strnicmp(str, "NONE", len) == 0)
-				level = 0;
-			else
-			{
-				if (*str == '-')
-				{
-					str++, len--;
-					neg = 1;
-				}
-				else
-					neg = 0;
-
-				for (i = 0, p = 1; i < NUMBER_OF_LEVELS;
-						i++, p *= 2)
-				{
-					if (!my_strnicmp(str, levels[i], len))
-					{
-						if (neg)
-							level &= (LOG_ALL ^ p);
-						else
-							level |= p;
-						break;
-					}
-				}
-				if (i == NUMBER_OF_LEVELS)
-				{
-					say("Unknown level: %s", str);
-					if (!warn)
-					{
-						warn_lastlog_levels();
-						warn = 1;
-					}
-				}
-			}
-		}
-		str = ptr;
-	    }
-	    str = rest;
-	}
-	return (level);
-}
-
-/*
- * set_lastlog_level: called whenever a "SET LASTLOG_LEVEL" is done.  It
- * parses the settings and sets the lastlog_level variable appropriately.  It
+ * set_lastlog_mask: called whenever a "SET LASTLOG_LEVEL" is done.  It
+ * parses the settings and sets the lastlog_mask variable appropriately.  It
  * also rewrites the LASTLOG_LEVEL variable to make it look nice 
  */
-void	set_lastlog_level (char *str)
+void	set_lastlog_mask (void *stuff)
 {
-	lastlog_level = parse_lastlog_level(str);
-	set_string_var(LASTLOG_LEVEL_VAR, bits_to_lastlog_level(lastlog_level));
-	current_window->lastlog_level = lastlog_level;
+	VARIABLE *v;
+	const char *str;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	str_to_mask(&lastlog_mask, str);
+	malloc_strcpy(&v->string, mask_to_str(&lastlog_mask));
+
+	current_window->lastlog_mask = lastlog_mask;
 }
 
-void	set_new_server_lastlog_level (char *str)
+void	set_new_server_lastlog_mask (void *stuff)
 {
-	new_server_lastlog_level = parse_lastlog_level(str);
-	set_string_var(NEW_SERVER_LASTLOG_LEVEL_VAR, 
-			bits_to_lastlog_level(new_server_lastlog_level));
+	VARIABLE *v;
+	const char *str;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	if (str)
+	{
+	    if (!new_server_lastlog_mask)
+	    {
+		new_server_lastlog_mask = (Mask *)new_malloc(sizeof(Mask));
+		mask_unsetall(new_server_lastlog_mask);
+	    }
+	    str_to_mask(new_server_lastlog_mask, str);
+	    malloc_strcpy(&v->string, mask_to_str(new_server_lastlog_mask));
+	}
+	else
+	    new_free(&new_server_lastlog_mask);
+}
+
+void	set_old_server_lastlog_mask (void *stuff)
+{
+	VARIABLE *v;
+	const char *str;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	if (str)
+	{
+	    if (!old_server_lastlog_mask)
+	    {
+		old_server_lastlog_mask = (Mask *)new_malloc(sizeof(Mask));
+		mask_unsetall(old_server_lastlog_mask);
+	    }
+	    str_to_mask(old_server_lastlog_mask, str);
+	    malloc_strcpy(&v->string, mask_to_str(old_server_lastlog_mask));
+	}
+	else
+	    new_free(&old_server_lastlog_mask);
+}
+
+Mask	real_notify_mask (void)
+{
+	return (notify_mask);
+}
+
+Mask	real_lastlog_mask (void)
+{
+	return (lastlog_mask);
+}
+
+void	set_notify_mask (void *stuff)
+{
+	VARIABLE *v;
+	const char *str;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	str_to_mask(&notify_mask, str);
+	malloc_strcpy(&v->string, mask_to_str(&notify_mask));
+
+	current_window->notify_mask = notify_mask;
+}
+
+void	set_current_window_mask (void *stuff)
+{
+	VARIABLE *v;
+	const char *str;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	str_to_mask(&current_window_mask, str);
+	malloc_strcpy(&v->string, mask_to_str(&current_window_mask));
 }
 
 
-void 	remove_from_lastlog (Window *window)
+/**********************************************************************/
+/*
+ * add_to_lastlog: adds the line to the lastlog.  If the LASTLOG_CONVERSATION
+ * variable is on, then only those lines that are user messages (private
+ * messages, channel messages, wall's, and any outgoing messages) are
+ * recorded, otherwise, everything is recorded 
+ */
+intmax_t	add_to_lastlog (Window *window, const char *line)
+{
+	Lastlog *new_l;
+
+	if (!window)
+		window = current_window;
+
+	new_l = (Lastlog *)new_malloc(sizeof(Lastlog));
+	new_l->refnum = global_lastlog_refnum++;
+	new_l->older = window->lastlog_newest;
+	new_l->newer = NULL;
+	new_l->level = who_level;
+	new_l->msg = malloc_strdup(line);
+	if (who_from)
+		new_l->target = malloc_strdup(who_from);
+	else
+		new_l->target = NULL;
+	time(&new_l->when);
+
+	/* * * */
+	if (window->lastlog_newest)
+		window->lastlog_newest->newer = new_l;
+	window->lastlog_newest = new_l;
+
+	if (!window->lastlog_oldest)
+		window->lastlog_oldest = window->lastlog_newest;
+
+	if (mask_isset(&window->lastlog_mask, who_level))
+	{
+		new_l->visible = 1;
+		window->lastlog_size++;
+		trim_lastlog(window);
+	}
+	else
+		new_l->visible = 0;
+
+	/* * * */
+#if 0
+	lastlog_newest->newer = new_l;
+	new_l->older = lastlog_newest;
+	lastlog_newest = new_l;
+
+	if (!lastlog_oldest)
+		lastlog_oldest = lastlog_newest;
+#endif
+
+	/* * * */
+	return new_l->refnum;
+}
+
+
+void 	trim_lastlog (Window *window)
 {
 	Lastlog *new_oldest;
 	Lastlog *being_removed;
 
-	if (window->lastlog_oldest)
+	while (window->lastlog_oldest)
 	{
+		/* All done! */
+		if (window->lastlog_size <= window->lastlog_max)
+			return;
+
 		being_removed = window->lastlog_oldest;
 		new_oldest = being_removed->newer;
 		window->lastlog_oldest = new_oldest;
@@ -196,12 +265,15 @@ void 	remove_from_lastlog (Window *window)
 			new_oldest->older = NULL;
 		else
 			window->lastlog_newest = NULL;
-		window->lastlog_size--;
-		new_free((char **)&(being_removed->msg));
+		if (being_removed->visible)
+			window->lastlog_size--;
+		new_free((char **)&being_removed->msg);
+		new_free((char **)&being_removed->target);
 		new_free((char **)&being_removed);
 	}
-	else
-		window->lastlog_size = 0;
+
+	/* Uh, the lastlog must be empty... */
+	window->lastlog_size = 0;
 }
 
 /*
@@ -209,35 +281,22 @@ void 	remove_from_lastlog (Window *window)
  * has gotten larger than it was before, all newer lastlog entries remain.
  * If it get smaller, some are deleted from the end. 
  */
-void	set_lastlog_size (int size)
+void	set_lastlog_size (void *stuff)
 {
-	int	i,
-		diff;
+	VARIABLE *v;
+	int	size;
 	Window	*window = NULL;
 
+	v = (VARIABLE *)stuff;
+	size = v->integer;
+
 	while (traverse_all_windows(&window))
 	{
-		if (window->lastlog_size > size)
-		{
-			diff = window->lastlog_size - size;
-			for (i = 0; i < diff; i++)
-				remove_from_lastlog(window);
-		}
 		window->lastlog_max = size;
+		trim_lastlog(window);
 	}
 }
 
-/* set_lastlog_msg_level: sets the message level for recording in the lastlog */
-int 	set_lastlog_msg_level (int level)
-{
-	int	old;
-
-	old = msg_level;
-	msg_level = level;
-	return (old);
-}
-
-
 /*
  * The /LASTLOG command:
  * Syntax:
@@ -299,33 +358,38 @@ int 	set_lastlog_msg_level (int level)
 BUILT_IN_COMMAND(lastlog)
 {
 	int		reverse = 0;
-	int		level_mask = 0;
+	Mask		level_mask;
 	int		skip = -1;
 	int		number = INT_MAX;
 	int		max = -1;
 	char *		match = NULL;
+	char *		target = NULL;
 	char *		regex = NULL;
 	Lastlog *	start;
 	Lastlog *	end;
 	Lastlog *	l;
+	Lastlog *	lastshown;
 	regex_t 	realreg;
 	regex_t *	reg = NULL;
 	int		cnt;
 	char *		arg;
 	int		header = 1;
-	int		save_level;
+	Mask		save_mask;
 	int		before = -1;
 	int		after = 0;
 	int		counter = 0;
 	int		show_separator = 0;
-	char *		separator = "----";
+	const char *	separator = "----";
 	char *		outfile = NULL;
 	FILE *		outfp = NULL;
+	int		mangler = 0;
+	int		lc;
 
-	message_from(NULL, LOG_CURRENT);
+	lc = message_setall(0, NULL, LEVEL_OTHER);
 	cnt = current_window->lastlog_size;
-	save_level = current_window->lastlog_level;
-	current_window->lastlog_level = 0;
+	save_mask = current_window->lastlog_mask;
+	mask_unsetall(&current_window->lastlog_mask);
+	mask_unsetall(&level_mask);
 
 	while ((arg = new_next_arg(args, &args)) != NULL)
 	{
@@ -349,6 +413,14 @@ BUILT_IN_COMMAND(lastlog)
 			goto bail;
 		}
 	    }
+	    else if (!my_strnicmp(arg, "-TARGET", len))
+	    {
+		if (!(target = new_next_arg(args, &args)))
+		{
+			yell("LASTLOG -TARGET requires an argument.");
+			goto bail;
+		}
+	    }
 	    else if (!my_strnicmp(arg, "-MAXIMUM", len))
 	    {
 		char *x = new_next_arg(args, &args);
@@ -364,6 +436,12 @@ BUILT_IN_COMMAND(lastlog)
 			goto bail;
 		}
 	    }
+	    else if (!my_strnicmp(arg, "-MANGLE", len))
+	    {
+		char *	x = new_next_arg(args, &args);
+
+		mangler = parse_mangle(x, mangler, NULL);
+	    }
 	    else if (!my_strnicmp(arg, "-SKIP", len))
 	    {
 		char *x = new_next_arg(args, &args);
@@ -400,10 +478,10 @@ BUILT_IN_COMMAND(lastlog)
 
 		x = new_next_arg(args, &args);
 		before_str = x;
-		if ((after_str = strchr(x, ',')))
+		if (x && (after_str = strchr(x, ',')))
 			*after_str++ = 0;
 
-		if (!is_number(before_str))
+		if (!x || !is_number(before_str))
 		{
 			yell("LASTLOG -CONTEXT requires a numeric argument.");
 			goto bail;
@@ -433,21 +511,15 @@ BUILT_IN_COMMAND(lastlog)
 	    else if (!my_strnicmp(arg, "-REVERSE", len))
 		reverse = 1;
 	    else if (!my_strnicmp(arg, "-ALL", len))
-		level_mask = LOG_ALL;
+		mask_setall(&level_mask);
 	    else if (!my_strnicmp(arg, "--ALL", len))
-		level_mask = 0;
+		mask_unsetall(&level_mask);
 	    else if (!my_strnicmp(arg, "--", 2))
 	    {
-		int	i;
-		for (i = 0; i < NUMBER_OF_LEVELS; i++)
-		{
-		    if (!my_strnicmp(levels[i], arg+2, len-2))
-		    {
-			level_mask &= ~(1 << i);
-			break;
-		    }
-		}
-		if (i == NUMBER_OF_LEVELS)
+		int	i = str_to_level(arg+2);
+		if (i != -1)
+			mask_unset(&level_mask, i);
+		else
 		{
 			say("Unknown flag: %s", arg);
 			goto bail;
@@ -455,16 +527,10 @@ BUILT_IN_COMMAND(lastlog)
 	    }
 	    else if (!my_strnicmp(arg, "-", 1))
 	    {
-		int	i;
-		for (i = 0; i < NUMBER_OF_LEVELS; i++)
-		{
-		    if (!my_strnicmp(levels[i], arg+1, len-1))
-		    {
-			level_mask |= (1 << i);
-			break;
-		    }
-		}
-		if (i == NUMBER_OF_LEVELS)
+		int	i = str_to_level(arg+1);
+		if (i != -1)
+			mask_set(&level_mask, i);
+		else
 		{
 			say("Unknown flag: %s", arg);
 			goto bail;
@@ -511,8 +577,12 @@ BUILT_IN_COMMAND(lastlog)
 	}
 	if (match)
 	{
-		char *	blah = alloca(strlen(match) + 3);
-		sprintf(blah, "*%s*", match);
+		size_t	size;
+		char *	blah;
+
+		size = strlen(match) + 3;
+		blah = alloca(size);
+		snprintf(blah, size, "*%s*", match);
 		match = blah;
 	}
 	if (regex)
@@ -533,14 +603,15 @@ BUILT_IN_COMMAND(lastlog)
 	if (x_debug & DEBUG_LASTLOG)
 	{
 		yell("Lastlog summary status:");
-		yell("Pattern: [%s]", match);
-		yell("Regex: [%s]", regex);
+		yell("Pattern: [%s]", match ? match : "<none>");
+		yell("Regex: [%s]", regex ? regex : "<none>");
+		yell("Target: [%s]", target ? target : "<none>");
 		yell("Header: %d", header);
 		yell("Reverse: %d", reverse);
 		yell("Skip: %d", skip);
 		yell("Number: %d", number);
 		yell("Max: %d", max);
-		yell("Mask: %d", level_mask);
+		yell("Mask: %s", mask_to_str(&level_mask));
 	}
 
 	if (outfile)
@@ -554,7 +625,7 @@ BUILT_IN_COMMAND(lastlog)
 
 	/* Iterate over the lastlog here */
 	if (header)
-		file_put_it(outfp, "%s Lastlog:", numeric_banner());
+		file_put_it(outfp, "%s Lastlog:", banner());
 
 	/*
 	 * Ugh.  This is way too complicated for its own good.  Let's
@@ -583,42 +654,57 @@ BUILT_IN_COMMAND(lastlog)
 
 	    end = current_window->lastlog_newest;
 	    start = end;
-	    for (i = 1; i < number; i++)
+	    for (i = 0; start != current_window->lastlog_oldest; )
 	    {
-		if (start == current_window->lastlog_oldest)
+		if (mask_isnone(&level_mask) || (mask_isset(&level_mask, start->level)))
+			i++;
+		if (i == number)
 			break;
 		start = start->older;
 	    }
 
+	    lastshown = NULL;
 	    for (l = start; l; (void)(l && (l = l->newer)))
 	    {
-		if (show_lastlog(&l, &skip, &number, level_mask, 
-				match, reg, &max))
+		if (show_lastlog(&l, &skip, &number, &level_mask, 
+				match, reg, &max, target, mangler))
 		{
-		    if (show_separator)
-		    {
-			file_put_it(outfp, "%s", separator);
-			show_separator = 0;
-		    }
-
 		    if (counter == 0 && before > 0)
 		    {
-			int i;
-
+			counter = 1;
 			for (i = 0; i < before; i++)
-			     if (l && l->older)
-				l = l->older;
-			counter = before + 1;
+			{
+			    if (l && l == lastshown)
+			    {
+				if (l->newer)
+				    l = l->newer;
+				show_separator = 0;
+				break;
+			    }
 
+			    if (l && l->older)
+			    {
+				l = l->older;
+				counter++;
+			    }
+			}
 		    }
 		    else if (after != -1)
 			counter = after + 1;
 		    else
 			counter = 1;
+
+		    if (show_separator)
+		    {
+			file_put_it(outfp, "%s", separator);
+			show_separator = 0;
+		    }
 		}
+
 		if (counter)
 		{
 			file_put_it(outfp, "%s", l->msg);
+			lastshown = l;
 			counter--;
 			if (counter == 0 && before != -1 && separator)
 				show_separator = 1;
@@ -634,40 +720,56 @@ BUILT_IN_COMMAND(lastlog)
 
 	    start = current_window->lastlog_newest;
 	    end = start;
-	    for (i = 1; i < number; i++)
+	    for (i = 0; end != current_window->lastlog_oldest; )
 	    {
-		if (end == current_window->lastlog_oldest)
+		if (mask_isnone(&level_mask) || (mask_isset(&level_mask, start->level)))
+			i++;
+		if (i == number)
 			break;
 		end = end->older;
 	    }
 
+	    lastshown = NULL;
 	    for (l = start; l; (void)(l && (l = l->older)))
 	    {
-		if (show_lastlog(&l, &skip, &number, level_mask, 
-				match, reg, &max))
+		if (show_lastlog(&l, &skip, &number, &level_mask, 
+				match, reg, &max, target, mangler))
 		{
-		    if (show_separator)
-		    {
-			file_put_it(outfp, "%s", separator);
-			show_separator = 0;
-		    }
-
 		    if (counter == 0 && before > 0)
 		    {
-			int i;
+			counter = 1;
 			for (i = 0; i < before; i++)
-			     if (l && l->newer)
+			{
+			    if (l && l == lastshown)
+			    {
+				if (l->older)
+				    l = l->older;
+				show_separator = 0;
+				break;
+			    }
+
+			    if (l && l->newer)
+			    {
 				l = l->newer;
-			counter = before + 1;
+				counter++;
+			    }
+			}
 		    }
 		    else if (after != -1)
 			counter = after + 1;
 		    else
 			counter = 1;
+
+		    if (show_separator)
+		    {
+			file_put_it(outfp, "%s", separator);
+			show_separator = 0;
+		    }
 		}
 		if (counter)
 		{
 			file_put_it(outfp, "%s", l->msg);
+			lastshown = l;
 			counter--;
 			if (counter == 0 && before != -1 && separator)
 				show_separator = 1;
@@ -677,15 +779,14 @@ BUILT_IN_COMMAND(lastlog)
 	    }
 	}
 	if (header)
-		file_put_it(outfp, "%s End of Lastlog", numeric_banner());
+		file_put_it(outfp, "%s End of Lastlog", banner());
 bail:
 	if (outfp)
 		fclose(outfp);
 	if (reg)
 		regfree(reg);
-	current_window->lastlog_level = save_level;
-	set_lastlog_msg_level(msg_level);
-	message_from(NULL, LOG_CRAP);
+	current_window->lastlog_mask = save_mask;
+	pop_message_from(lc);
 	return;
 }
 
@@ -693,8 +794,17 @@ bail:
  * This returns 1 if the current item pointed to by 'l' is something that
  * should be displayed based on the criteron provided.
  */
-static int	show_lastlog (Lastlog **l, int *skip, int *number, int level_mask, char *match, regex_t *reg, int *max)
+static int	show_lastlog (Lastlog **l, int *skip, int *number, Mask *level_mask, char *match, regex_t *reg, int *max, const char *target, int mangler)
 {
+	const char *str = NULL;
+
+	if ((*l)->visible == 0)
+	{
+		if (x_debug & DEBUG_LASTLOG)
+			yell("Lastlog item is not visible");
+		return 0;
+	}
+
 	if (*skip > 0)
 	{
 		if (x_debug & DEBUG_LASTLOG)
@@ -702,37 +812,46 @@ static int	show_lastlog (Lastlog **l, int *skip, int *number, int level_mask, ch
 		(*skip)--;	/* Have not skipped enough leading records */
 		return 0;
 	}
-#if 0
-	if (*number == 0)
-	{
-		if (x_debug & DEBUG_LASTLOG)
-			yell("Number == 0");
-		*l = NULL;	/* Have iterated over the max num. records */
-		return 0;
-	}
-	if (*number > 0)
-		(*number)--;	/* Have not yet iterated over max records */
-#endif
 
-	if (level_mask && (((*l)->level & level_mask) == 0))
+	if (!mask_isnone(level_mask) && !mask_isset(level_mask, (*l)->level))
 	{
 		if (x_debug & DEBUG_LASTLOG)
-			yell("Level_mask != level ([%d] [%d])",
-				level_mask, (*l)->level);
+			yell("Level_mask != level ([%s] [%s])",
+				mask_to_str(level_mask), level_to_str((*l)->level));
 		return 0;			/* Not of proper level */
 	}
-	if (match && !wild_match(match, (*l)->msg))
+
+	if (mangler)
+	{
+		char *	output, *result;
+
+		result = new_normalize_string((*l)->msg, 1, mangler);
+		output = LOCAL_COPY(result);
+		new_free(&result);
+		str = output;
+	}
+	else
+		str = (*l)->msg;
+
+	if (match && !wild_match(match, str))
 	{
 		if (x_debug & DEBUG_LASTLOG)
-			yell("Line [%s] not matched [%s]", (*l)->msg, match);
+			yell("Line [%s] not matched [%s]", str, match);
 		return 0;			/* Pattern match failed */
 	}
-	if (reg && regexec(reg, (*l)->msg, 0, NULL, 0))
+	if (reg && regexec(reg, str, 0, NULL, 0))
 	{
 		if (x_debug & DEBUG_LASTLOG)
-			yell("Line [%s] not regexed", (*l)->msg);
+			yell("Line [%s] not regexed", str);
 		return 0;			/* Regex match failed */
 	}
+	if (target && (!(*l)->target || !wild_match(target, (*l)->target)))
+	{
+		if (x_debug & DEBUG_LASTLOG)
+			yell("Target [%s] not matched [%s]", 
+					(*l)->target, target);
+		return 0;			/* Target match failed */
+	}
 	if (*max == 0)
 	{
 		if (x_debug & DEBUG_LASTLOG)
@@ -747,75 +866,18 @@ static int	show_lastlog (Lastlog **l, int *skip, int *number, int level_mask, ch
 }
 
 /*
- * add_to_lastlog: adds the line to the lastlog.  If the LASTLOG_CONVERSATION
- * variable is on, then only those lines that are user messages (private
- * messages, channel messages, wall's, and any outgoing messages) are
- * recorded, otherwise, everything is recorded 
+ * reconstitute_scrollback: walk through the lastlog, and put_it everything,
+ * making sure to reset the level and all that jazz.  This will cause the 
+ * scrollback to be rebroken, etc.
  */
-void 	add_to_lastlog (Window *window, const char *line)
+void	reconstitute_scrollback (Window *window)
 {
-	Lastlog *new_l;
+	Lastlog *li;
 
-	if (!window)
-		window = current_window;
-
-	if (window->lastlog_level & msg_level)
-	{
-		new_l = (Lastlog *)new_malloc(sizeof(Lastlog));
-		new_l->older = window->lastlog_newest;
-		new_l->newer = NULL;
-		new_l->level = msg_level;
-		new_l->msg = m_strdup(line);
-
-		if (window->lastlog_newest)
-			window->lastlog_newest->newer = new_l;
-		window->lastlog_newest = new_l;
-
-		if (!window->lastlog_oldest)
-			window->lastlog_oldest = window->lastlog_newest;
-
-		if (window->lastlog_size++ >= window->lastlog_max)
-			remove_from_lastlog(window);
-	}
+	for (li = window->lastlog_oldest; li; li = li->newer)
+		add_to_window_scrollback(window, li->msg, li->refnum);
 }
-
-int	real_notify_level (void)
-{
-	return (notify_level);
-}
-
-int	real_lastlog_level (void)
-{
-	return (lastlog_level);
-}
-
-void	set_notify_level (char *str)
-{
-	notify_level = parse_lastlog_level(str);
-	set_string_var(NOTIFY_LEVEL_VAR, bits_to_lastlog_level(notify_level));
-	current_window->notify_level = notify_level;
-}
-
-void 	set_beep_on_msg (char *str)
-{
-	beep_on_level = parse_lastlog_level(str);
-	set_string_var(BEEP_ON_MSG_VAR, bits_to_lastlog_level(beep_on_level));
-}
-
-void	set_current_window_level (char *str)
-{
-	current_window_level = parse_lastlog_level(str);
-	set_string_var(CURRENT_WINDOW_LEVEL_VAR, 
-			bits_to_lastlog_level(current_window_level));
-}
-
-#define EMPTY empty_string
-#define RETURN_EMPTY return m_strdup(EMPTY)
-#define RETURN_IF_EMPTY(x) if (empty( x )) RETURN_EMPTY
-#define GET_INT_ARG(x, y) {RETURN_IF_EMPTY(y); x = my_atol(safe_new_next_arg(y, &y));}
-#define GET_STR_ARG(x, y) {RETURN_IF_EMPTY((y)); x = new_next_arg((y), &(y));RETURN_IF_EMPTY((x));}
-#define RETURN_STR(x) return m_strdup(x ? x : EMPTY);
-
+	
 /*
  * $line(<line number> [window number])
  * Returns the text of logical line <line number> from the lastlog of 
@@ -826,14 +888,17 @@ void	set_current_window_level (char *str)
  * Lines are numbered from 1, starting at the most recent line in the buffer.
  * Contributed by Crackbaby (Matt Carothers) on March 19, 1998.
  */
-char 	*function_line (char *word)
+BUILT_IN_FUNCTION(function_line, word)
 {
 	int	line = 0;
-	char *	windesc = zero;
+	const char *	windesc = zero;
 	Lastlog	*start_pos;
 	Window	*win;
 	char	*extra;
 	int	do_level = 0;
+	int	do_timestamp = 0;
+	char *	retval = NULL;
+	size_t	clue = 0;
 
 	GET_INT_ARG(line, word);
 
@@ -843,6 +908,8 @@ char 	*function_line (char *word)
 
 		if (!my_stricmp(extra, "-LEVEL"))
 			do_level = 1;
+		else if (!my_stricmp(extra, "-TIME"))
+			do_timestamp = 1;
 		else
 			windesc = extra;
 	}
@@ -857,18 +924,33 @@ char 	*function_line (char *word)
 
 	/* Get the line from the lastlog */
 	for (start_pos = win->lastlog_newest; line; start_pos = start_pos->older)
-		line--;
+	{
+		if (start_pos->visible)
+			line--;
+	}
 
 	if (!start_pos)
 		start_pos = win->lastlog_oldest;
 	else
 		start_pos = start_pos->newer;
 
+	while (!start_pos->visible && start_pos->newer)
+		start_pos = start_pos->newer;
+
+	/* If there are no visible lastlog items, punt */
+	if (!start_pos)
+		RETURN_EMPTY;
+
+	malloc_strcat_c(&retval, start_pos->msg, &clue);
+
 	if (do_level)
-		return m_sprintf("%s %s", start_pos->msg, 
-					levels[start_pos->level]);
-	else
-		RETURN_STR(start_pos->msg);
+		malloc_strcat_wordlist_c(&retval, space, 
+					level_to_str(start_pos->level), &clue);
+	if (do_timestamp)
+		malloc_strcat_wordlist_c(&retval, space, 
+					NUMSTR(start_pos->when), &clue);
+
+	RETURN_MSTR(retval);
 }
 
 /*
@@ -878,20 +960,20 @@ char 	*function_line (char *word)
  * levels as represented by the lastlog levels. If the window number is 
  * invalid, the function will return the false value.
  */
-char *function_lastlog (char *word)
+BUILT_IN_FUNCTION(function_lastlog, word)
 {
-	char *	windesc = zero;
+	const char *	windesc = zero;
 	char *	pattern = NULL;
 	char *	retval = NULL;
 	Lastlog	*iter;
 	Window *win;
-	int	levels;
+	Mask	lastlog_levels;
 	int	line = 1;
 	size_t	rvclue = 0;
 
 	GET_STR_ARG(windesc, word);
 	GET_STR_ARG(pattern, word);
-	levels = parse_lastlog_level(word);
+	str_to_mask(&lastlog_levels, word);
 
 	/* Get the current window, default to current window */
 	if (!(win = get_window_by_desc(windesc)))
@@ -899,9 +981,10 @@ char *function_lastlog (char *word)
 
 	for (iter = win->lastlog_newest; iter; iter = iter->older, line++)
 	{
-		if (iter->level & levels)
+		if (iter->visible)
+		    if (mask_isset(&lastlog_levels, iter->level))
 			if (wild_match(pattern, iter->msg))
-				m_sc3cat(&retval, space, ltoa(line), &rvclue);
+				malloc_strcat_word_c(&retval, space, ltoa(line), &rvclue);
 	}
 
 	if (retval)
@@ -911,3 +994,131 @@ char *function_lastlog (char *word)
 }
 
 
+#if 0
+/*
+ * Here's the plan:
+ * 	$lastlogctl(REFNUMS <windesc> <levels>)
+ *	  Like $lastlog(), returns all lastlog refnums for <windesc> that
+ *	  belong to one of the level(s).  The refnums returned do not work
+ *	  with $line() (but work with $lastlogctl())
+ *
+ *	$lastlogctl(GET <refnum> LEVEL)
+ *	  The lastlog item's level, eg, MSGS, PUBLICS, OTHER.
+ *	$lastlogctl(GET <refnum> TARGET)
+ *	  The item's target: 
+ *		For messages sent to you, the sender.
+ *	  	For messages you sent, the receiver.  
+ *		For messages sent to a channel, the channel.
+ *		For everything else, the empty string.
+ *	$lastlogctl(GET <refnum> TEXT)
+ *	$lastlogctl(GET <refnum> TIMESTAMP)
+ *	$lastlogctl(GET <refnum> VISIBLE)
+ *	$lastlogctl(GET <refnum> WINDOW)
+ *
+ *	$lastlogctl(SET <refnum> LEVEL <level>)
+ *	$lastlogctl(SET <refnum> TARGET <target>)
+ *	$lastlogctl(SET <refnum> TEXT <text>)
+ *	$lastlogctl(SET <refnum> TIMESTAMP <timestamp>)
+ *	$lastlogctl(SET <refnum> VISIBLE [0|1])
+ *	$lastlogctl(SET <refnum> WINDOW)
+ */
+BUILT_IN_FUNCTION(function_lastlogctl, input)
+{
+	RETURN_STR(empty_string);
+}
+#endif
+
+/************************************************************************/
+#if 0
+int	oldest_lastlog_for_window (Lastlog **item, int winref)
+{
+	*item = NULL;
+	return newer_lastlog_entry(item, winref);
+}
+
+int	newer_lastlog_entry (Lastlog **item, int winref)
+{
+	Lastlog *i;
+
+	if (*item)
+		i = (*item)->newer;
+	else
+		i = lastlog_oldest;
+
+	while (i && i->winref != winref)
+		i = i->newer;
+	*item = i;
+
+	return i ? 1 : 0;
+}
+
+int	older_lastlog_entry (Lastlog **item, int winref)
+{
+	Lastlog *i;
+
+	if (*item)
+		i = (*item)->older;
+	else
+		i = lastlog_newest;
+
+	while (i && i->winref != winref)
+		i = i->older;
+	*item = i;
+
+	return i ? 1 : 0;
+}
+
+int	newest_lastlog_for_window (Lastlog **item, int winref)
+{
+	*item = NULL;
+	return older_lastlog_entry(item, winref);
+}
+
+int	remove_lastlog_item (Lastlog *item)
+{
+	if (item == lastlog_oldest)
+	{
+		if (item->older != NULL)
+			panic("Oldest lastlog item %jd has older item %jd",
+				item->refnum, item->older->refnum);
+
+		item->newer->older = NULL;
+		lastlog_oldest = item->newer;
+		item->newer = NULL;
+	}
+	else if (item == lastlog_newest)
+	{
+		if (item->newer != NULL)
+			panic("Newest lastlog item %jd has newer item %jd",
+				item->refnum, item->newer->refnum);
+
+		item->older->newer = NULL;
+		lastlog_newest = item->older;
+		item->older = NULL;
+	}
+	else
+	{
+		item->older->newer = item->newer;
+		item->newer->older = item->older;
+		item->newer = item->older = NULL;
+	}
+}
+
+int	remove_oldest_lastlog_for_window (int winref)
+{
+	Lastlog *item = NULL;
+
+	if (oldest_lastlog_for_window(&item, winref))
+		remove_lastlog_item(item);
+}
+
+int	switch_lastlog_window (Lastlog *item, int newref)
+{
+	/* Mark the old window's scrollback for reconstitution */
+	window_scrollback_needs_rebuild(item->winref);
+	item->winref = newref;
+	/* Mark the new window's scrollback for reconstitution */
+	window_scrollback_needs_rebuild(item->winref);
+}
+#endif
+
