@@ -1,4 +1,4 @@
-/* $EPIC: logfiles.c,v 1.23 2004/10/25 15:08:02 jnelson Exp $ */
+/* $EPIC: logfiles.c,v 1.33 2006/05/27 18:45:59 jnelson Exp $ */
 /*
  * logfiles.c - General purpose log files
  *
@@ -31,6 +31,7 @@
  * SUCH DAMAGE.
  */
 #include "irc.h"
+#include "levels.h"
 #include "log.h"
 #include "vars.h"
 #include "output.h"
@@ -61,7 +62,7 @@ struct Logfile {
 	int	type;
 	WNickList *targets;
 	int	refnums[MAX_TARGETS];
-	int	level;
+	Mask	mask;
 
 	char *	rewrite;
 	int	mangler;
@@ -77,11 +78,20 @@ int	logref = 0;
 
 static Logfile *	new_logfile (void)
 {
-	Logfile *log;
+	Logfile *log, *ptr;
 	int	i;
 
 	log = (Logfile *)new_malloc(sizeof(Logfile));
-	log->next = logfiles;
+
+	/* Move it to the end of the list. */
+	for (ptr = logfiles; ptr && ptr->next;)
+		ptr = ptr->next;
+	if (ptr)
+		ptr->next = log;
+	else
+		logfiles = log;
+
+	log->next = NULL;
 	log->refnum = ++logref;
 	log->name = malloc_sprintf(NULL, "%d", log->refnum);
 	log->filename = NULL;
@@ -91,14 +101,13 @@ static Logfile *	new_logfile (void)
 	log->targets = NULL;
 	for (i = 0; i < MAX_TARGETS; i++)
 		log->refnums[i] = -1;
-	log->level = LOG_ALL;
+	mask_setall(&log->mask);
 	log->rewrite = NULL;
 	log->mangler = 0;
 	log->mangle_desc = NULL;
 	log->active = 0;
 	time(&log->activity);
 
-	logfiles = log;
 	return log;
 }
 
@@ -315,7 +324,7 @@ static Logfile *	logfile_describe (Logfile *log, char **args)
 	say("\t         Type: %s", logtype[log->type]);
 	say("\t       Server: %d", log->servref);
 	say("\tTarget/Refnum: %s", targets ? targets : "<NONE>");
-	say("\t        Level: %s", bits_to_lastlog_level(log->level));
+	say("\t        Level: %s", mask_to_str(&log->mask));
 	say("\t Rewrite Rule: %s", log->rewrite ? log->rewrite : "<NONE>");
 	say("\t Mangle rules: %s", log->mangle_desc ? log->mangle_desc : "<NONE>");
 
@@ -374,7 +383,7 @@ static Logfile *	logfile_level (Logfile *log, char **args)
 		return NULL;
 	}
 
-	log->level = parse_lastlog_level(arg);
+	str_to_mask(&log->mask, arg);
 	return log;
 }
 
@@ -586,7 +595,7 @@ static Logfile *	logfile_server (Logfile *log, char **args)
 	if (!is_number(arg))
 		say("SERVER: The log's server needs to be a number");
 	else
-		log->servref = parse_server_index(arg, 1);
+		log->servref = str_to_servref(arg);
 
 	return log;
 }
@@ -682,7 +691,7 @@ BUILT_IN_COMMAND(logcmd)
 }
 
 /****************************************************************************/
-void	add_to_logs (int winref, int servref, const char *target, int level, const char *orig_str)
+void	add_to_logs (long winref, int servref, const char *target, int level, const char *orig_str)
 {
 	Logfile *log;
 	int	i;
@@ -693,10 +702,10 @@ void	add_to_logs (int winref, int servref, const char *target, int level, const
 	    {
 		for (i = 0; i < MAX_TARGETS; i++) {
 		    if (log->refnums[i] == winref) {
-			if (log->level && (log->level & level) == 0)
+			if (!mask_isset(&log->mask, level))
 				continue;
 			time(&log->activity);
-			add_to_log(log->log, winref, orig_str, log->mangler, log->rewrite);
+			add_to_log(log->refnum, log->log, winref, orig_str, log->mangler, log->rewrite);
 		    }
 		}
 	    }
@@ -705,10 +714,10 @@ void	add_to_logs (int winref, int servref, const char *target, int level, const
 	    {
 		for (i = 0; i < MAX_TARGETS; i++) {
 		    if (log->refnums[i] == servref) {
-			if (log->level && (log->level & level) == 0)
+			if (!mask_isset(&log->mask, level))
 				continue;
 			time(&log->activity);
-			add_to_log(log->log, winref, orig_str, log->mangler, log->rewrite);
+			add_to_log(log->refnum, log->log, winref, orig_str, log->mangler, log->rewrite);
 		    }
 		}
 	    }
@@ -718,7 +727,7 @@ void	add_to_logs (int winref, int servref, const char *target, int level, const
 		if (log->servref != NOSERV && (log->servref != servref))
 			continue;
 
-		if (log->level && (log->level & level) == 0)
+		if (!mask_isset(&log->mask, level))
 			continue;
 
 		if (log->targets && !target)
@@ -729,7 +738,7 @@ void	add_to_logs (int winref, int servref, const char *target, int level, const
 
 		/* OK!  We want to log it now! */
 		time(&log->activity);
-		add_to_log(log->log, winref, orig_str, log->mangler, log->rewrite);
+		add_to_log(log->refnum, log->log, winref, orig_str, log->mangler, log->rewrite);
 	    }
 	}
 }
@@ -822,7 +831,7 @@ char *logctl	(char *input)
 			char *ret = logfile_get_targets(log);
 			RETURN_MSTR(ret);
                 } else if (!my_strnicmp(listc, "LEVEL", 3)) {
-			char *ret = bits_to_lastlog_level(log->level);
+			char *ret = mask_to_str(&log->mask);
 			RETURN_STR(ret);
                 } else if (!my_strnicmp(listc, "REWRITE", 3)) {
 			RETURN_STR(log->rewrite);
@@ -881,9 +890,35 @@ char *logctl	(char *input)
                 RETURN_EMPTY;           /* Not implemented for now. */
         } else if (!my_strnicmp(listc, "PMATCH", 1)) {
                 RETURN_EMPTY;           /* Not implemented for now. */
+        } else if (!my_strnicmp(listc, "CURRENT", 1)) {
+		RETURN_INT(current_log_refnum);
         } else
                 RETURN_EMPTY;
 
         RETURN_EMPTY;
 }
 
+/*
+ * The /WINDOW NUMBER command actually swaps the refnums of two windows:
+ * It's possible that 'newref' isn't in use, so that's ok.
+ */
+void    logfiles_swap_winrefs (int oldref, int newref)
+{
+	Logfile *log;
+	int	i;
+
+	for (log = logfiles; log; log = log->next)
+        {
+		if (log->type != LOG_WINDOWS)
+			continue;
+
+		for (i = 0; i < MAX_TARGETS; i++)
+		{
+			if (log->refnums[i] == newref)
+				log->refnums[i] = oldref;
+			else if (log->refnums[i] == oldref)
+				log->refnums[i] = newref;
+		}
+        }
+}
+
