@@ -1,4 +1,4 @@
-/* $EPIC: mail.c,v 1.16 2014/03/31 13:57:22 jnelson Exp $ */
+/* $EPIC: mail.c,v 1.23 2006/06/24 15:54:25 jnelson Exp $ */
 /*
  * mail.c -- a gross simplification of mail checking.
  * Only unix maildrops (``mbox'') are supported.
@@ -86,7 +86,6 @@ static int	init_mbox_checking (void)
 	else
 	{
 		say("I can't find your mailbox.");
-		set_int_var(MAIL_VAR, 0);
 		return 0;
 	}
 
@@ -194,10 +193,9 @@ static void	update_mail_level1_mbox (void)
 			mail_latch++;
 			if (do_hook(MAIL_LIST, "You have new email"))
 			{
-				int lastlog_level = 
-				set_lastlog_msg_level(LOG_CRAP);
+				int l = message_from(NULL, LEVEL_OTHER);
 				say("You have new email.");
-				set_lastlog_msg_level(lastlog_level);
+				pop_message_from(l);
 			}
 			mail_latch--;
 		}
@@ -259,10 +257,9 @@ static void	update_mail_level2_mbox (void)
 			if (do_hook(MAIL_LIST, "%d %d", 
 					count - mail_last_count, count))
 			{
-			    int lastlog_level = 
-					set_lastlog_msg_level(LOG_CRAP);
+			    int l = message_from(NULL, LEVEL_OTHER);
 			    say("You have new email.");
-			    set_lastlog_msg_level(lastlog_level);
+			    pop_message_from(l);
 			}
 			mail_latch--;
 		    }
@@ -321,7 +318,6 @@ static int	init_maildir_checking (void)
 		{
 			say("Can't find your maildir -- Both your MAILDIR "
 				"and MAIL environment variables are unset.");
-			set_int_var(MAIL_VAR, 0);
 			return 0;
 		}
 	}
@@ -330,7 +326,6 @@ static int	init_maildir_checking (void)
 	{
 		say("The file in your %s environment variable "
 			"does not exist.", envvar);
-		set_int_var(MAIL_VAR, 0);
 		return 0;
 	}
 
@@ -338,7 +333,6 @@ static int	init_maildir_checking (void)
 	{
 		say("The file in your %s environment variable "
 			"is not a directory.", envvar);
-		set_int_var(MAIL_VAR, 0);
 		return 0;
 	}
 
@@ -349,11 +343,10 @@ static int	init_maildir_checking (void)
 		say("The directory in your %s environment variable "
 			"does not contain a sub-directory called 'new'", 
 				envvar);
-		set_int_var(MAIL_VAR, 0);
 		return 0;
 	}
 
-	maildir_path = malloc_strdup(maildir);
+	maildir_path = malloc_strdup(tmp_maildir_path);
 	maildir_last_changed = -1;
 	return 1;
 }
@@ -375,29 +368,13 @@ static int	maildir_count (void)
 {
 	int	count = 0;
 	DIR *	dir;
-	Filename 	tmp_maildir_path;
-	struct dirent*	dir_data;
 
-	strlcpy(tmp_maildir_path, maildir_path, sizeof(Filename));
-	strlcat(tmp_maildir_path, "/new", sizeof(Filename));
-	if ((dir = opendir(tmp_maildir_path)))
+	if ((dir = opendir(maildir_path)))
 	{
-		while ((dir_data = readdir(dir)) != NULL) {
-			if (dir_data->d_name[0] != '.')
-				count++;
-		}
-		closedir(dir);
-	}
-
-	strlcpy(tmp_maildir_path, maildir_path, sizeof(Filename));
-	strlcat(tmp_maildir_path, "/cur", sizeof(Filename));
-	if ((dir = opendir(tmp_maildir_path)))
-	{
-		while ((dir_data = readdir(dir)) != NULL) {
-			if (dir_data->d_name[0] != '.')
-				count++;
-		}
+		while (readdir(dir) != NULL)
+			count++;
 		closedir(dir);
+		count -= 2;	/* Don't count . or .. */
 	}
 
 	return count;
@@ -414,7 +391,6 @@ static int	poll_maildir_status (void *ptr)
 {
 	Stat	sb;
 	Stat *	stat_buf;
-	Filename 	tmp_maildir_path;
 
 	if (ptr)
 		stat_buf = (Stat *)ptr;
@@ -425,11 +401,8 @@ static int	poll_maildir_status (void *ptr)
 		if (!init_maildir_checking())
 			return 0;		/* Can't find maildir */
 
-	strlcpy(tmp_maildir_path, maildir_path, sizeof(Filename));
-	strlcat(tmp_maildir_path, "/new", sizeof(Filename));
-
 	/* If there is no mailbox, there is no mail! */
-	if (stat(tmp_maildir_path, stat_buf) == -1)
+	if (stat(maildir_path, stat_buf) == -1)
 		return 0;
 
 	/* 
@@ -477,10 +450,9 @@ static void	update_mail_level1_maildir (void)
 				mail_latch++;
 				if (do_hook(MAIL_LIST, "You have new email"))
 				{
-					int lastlog_level = 
-						set_lastlog_msg_level(LOG_CRAP);
-					say("You have new email.");
-					set_lastlog_msg_level(lastlog_level);
+				    int l = message_from(NULL, LEVEL_OTHER);
+				    say("You have new email.");
+				    pop_message_from(l);
 				}
 				mail_latch--;
 			}
@@ -541,10 +513,9 @@ static void	update_mail_level2_maildir (void)
 			if (do_hook(MAIL_LIST, "%d %d", 
 					count - mail_last_count, count))
 			{
-			    int lastlog_level = 
-					set_lastlog_msg_level(LOG_CRAP);
+			    int l = message_from(NULL, LEVEL_OTHER);
 			    say("You have new email.");
-			    set_lastlog_msg_level(lastlog_level);
+			    pop_message_from(l);
 			}
 			mail_latch--;
 		    }
@@ -567,10 +538,6 @@ static void	update_mail_level3_maildir (void)
 	update_mail_level2_maildir();
 	if (status == 2)
 	{
-		Filename 	tmp_maildir_path;
-		strlcpy(tmp_maildir_path, maildir_path, sizeof(Filename));
-		strlcat(tmp_maildir_path, "/new", sizeof(Filename));
-
 		/* XXX Ew.  Evil. Gross. */
 		ts.actime = stat_buf.st_atime;
 		ts.modtime = stat_buf.st_mtime;
@@ -595,14 +562,14 @@ struct mail_checker mail_types[] = {
 			update_mail_level2_mbox,
 			update_mail_level3_mbox,
 			deinit_mbox_checking },
-	{ "MAILDIR",    init_maildir_checking,
+	{ "MAILDIR",	init_maildir_checking, 
 			update_mail_level1_maildir,
 			update_mail_level2_maildir,
 			update_mail_level3_maildir,
 			deinit_maildir_checking },
 	{ NULL,		NULL, NULL, NULL, NULL, NULL }
 };
-struct mail_checker *checkmail = &mail_types[MAIL_CHECKING_TYPE];
+struct mail_checker *checkmail = NULL;
 
 /**************************************************************************/
 const char *	check_mail (void)
@@ -616,6 +583,9 @@ void	mail_systimer (void)
 {
 	int	x;
 
+	if (checkmail == NULL)
+		return;		/* Whatever */
+
 	switch ((x = get_int_var(MAIL_VAR)))
 	{
 		case 1:
@@ -636,27 +606,38 @@ void	mail_systimer (void)
 	return;
 }
 
-void    set_mail_interval (const void *stuff)
+void    set_mail_interval (void *stuff)
 {
 	update_system_timer(mail_timeref);
 }
 
-void	set_mail (const void *stuff)
+void	set_mail (void *stuff)
 {
-	int	value = *(const int *)stuff;
+	VARIABLE *v;
+	int	value;
+
+	v = (VARIABLE *)stuff;
+	value = v->integer;
 
 	if (value < 0 || value > 3)
 	{
 		say("/SET MAIL must be 0, 1, 2, or 3");
-		set_int_var(MAIL_VAR, 0);
-		return;
+		v->integer = value = 0;
 	}
-	else if (value == 0)
-		checkmail->deinit();
-	else
+	if (value != 0)
 	{
-		if (!checkmail->init())
+		if (checkmail == NULL)
 			return;
+		if (!checkmail->init())
+			value = v->integer = 0;
+	}
+	if (value == 0)
+	{
+		if (checkmail == NULL)
+			return;
+		checkmail->deinit();
+		mail_last_count = -1;
+		mail_last_count_str = NULL;
 	}
 
 	update_system_timer(mail_timeref);
@@ -664,13 +645,15 @@ void	set_mail (const void *stuff)
 	cursor_to_input();
 }
 
-void	set_mail_type (const void *stuff)
+void	set_mail_type (void *stuff)
 {
+	VARIABLE *v;
 	const char *	value;
 	struct mail_checker *new_checker;
 	char	old_mailval[16];
 
-	value = (const char *)stuff;
+	v = (VARIABLE *)stuff;
+	value = v->string;
 
 	if (value == NULL)
 		new_checker = NULL;
@@ -685,8 +668,8 @@ void	set_mail_type (const void *stuff)
 	}
 
 	snprintf(old_mailval, sizeof(old_mailval), "%d", get_int_var(MAIL_VAR));
-	set_var_value(MAIL_VAR, zero);
+	set_var_value(MAIL_VAR, zero, 0);
 	checkmail = new_checker;
-	set_var_value(MAIL_VAR, old_mailval);
+	set_var_value(MAIL_VAR, old_mailval, 0);
 }
 
