@@ -1,4 +1,4 @@
-/* $EPIC: names.c,v 1.46 2003/10/28 05:53:57 jnelson Exp $ */
+/* $EPIC: names.c,v 1.74 2006/05/27 18:45:59 jnelson Exp $ */
 /*
  * names.c: This here is used to maintain a list of all the people currently
  * on your channel.  Seems to work 
@@ -46,6 +46,7 @@
 #include "server.h"
 #include "list.h"
 #include "hook.h"
+#include "parse.h"
 
 typedef struct nick_stru
 {
@@ -78,59 +79,28 @@ struct	channel_stru *	prev;		/* pointer to previous channel */
 	int		server;		/* The server the channel is "on" */
 	int		winref;		/* The window the channel is "on" */
 	int		curr_count;	/* Current channel precedence */
-	int		waiting;	/* just acting as a placeholder... */
-	int		inactive;	/* waiting on a reconnect... */
-	int		bound;		/* Bound to this window? */
-	int		current;	/* Current to this window? */
+	int		waiting;	/* Syncing, waiting for names/who */
 	NickList	nicks;		/* alist of nicks on channel */
 
-	unsigned long	mode;		/* Current mode settings for channel */
-	unsigned long	i_mode;		/* channel mode for cached string */
-	char *		s_mode;		/* cached string version of modes */
+	char 		base_modes[54];	/* Just the modes w/o args */
 	int		limit;		/* max users for the channel */
 	char *		key;		/* key for this channel */
-	char		chop;		/* true if you are chop */
-	char		voice;		/* true if you are voiced */
-	char		half_assed;	/* true if you are a helper */
+	char *		modestr;	/* The whole mode string */
+
+	char		chop;		/* true if i'm a channel operator */
+	char		voice;		/* true if i'm a channel voice */
+	char		half_assed;	/* true if i'm a channel helper */
 	Timeval		join_time;	/* When we joined the channel */
 }	Channel;
 
 
-/*
- * The variable "defined_mode_str" must correspond in order to the modes 
- * defined here, or all heck will break loose.  You have been warned.
- */
-static	char	defined_mode_str[] = "aciklmnprstzDMOR";
-
-const int	MODE_ANONYMOUS	= 1 << 0;	/* av2.9 */
-const int	MODE_C		= 1 << 1;	/* erf/TS4 */
-const int 	MODE_INVITE 	= 1 << 2;	/* RFC */
-const int 	MODE_KEY    	= 1 << 3;	/* RFC */
-const int	MODE_LIMIT	= 1 << 4;	/* RFC */
-const int 	MODE_MODERATED	= 1 << 5;	/* RFC */
-const int	MODE_MSGS	= 1 << 6;	/* RFC */
-const int	MODE_PRIVATE	= 1 << 7;	/* RFC */
-const int	MODE_REGISTERED = 1 << 8;	/* Duhnet */
-const int	MODE_SECRET	= 1 << 9;	/* RFC */
-const int	MODE_TOPIC	= 1 << 10;	/* RFC */
-const int	MODE_Z		= 1 << 11;	/* erf/TS4 */
-const int	MODE_D		= 1 << 12;	/* special request */
-const int	MODE_M          = 1 << 13;	/* Duhnet */
-const int	MODE_OPER_ONLY	= 1 << 14;	/* Duhnet */
-const int	MODE_RESTRICTED = 1 << 15;	/* Duhnet */
-const int	MODE_C2		= 1 << 16;	/* Quakenet */
-
-
 /* channel_list: list of all the channels you are currently on */
 static	Channel *	channel_list = NULL;
 
 #if 0
-/* For new eu2.9 !channels */
-static	char		new_channel_format[BIG_BUFFER_SIZE];
-#else
 static	int	match_chan_with_id (const char *chan, const char *match);
 #endif
-	void	channel_hold_election (int winref);
+static	void	channel_hold_election (int winref);
 
 
 /*
@@ -178,8 +148,8 @@ static Channel *find_channel (const char *channel, int server)
 			return NULL;		/* sb colten */
 
 	while (traverse_all_channels(&ch, server, 1))
-		if (!my_stricmp(ch->channel, channel))
-			return ch;
+	    if (!server_stricmp(ch->channel, channel, server))
+		return ch;
 
 	return NULL;
 }
@@ -193,18 +163,17 @@ static Channel *create_channel (const char *name, int server)
 	new_c->channel = malloc_strdup(name);
 	new_c->server = server;
 	new_c->waiting = 0;
-	new_c->inactive = 0;
 	new_c->winref = -1;
-	new_c->bound = 0;
-	new_c->current = 0;
 	new_c->nicks.max_alloc = new_c->nicks.max = 0;
 	new_c->nicks.list = NULL;
-	new_c->nicks.func = (alist_func) my_strnicmp;
+	if (get_server_stricmp_table(server) == 0)
+		new_c->nicks.func = (alist_func) ascii_strnicmp;
+	else
+		new_c->nicks.func = (alist_func) rfc1459_strnicmp;
 	new_c->nicks.hash = HASH_INSENSITIVE;
 
-	new_c->mode = 0;
-	new_c->i_mode = 0;
-	new_c->s_mode = NULL;
+	new_c->base_modes[0] = 0;
+	new_c->modestr = NULL;
 	new_c->limit = 0;
 	new_c->key = NULL;
 	new_c->chop = 0;
@@ -280,9 +249,7 @@ static void 	destroy_channel (Channel *chan)
 	if (chan->nicks.max_alloc)
 		clear_channel(chan);
 
-	chan->mode = 0;
-	chan->i_mode = 0;
-	new_free(&chan->s_mode);
+	new_free(&chan->modestr);
 	chan->limit = 0;
 	new_free(&chan->key); 
 	chan->chop = 0;
@@ -297,7 +264,6 @@ void 	add_channel (const char *name, int server)
 {
 	Channel *new_c;
 	int	was_window = -1;
-	Window	*tmp = NULL;
 
 	if ((new_c = find_channel(name, server)))
 	{
@@ -309,36 +275,11 @@ void 	add_channel (const char *name, int server)
 	else
 		new_c = create_channel(name, server);
 
-	new_c->inactive = 0;		/* Channel no longer is a placeholder */
 	new_c->waiting = 1;		/* This channel is "syncing" */
 	get_time(&new_c->join_time);
 
 	if (was_window == -1)
-	{
-	    /* Try to find "our" window for this channel. */
-	    while (traverse_all_windows(&tmp))
-	    {
-		if (tmp->server != from_server)
-			continue;
-		if (!tmp->waiting_channel && !tmp->bind_channel)
-			 continue;
-
-		if (tmp->bind_channel && 
-			!match_chan_with_id(name, tmp->bind_channel))
-		{
-			was_window = tmp->refnum;
-			break;
-		}
-
-		if (tmp->waiting_channel &&
-			!match_chan_with_id(name, tmp->waiting_channel))
-		{
-			new_free(&tmp->waiting_channel);
-			was_window = tmp->refnum;
-			break;
-		}
-	    }
-	}
+		was_window = claim_waiting_channel(name, server);
 
 	if (was_window == -1)
 		was_window = get_winref_by_servref(from_server);
@@ -619,7 +560,7 @@ void 	remove_from_channel (const char *channel, const char *nick, int server)
 	while (traverse_all_channels(&chan, server, 1))
 	{
 		/* This is correct, dont change it! */
-		if (channel && my_stricmp(channel, chan->channel))
+		if (channel && server_stricmp(channel, chan->channel, server))
 			continue;
 
 		if ((tmp = (Nick *)remove_from_array((array *)&chan->nicks, nick)))
@@ -648,6 +589,7 @@ void 	rename_nick (const char *old_nick, const char *new_nick, int server)
 		if ((tmp = (Nick *)remove_from_array((array *)&chan->nicks, old_nick)))
 		{
 			malloc_strcpy(&tmp->nick, new_nick);
+			malloc_strcpy(&tmp->userhost, FromUserHost);
 			add_to_array((array *)&chan->nicks, (array_item *)tmp);
 		}
 	}
@@ -663,14 +605,6 @@ const char *	check_channel_type (const char *channel)
 {
 	/* Grumblesmurf */
 	return channel;
-#if 0
-	if (*channel != '!' || strlen(channel) < 6)
-		return channel;
-
-	snprintf(new_channel_format, sizeof new_channel_format,
-			"[%.6s] %s", channel, channel + 6);
-	return new_channel_format;
-#endif
 }
 
 int 	im_on_channel (const char *channel, int refnum)
@@ -790,40 +724,9 @@ char	*create_nochops_list (const char *name, int server)
  * get_cmode: Get the current mode string for the specified server.
  * We only do a refresh if the mode has changed since we generated it.
  */
-static char *	get_cmode (Channel *chan)
+static const char *	get_cmode (Channel *chan)
 {
-	int	mode_pos = 0,
-		str_pos = 0;
-	char	local_buffer[BIG_BUFFER_SIZE];
-
-	/* Used the cache value if its still good */
-	if ((chan->mode == chan->i_mode) && chan->s_mode)
-		return chan->s_mode;
-
-	chan->i_mode = chan->mode;
-
-	local_buffer[0] = 0;
-	while (chan->mode >= (1UL << mode_pos))
-	{
-		if (chan->mode & (1UL << mode_pos))
-			local_buffer[str_pos++] = defined_mode_str[mode_pos];
-		mode_pos++;
-	}
-	local_buffer[str_pos] = 0;
-
-	if (chan->key)
-	{
-		strlcat(local_buffer, " ", sizeof local_buffer);
-		strlcat(local_buffer, chan->key, sizeof local_buffer);
-	}
-	if (chan->limit)
-	{
-		strlcat(local_buffer, " ", sizeof local_buffer);
-		strlcat(local_buffer, ltoa(chan->limit), sizeof local_buffer);
-	}
-
-	malloc_strcpy(&chan->s_mode, local_buffer);
-	return (chan->s_mode);
+	return chan->modestr;
 }
 
 /*
@@ -882,8 +785,9 @@ static void	decifer_mode (const char *modes, Channel *chan)
 	int	add = 0;
 	char *	rest;
 	Nick *	nick;
-	int	value = 0;
 	char *	mode_str;
+	char	local_buffer[BIG_BUFFER_SIZE];
+	int	type;
 
 	/* Make a copy of it.*/
 	mode_str = LOCAL_COPY(modes);
@@ -892,11 +796,14 @@ static void	decifer_mode (const char *modes, Channel *chan)
 	if (!(mode_str = next_arg(mode_str, &rest)))
 		return;
 
+	/* Now walk through and look for special stuff */
 	for (; *mode_str; mode_str++)
 	{
 	    const char *arg = NULL;
 
-	    switch (chanmodetype(*mode_str))
+	    /* Grab an argument if this mode takes one */
+	    type = chanmodetype(*mode_str);
+	    switch (type)
 	    {
 		case 6: case 1:
 			break;
@@ -911,6 +818,7 @@ static void	decifer_mode (const char *modes, Channel *chan)
 			yell("Defaulting %c%c to CHANMODE type D", add?'+':'-', *mode_str);
 	    }
 
+	    /* Now look for modes that change state that we track */
 	    switch (*mode_str)
 	    {
 		case '+':
@@ -920,51 +828,6 @@ static void	decifer_mode (const char *modes, Channel *chan)
 			add = 0;
 			continue;
 
-		case 'a':
-			value = MODE_ANONYMOUS;
-			break;
-		case 'c':
-			value = MODE_C;
-			break;
-		case 'C':
-			value = MODE_C2;
-			break;
-		case 'D':
-			value = MODE_D;
-			break;
-		case 'i':
-			value = MODE_INVITE;
-			break;
-		case 'm':
-			value = MODE_MODERATED;
-			break;
-		case 'n':
-			value = MODE_MSGS;
-			break;
-		case 'p':
-			value = MODE_PRIVATE;
-			break;
-		case 'r':
-			value = MODE_REGISTERED;
-			break;
-		case 'R':
-			value = MODE_RESTRICTED;
-			break;
-		case 's':
-			value = MODE_SECRET;
-			break;
-		case 't':
-			value = MODE_TOPIC;
-			break;
-		case 'z':		/* Erf/TS4 "zapped" */
-			value = MODE_Z;
-			break;
-		case 'M':		/* Duhnet's mute-mode */
-			value = MODE_M;
-			break;
-		case 'O':		/* Duhhnet's oper-only */
-			value = MODE_OPER_ONLY;
-			break;
 		case 'k':
 		{
 			if (!arg)
@@ -975,24 +838,31 @@ static void	decifer_mode (const char *modes, Channel *chan)
 			    continue;
 			}
 
-			value = MODE_KEY;
-
 			if (add)
+			{
 				malloc_strcpy(&chan->key, arg);
+				add_mode_to_str(chan->base_modes, 54, *mode_str);
+			}
 			else
+			{
 				new_free(&chan->key);
+				remove_mode_from_str(chan->base_modes, 54, *mode_str);
+			}
 
-			chan->i_mode = -1;	/* Revoke old cache */
 			break;	
 		}
+
 		case 'l':
 		{
-			value = MODE_LIMIT;
 			if (!add)
 				arg = zero;
 
 			chan->limit = my_atol(arg);
-			chan->i_mode = -1;	/* Revoke old cache */
+			if (chan->limit)
+				add_mode_to_str(chan->base_modes, 54, *mode_str);
+			else
+				remove_mode_from_str(chan->base_modes, 54, *mode_str);
+
 			continue;
 		}
 
@@ -1046,26 +916,34 @@ static void	decifer_mode (const char *modes, Channel *chan)
 				nick->half_assed = add;
 			continue;
 		}
-		case 'b':
-		case 'e':	/* borked erfnet ban exceptions */
-		case 'I':	/* borked ircnet invite exceptions */
+
+		default:
 		{
-			continue;
+		    if (type == 2 || type == 3 || type == 4)
+			continue;	/* Skip modes with args */
+
+		    if (add)
+			add_mode_to_str(chan->base_modes, 54, *mode_str);
+		    else
+			remove_mode_from_str(chan->base_modes, 54, *mode_str);
 		}
 	    }
-
-	    if (add)
-		chan->mode |= value;
-	    else
-		chan->mode &= ~value;
 	}
-	if (rest && *rest)
-		yell("WARNING:  Mode parser or server is BROKE.  Remaining args: %s", rest);
+	if (!is_string_empty(rest))
+		yell("WARNING:  Extra unhandled arguments found in MODE: %s", rest);
 
-	if (!chan->limit)
-		chan->mode &= ~MODE_LIMIT;
-	else
-		chan->mode |= MODE_LIMIT;
+	strlcpy(local_buffer, chan->base_modes, sizeof local_buffer);
+	if (chan->key)
+	{
+		strlcat(local_buffer, " ", sizeof local_buffer);
+		strlcat(local_buffer, chan->key, sizeof local_buffer);
+	}
+	if (chan->limit)
+	{
+		strlcat(local_buffer, " ", sizeof local_buffer);
+		strlcat(local_buffer, ltoa(chan->limit), sizeof local_buffer);
+	}
+	malloc_strcpy(&chan->modestr, local_buffer);
 }
 
 /* XXX Probably doesnt belong here. im tired, though */
@@ -1082,16 +960,17 @@ int	channel_is_syncing (const char *channel, int server)
 void	channel_not_waiting (const char *channel, int server)
 {
 	Channel *tmp = find_channel(channel, server);
+	int	l;
 
 	if (tmp)
 	{
 		tmp->waiting = 0;
-		message_from(channel, LOG_CRAP);
+		l = message_from(channel, LEVEL_OTHER);
 		do_hook(CHANNEL_SYNC_LIST, "%s %f %d",
 			tmp->channel, 
 			time_diff(tmp->join_time, get_time(NULL)),
 			tmp->server);
-		message_from(NULL, LOG_CRAP);
+		pop_message_from(l);
 	}
 	else
 		yell("Channel_sync -- didn't find [%s:%d]",
@@ -1118,7 +997,7 @@ const char 	*get_channel_key (const char *channel, int server)
 }
 
 
-char	*get_channel_mode (const char *channel, int server)
+const char	*get_channel_mode (const char *channel, int server)
 {
 	Channel *tmp = find_channel(channel, server);
 
@@ -1138,9 +1017,8 @@ int 	is_channel_private (const char *channel, int server_index)
 	Channel *tmp = find_channel(channel, server_index);
 
 	if (tmp)
-		return (tmp->mode & (MODE_PRIVATE | MODE_SECRET));
-	else
-		return 0;
+		return (sindex(tmp->base_modes, "ps") ? 1 : 0);
+	return 0;
 }
 
 int	is_channel_nomsgs (const char *channel, int server_index)
@@ -1148,7 +1026,17 @@ int	is_channel_nomsgs (const char *channel, int server_index)
 	Channel *tmp = find_channel(channel, server_index);
 
 	if (tmp)
-		return (tmp->mode & (MODE_MSGS));
+		return (strchr(tmp->base_modes, 'n') ? 1 : 0);
+	else
+		return 0;
+}
+
+int	is_channel_anonymous (const char *channel, int server_index)
+{
+	Channel *tmp = find_channel(channel, server_index);
+
+	if (tmp)
+		return (strchr(tmp->base_modes, 'a') ? 1 : 0);
 	else
 		return 0;
 }
@@ -1271,7 +1159,7 @@ void 	list_channels (void)
 
 
 /* This is a keybinding */
-void 	switch_channels (char dumb, char *dumber)
+BUILT_IN_KEYBINDING(switch_channels)
 {
 	int	lowcount = current_channel_counter;
 	Char *	winner = NULL;
@@ -1337,8 +1225,6 @@ static Channel *window_current_channel_internal (int window, int server)
         {
                 if (tmp->winref != window)
                         continue;
-                if (tmp->inactive)
-                        continue;
                 if (tmp->curr_count > maxcount)
                 {
                         maxcount = tmp->curr_count;
@@ -1364,19 +1250,19 @@ const char *  window_current_channel (int window, int server)
         return tmp->channel;
 }
 
-void 	change_server_channels (int old_s, int new_s)
+char *	window_all_channels (int window, int server)
 {
+	char *str = NULL;
 	Channel *tmp = NULL;
+	size_t	clue = 0;
 
-	if (old_s < 0) return;		/* Sanity check */
-
-	while (traverse_all_channels(&tmp, old_s, 1))
+	while (traverse_all_channels(&tmp, server, 1))
 	{
-		tmp->server = new_s;
-		tmp->inactive = 1;
-		/* Is there any case where this is bad? */
-		clear_channel(tmp);	
+		if (tmp->winref != window)
+			continue;
+		malloc_strcat_word_c(&str, space, tmp->channel, &clue);
 	}
+	return str;
 }
 
 int     is_current_channel (const char *channel, int server)
@@ -1386,16 +1272,14 @@ int     is_current_channel (const char *channel, int server)
  
         if ((window = get_channel_winref(channel, server)) > 0)
                 if ((name = window_current_channel(window, server)))
-                        if (!my_stricmp(name, channel)) 
+                        if (!server_stricmp(name, channel, server)) 
                                 return 1;
         return 0; 
 }
 
 /*
- * This is called by connect_to_new_server(), if the new server we are going
- * to attach to is already an established connection; by close_server(), if
- * it was asked to have the server's channels thrown away; by 
- * reconnect_all_channels() after reconnect and the JOINs have been sent off.
+ * This is called by close_server(), when a server connection is being 
+ * abandoned.  Channels are never retained across reconnections.
  */
 void 	destroy_server_channels (int server)
 {
@@ -1426,72 +1310,6 @@ void 	destroy_server_channels (int server)
 }
 
 
-/*
- * reconnect_all_channels:  the main auto-rejoin-on-reconnect function.
- * Called after you connect to a server as part of the "continuing action"
- * of a disconnect handling event.  It gloms up the channel names on the
- * list for the current server (which at this point are just acting as 
- * placeholders), and makes sure that all the windows are all happy with 
- * their current channels, and then spits out a JOIN for all of the 
- * channels, and then trashcans the channels.
- *
- * This probably never works quite like people expect.
- */
-void 	reconnect_all_channels (void)
-{
-	Channel *tmp = NULL;
-	char	*channels = NULL;
-	char	*keyed_channels = NULL;
-	char	*keys = NULL;
-	size_t	chan_clue = 0, kc_clue = 0, key_clue = 0;
-
-	/* Oh, what the heck. */
-	if (!get_int_var(AUTO_REJOIN_CONNECT_VAR)) {
-		destroy_server_channels(from_server);
-		return;
-	}
-
-	while (traverse_all_channels(&tmp, from_server, 1))
-	{
-		if (!tmp->inactive)
-			yell("Ack.  Reconnecting channel [%s] on server [%d] "
-			     "but it isn't inactive!", 
-				tmp->channel, tmp->server);
-
-		if (tmp->key)
-		{
-			malloc_strcat_wordlist_c(&keyed_channels, ",", tmp->channel, &kc_clue);
-			malloc_strcat_wordlist_c(&keys, ",", tmp->key, &key_clue);
-		}
-		else
-			malloc_strcat_wordlist_c(&channels, ",", tmp->channel, &chan_clue);
-
-		clear_channel(tmp);
-	}
-
-	/* This is probably useless, but its harmless. */
-	save_channels(from_server);
-
-	/*
-	 * Interestingly enough, black magic on the server's part makes
-	 * this work.  I sure hope they dont "break" this in the future...
-	 */
-	if (keyed_channels)
-		send_to_server("JOIN %s %s", keyed_channels, keys);
-	if (channels)
-		send_to_server("JOIN %s", channels);
-
-	new_free(&channels);
-	new_free(&keyed_channels);
-	new_free(&keys);
-
-	/*
-	 * I wish i didnt have to do this... :/
-	 */
-	destroy_server_channels(from_server);
-}
-
-
 const char *	what_channel (const char *nick)
 {
 	Channel *tmp = NULL;
@@ -1521,19 +1339,20 @@ const char *	walk_channels (int init, const char *nick)
 	return NULL;
 }
 
-const char *	fetch_userhost (int server, const char *nick)
+const char *	fetch_userhost (int server, const char *chan, const char *nick)
 {
 	Channel *tmp = NULL;
 	Nick *user = NULL;
 
 	if (server == NOSERV) return NULL;		/* Sanity check */
 
-	while (traverse_all_channels(&tmp, server, 1))
-	{
+	if (chan && (tmp = find_channel(chan, server)) &&
+			(user = find_nick_on_channel(tmp, nick)))
+		return user->userhost;
+	else while (traverse_all_channels(&tmp, server, 1))
 		if ((user = find_nick_on_channel(tmp, nick)) && 
 				user->userhost)
 			return user->userhost;
-	}
 
 	return NULL;
 }
@@ -1659,7 +1478,6 @@ void 	set_channel_window (const char *channel, int server, int winref, int as_cu
 void   move_channel_to_window (const char *chan, int server, int old_w, int new_w)
 {
 	Channel *tmp = NULL;
-	const char *x;
 
 	if (chan == NULL || strcmp(chan, zero) == 0)
 		return;
@@ -1676,27 +1494,12 @@ void   move_channel_to_window (const char *chan, int server, int old_w, int new_
 		panic("Channel [%s:%d] is on window [%d] not on window [%d] (moving to [%d])",
 			chan, server, tmp->winref, old_w, new_w);
 
-	/* 
-	 * Unbind this channel unconditionally unless the new window has
-	 * already bound it.
-	 */
-	x = get_bound_channel_by_refnum(new_w);
-	if (x && my_stricmp(chan, x))
-		unbind_channel(chan, tmp->server);
-
-	/*
-	 * If the old window is waiting for this channel, make the new
-	 * window wait for this channel.
-	 */
-	if (is_window_waiting_for_channel(old_w, chan))
-		move_waiting_channel(old_w, new_w);
-
 	/* set_channel_window calls new elections */
 	set_channel_window(chan, server, new_w, 1);
 }
 
 
-void	channel_hold_election (int winref)
+static void	channel_hold_election (int winref)
 {
 	if (winref <= 0)
 		return;			/* Whatever.  Probably should panic */
@@ -1716,12 +1519,10 @@ void	reassign_window_channels (int window)
 	Window *w = NULL;
 	int	caution = 0;
 	int	winserv;
-	int	winoldserv;
 
 	winserv = get_window_server(window);
-	winoldserv = get_window_oldserver(window);
 
-	if (winserv == NOSERV && winoldserv == NOSERV)
+	if (winserv == NOSERV)
 		caution = 1;
 
 	for (tmp = channel_list; tmp; tmp = tmp->next)
@@ -1736,7 +1537,7 @@ void	reassign_window_channels (int window)
 			caution = 0;
 		}
 
-		if (tmp->server != winserv && tmp->server != winoldserv)
+		if (tmp->server != winserv)
 		{
 			yell("Channel [%s:%d] is connected to window [%d] "
 				"which is apparantly on another server", 
@@ -1776,82 +1577,15 @@ char *	create_channel_list (int server)
 	return retval ? retval : malloc_strdup(empty_string);
 }
 
-/*
- * This is (supposed to be) only ever called by close_server, and is called
- * whenever it was requested that the channels on a closing server not be
- * thrown away, but instead converted into placeholders.  So that's what we
- * do here.
- */
-void 	save_channels (int servref)
-{
-	Window	*tmp = NULL;
-	Channel *tmpc = NULL;
-	const char *chan;
-
-	/*
-	 * Go through all of the windows for this server
-	 * and mark their current channels as WAITING.
-	 */
-	while (traverse_all_windows(&tmp))
-	{
-		if (tmp->server != servref)
-			continue;
-
-		if (tmp->waiting_channel)
-			continue;		/* Yea yea yea */
-
-		if ((chan = get_echannel_by_refnum(tmp->refnum)))
-			tmp->waiting_channel = malloc_strdup(chan);
-	}
-
-	/*
-	 * Go through all the channels for this server
-	 * and mark them as WAITING.
-	 */
-	while (traverse_all_channels(&tmpc, servref, 1))
-		tmpc->inactive = 1;
-}
-
 /* I don't know if this belongs here. */
 void	cant_join_channel (const char *channel, int server)
 {
-	Window *w = NULL;
-	while (traverse_all_windows(&w))
-	{
-		if (w->server != server)
-			continue;
-		if (!w->waiting_channel)
-			continue;
-		if (!my_stricmp(w->waiting_channel, channel))
-			new_free(&w->waiting_channel);
-	}
+	/* Eh, just throw the channel away, eh? */
+	claim_waiting_channel(channel, server);
 	update_all_windows();
 }
 
-/**************************************************************************/
-/* 
- * Im sure this doesnt belong here, but im not sure where it does belong.
- */
-int 	auto_rejoin_callback (void *d)
-{
-	char *	data    = (char *) d;
-	char *	channel	= next_arg(data, &data);
-	int 	server	= parse_server_index(next_arg(data, &data), 0);
-	Window *window 	= get_window_by_refnum(my_atol(next_arg(data, &data)));
-	char *	key    	= next_arg(data, &data);
-
-	if (key && *key)
-		send_to_aserver(server, "JOIN %s %s", channel, key);
-	else
-		send_to_aserver(server, "JOIN %s", channel);
-
-	if (window)
-		malloc_strcpy(&window->waiting_channel, channel);
-	new_free((char **)&d);
-
-	return 0;
-}
-
+#if 0
 /*
  * match_chan_with_id: check if the channel matches.  It checks also takes
  * the ID into account for !channels
@@ -1884,7 +1618,7 @@ static int	match_chan_with_id (const char *chan, const char *match)
 
 	return my_stricmp(chan, match);
 }
-
+#endif
 
 /*
  *
@@ -1940,11 +1674,11 @@ void	channel_check_windows (void)
 		if (tmp->winref <= 0)
 			panic("I thought we just checked for this! [1]");
 
-		if (get_window_server(tmp->winref) == NOSERV && 
-			     tmp->server == get_window_oldserver(tmp->winref))
+		if (get_window_server(tmp->winref) == NOSERV)
 			continue;			/* This is OK. */
 
 		if (tmp->server != get_window_server(tmp->winref))
+		    if (get_server_status(tmp->server) != SERVER_CLOSING)
 			panic("Referential integrity failure: "
 			      "Channel [%s] on server [%d] is connected "
 			      "to window [%d] on server [%d]",
@@ -1960,6 +1694,7 @@ void	channel_check_windows (void)
 			panic("I thought we just checked for this! [2]");
 
 		if (!get_window_by_refnum(tmp->winref))
+		    if (get_server_status(tmp->server) != SERVER_CLOSING)
 			panic("Referential integrity failure: "
 			      "Channel [%s] on server [%d] is connected "
 			      "to window [%d] "
@@ -1968,25 +1703,26 @@ void	channel_check_windows (void)
 				tmp->winref);
 	}
 
-	/* Do test #7 -- check for abandoned channels */
-	for (tmp = channel_list; tmp;
-		reset ? (tmp = channel_list) : (tmp = tmp->next))
-	{
-		if (tmp->winref <= 0)
-			panic("I thought we just checked for this! [4]");
-
-		if (did_server_rejoin_channels(tmp->server) && tmp->inactive)
-			panic("Channel [%s] on server [%d] is inactive "
-				"even though this server is connected!",
-				tmp->channel, tmp->server);
-
-		if (is_server_registered(tmp->server) &&
-				!did_server_rejoin_channels(tmp->server) && 
-				!tmp->inactive)
-			panic("Channel [%s] on server [%d] is NOT inactive "
-				"even though this server is NOT connected!",
-				tmp->channel, tmp->server);
-	}
-
 	return;
 }
+
+/*
+ * The /WINDOW NUMBER command actually swaps the refnums of two windows:
+ * It's possible that 'newref' isn't in use, so that's ok.
+ */
+void	channels_swap_winrefs (int oldref, int newref)
+{
+	Channel	*tmp = NULL;
+	Window *w = NULL;
+	int	reset = 0;
+
+	for (tmp = channel_list; tmp; 
+		reset ? (tmp = channel_list) : (tmp = tmp->next))
+	{
+		if (tmp->winref == newref)
+			tmp->winref = oldref;
+		else if (tmp->winref == oldref)
+			tmp->winref = newref;
+	}
+}
+
