@@ -1,4 +1,4 @@
-/* $EPIC: parse.c,v 1.53 2013/07/16 01:03:57 jnelson Exp $ */
+/* $EPIC: parse.c,v 1.78 2006/06/23 05:03:11 jnelson Exp $ */
 /*
  * parse.c: handles messages from the server.   Believe it or not.  I
  * certainly wouldn't if I were you. 
@@ -46,7 +46,7 @@
 #include "lastlog.h"
 #include "ircaux.h"
 #include "sedcrypt.h"
-#include "termx.h"
+#include "term.h"
 #include "flood.h"
 #include "window.h"
 #include "screen.h"
@@ -54,7 +54,6 @@
 #include "numbers.h"
 #include "parse.h"
 #include "notify.h"
-#include "notice.h"
 #include "timer.h"
 
 #define STRING_CHANNEL 	'+'
@@ -124,10 +123,9 @@ const char *	PasteArgs (const char **arg_list, int paste_point)
  *
  * This doesnt strip out extraneous spaces any more.
  */
-static void 	BreakArgs (char *Input, char **Sender, const char **OutPut)
+static void 	BreakArgs (char *Input, const char **Sender, const char **OutPut)
 {
 	int	ArgCount;
-	char	*fuh;
 
 	/*
 	 * Paranoia.  Clean out any bogus ptrs still left on OutPut...
@@ -146,7 +144,9 @@ static void 	BreakArgs (char *Input, char **Sender, const char **OutPut)
 	 */
 	if (*Input == ':')
 	{
-		*Sender = ++Input;
+		char	*fuh;
+
+		fuh = ++Input;
 		while (*Input && *Input != space)
 			Input++;
 		if (*Input == space)
@@ -155,27 +155,28 @@ static void 	BreakArgs (char *Input, char **Sender, const char **OutPut)
 		/*
 		 * Look to see if the optional !user@host is present.
 		 */
-		fuh = *Sender;
+		*Sender = fuh;
 		while (*fuh && *fuh != '!')
 			fuh++;
 		if (*fuh == '!')
 			*fuh++ = 0;
+		FromUserHost = fuh;
 	}
 	/*
 	 * No sender present.
 	 */
 	else
-		*Sender = fuh = empty_string;
-
-	FromUserHost = fuh;
+		*Sender = FromUserHost = empty_string;
 
 	/*
 	 * Now we go through the argument list...
 	 */
 	for (;;)
 	{
+/*
 		while (*Input && *Input == space)
 			Input++;
+*/
 
 		if (!*Input)
 			break;
@@ -201,42 +202,34 @@ static void 	BreakArgs (char *Input, char **Sender, const char **OutPut)
 /* in response to a TOPIC message from the server */
 static void	p_topic (const char *from, const char *comm, const char **ArgList)
 {
-	const char 	*high, *channel, *new_topic;
+	const char 	*channel, *new_topic;
+	int	l;
 
 	if (!(channel = ArgList[0])) 
 		{ rfc1459_odd(from, comm, ArgList); return; }
 	if (!(new_topic = ArgList[1])) 
 		{ rfc1459_odd(from, comm, ArgList); return; }
 
-	switch (check_ignore_channel(from, FromUserHost, 
-					channel, IGNORE_TOPICS))
-	{
-		case IGNORED:
-			return;
-		case HIGHLIGHTED:
-			high = highlight_char;
-			break;
-		default:
-			high = empty_string;
-			break;
-	}
-
-	if (new_check_flooding(from, FromUserHost, channel, new_topic, TOPIC_FLOOD))
+	if (check_ignore_channel(from, FromUserHost, 
+				channel, LEVEL_TOPIC) == IGNORED)
 		return;
 
-	message_from(channel, LOG_CRAP);
+	if (new_check_flooding(from, FromUserHost, 
+				channel, new_topic, LEVEL_TOPIC))
+		return;
+
+	l = message_from(channel, LEVEL_TOPIC);
 	if (do_hook(TOPIC_LIST, "%s %s %s", from, channel, new_topic))
 		say("%s has changed the topic on channel %s to %s",
 			from, check_channel_type(channel), new_topic);
-	message_from(NULL, LOG_CRAP);
+	pop_message_from(l);
 }
 
 static void	p_wallops (const char *from, const char *comm, const char **ArgList)
 {
 	const char 	*message;
 	int 	server_wallop;
-	int	level;
-	char	*high;
+	int	l;
 
 	if (!(message = ArgList[0]))
 		{ rfc1459_odd(from, comm, ArgList); return; }
@@ -244,54 +237,34 @@ static void	p_wallops (const char *from, const char *comm, const char **ArgList)
 	server_wallop = strchr(from, '.') ? 1 : 0;
 
 	/* Check for ignores... */
-	switch (check_ignore(from, FromUserHost, IGNORE_WALLOPS))
-	{
-		case (IGNORED):
-			return;
-		case (HIGHLIGHTED):
-			high = highlight_char;
-			break;
-		default:
-			high = empty_string;
-			break;
-	}
+	if (check_ignore(from, FromUserHost, LEVEL_WALLOP) == IGNORED)
+		return;
 
 	/* Check for floods... servers are exempted from flood checks */
 	if (!server_wallop && check_flooding(from, FromUserHost, 
-						WALLOP_FLOOD, message))
+						LEVEL_WALLOP, message))
 		return;
 
 
-	message_from(from, LOG_WALLOP);
-	level = set_lastlog_msg_level(LOG_WALLOP);
-
+	l = message_from(from, LEVEL_WALLOP);
 	if (do_hook(WALLOP_LIST, "%s %c %s", 
 				from, 
 				server_wallop ? 'S' : '*', 
 				message))
-		put_it("%s!%s%s!%s %s", 
-				high, from, 
-				server_wallop ? empty_string : star, 
-				high, message);
-
-	if (beep_on_level & LOG_WALLOP)
-		beep_em(1);
-
-	set_lastlog_msg_level(level);
-	message_from(NULL, LOG_CRAP);
+		put_it("!%s%s! %s", 
+				from, server_wallop ? empty_string : star, 
+				message);
+	pop_message_from(l);
 }
 
 static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 {
 	const char	*target, *message;
-	int		level,
-			hook_type,
-			flood_type,
-			log_type;
+	int		hook_type,
+			level;
 	const char	*hook_format;
 	const char	*flood_channel = NULL;
-	unsigned char	ignore_type;
-	char		*high;
+	int	l;
 
 	PasteArgs(ArgList, 1);
 	if (!(target = ArgList[0]))
@@ -320,39 +293,33 @@ static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 	/* ooops. cant just do is_channel(to) because of # walls... */
 	if (is_channel(target) && im_on_channel(target, from_server))
 	{
-		log_type = LOG_PUBLIC;
+		level = LEVEL_PUBLIC;
 		flood_channel = target;
-		flood_type = PUBLIC_FLOOD;
-		ignore_type = IGNORE_PUBLIC;
 
 		if (!is_channel_nomsgs(target, from_server) && 
 				!is_on_channel(target, from)) {
 			hook_type = PUBLIC_MSG_LIST;
-			hook_format = "%s(%s/%s)%s %s";
+			hook_format = "(%s/%s) %s";
 		} else if (is_current_channel(target, from_server)) {
 			hook_type = PUBLIC_LIST;
-			hook_format = "%s<%s%.0s>%s %s";
+			hook_format = "<%s%.0s> %s";
 		} else {
 			hook_type = PUBLIC_OTHER_LIST;
-			hook_format = "%s<%s:%s>%s %s";
+			hook_format = "<%s:%s> %s";
 		}
 	}
 	else if (!is_me(from_server, target))
 	{
-		log_type = LOG_WALL;
+		level = LEVEL_WALL;
 		flood_channel = NULL;
-		flood_type = WALL_FLOOD;
-		ignore_type = IGNORE_WALLS;
 
 		hook_type = MSG_GROUP_LIST;
-		hook_format = "%s<-%s:%s->%s %s";
+		hook_format = "<-%s:%s-> %s";
 	}
 	else
 	{
-		log_type = LOG_MSG;
+		level = LEVEL_MSG;
 		flood_channel = NULL;
-		flood_type = MSG_FLOOD;
-		ignore_type = IGNORE_MSGS;
 
 		hook_type = MSG_LIST;
 		hook_format = NULL;	/* See below */
@@ -362,17 +329,10 @@ static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 	if (!target || !*target)
 		target = from;		/* Target is actually sender here */
 
-	switch (check_ignore_channel(from, FromUserHost, target, ignore_type))
+	if (check_ignore_channel(from, FromUserHost, target, level) == IGNORED)
 	{
-		case IGNORED:
-			set_server_doing_privmsg(from_server, 0);
-			return;
-		case HIGHLIGHTED:
-			high = highlight_char;
-			break;
-		default:
-			high = empty_string;
-			break;
+		set_server_doing_privmsg(from_server, 0);
+		return;
 	}
 
 	/* Encrypted privmsgs are specifically exempted from flood control */
@@ -381,16 +341,11 @@ static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 		int	do_return = 1;
 
 		sed = 0;
-
-		level = set_lastlog_msg_level(log_type);
-		message_from(target, log_type);
-
+		l = message_from(target, level);
 		if (do_hook(ENCRYPTED_PRIVMSG_LIST, "%s %s %s", from, target,
 				message))
 			do_return = 0;
-
-		set_lastlog_msg_level(level);
-		message_from(NULL, LOG_CRAP);
+		pop_message_from(l);
 
 		if (do_return) {
 			set_server_doing_privmsg(from_server, 0);
@@ -399,7 +354,7 @@ static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 	}
 
 	if (new_check_flooding(from, FromUserHost, flood_channel, 
-				message, flood_type)) {
+				message, level)) {
 		set_server_doing_privmsg(from_server, 0);
 		return;
 	}
@@ -412,13 +367,8 @@ static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 	else if (hook_type == MSG_LIST)
 		set_server_recv_nick(from_server, from);
 
-	/* Beep the user if they asked us to */
-	if (beep_on_level & log_type)
-	    beep_em(1);
-
 	/* Go ahead and throw it to the user */
-	level = set_lastlog_msg_level(log_type);
-	message_from(target, log_type);
+	l = message_from(target, level);
 
 	if (do_hook(GENERAL_PRIVMSG_LIST, "%s %s %s", from, target, message))
 	{
@@ -426,31 +376,24 @@ static void	p_privmsg (const char *from, const char *comm, const char **ArgList)
 	    {
 		const char *away = get_server_away(NOSERV);
 
-		if (away)
-			beep_em(get_int_var(BEEP_WHEN_AWAY_VAR));
-
 		if (do_hook(hook_type, "%s %s", from, message))
 		{
 		    if (away)
 		    {
 			time_t blah = time(NULL);
-			put_it("%s*%s*%s %s <%.16s>", 
-				high, from, high, message, 
-				ctime(&blah));
+			put_it("*%s* %s <%.16s>", from, message, ctime(&blah));
 		    }
 		    else
-			put_it("%s*%s*%s %s", high, from, high,message);
+			put_it("*%s* %s", from, message);
 		}
 	    }
 
 	    else if (do_hook(hook_type, "%s %s %s", from, target, message))
-		put_it(hook_format, high, from, check_channel_type(target), 
-				high, message);
+		put_it(hook_format, from, check_channel_type(target), message);
 	}
 
 	/* Clean up and go home. */
-	set_lastlog_msg_level(level);
-	message_from(NULL, LOG_CRAP);
+	pop_message_from(l);
 	set_server_doing_privmsg(from_server, 0);
 
 	/* Alas, this is not protected by protocol enforcement. :( */
@@ -462,12 +405,11 @@ static void	p_quit (const char *from, const char *comm, const char **ArgList)
 	const char *	quit_message;
 	int		one_prints = 1;
 	const char *	chan;
+	int		l;
 
 	if (!(quit_message = ArgList[0]))
 		{ rfc1459_odd(from, comm, ArgList); return; }
 
-    do
-    {
 	/*
 	 * Normally, we do not throw the user a hook until after we
 	 * have taken care of administrative details.  But in this case,
@@ -475,32 +417,34 @@ static void	p_quit (const char *from, const char *comm, const char **ArgList)
 	 * so we cannot remove them from the channel until after we have
 	 * thrown the hook.  That is the only reason this is out of order.
 	 */
-	if (check_ignore(from, FromUserHost, IGNORE_CRAP) == IGNORED)
-		break;
+	if (check_ignore(from, FromUserHost, LEVEL_QUIT) == IGNORED)
+		goto remove_quitter;
 
-	for (chan = walk_channels(1, from); chan; 
-		 chan = walk_channels(0, from))
+	if (check_flooding(from, FromUserHost, LEVEL_QUIT, quit_message))
+		goto remove_quitter;
+
+	for (chan = walk_channels(1, from); chan; chan = walk_channels(0, from))
 	{
-		if (check_ignore_channel(from, FromUserHost, chan,
-						IGNORE_CRAP) == IGNORED)
-		{
-			one_prints = 0;
-			continue;
-		}
+	    if (check_ignore_channel(from, FromUserHost, 
+					chan, LEVEL_QUIT) == IGNORED)
+	    {
+		one_prints = 0;
+		continue;
+	    }
 
-		message_from(chan, LOG_CRAP);
-		if (!do_hook(CHANNEL_SIGNOFF_LIST, "%s %s %s", chan, from, 
-								quit_message))
-			one_prints = 0;
-		message_from(NULL, LOG_CRAP);
+	    l = message_from(chan, LEVEL_QUIT);
+	    if (!do_hook(CHANNEL_SIGNOFF_LIST, "%s %s %s", chan, from, 
+							quit_message))
+		one_prints = 0;
+	    pop_message_from(l);
 	}
 
 	if (one_prints)
 	{
-		message_from(what_channel(from), LOG_CRAP);
+		l = message_from(what_channel(from), LEVEL_QUIT);
 		if (do_hook(SIGNOFF_LIST, "%s %s", from, quit_message))
 			say("Signoff: %s (%s)", from, quit_message);
-		message_from(NULL, LOG_CRAP);
+		pop_message_from(l);
 	}
 
 	/*
@@ -511,9 +455,8 @@ static void	p_quit (const char *from, const char *comm, const char **ArgList)
 	 * until the top of the next minute.
 	 */
 	notify_mark(from_server, from, 0, 0);
-    }
-    while (0);
 
+remove_quitter:
 	/* Send all data about this unperson to the memory hole. */
 	remove_from_channel(NULL, from, from_server);
 }
@@ -545,7 +488,7 @@ static void	p_pong (const char *from, const char *comm, const char **ArgList)
 			return;
 	}
 
-	if (check_ignore(from, FromUserHost, IGNORE_CRAP) == IGNORED)
+	if (check_ignore(from, FromUserHost, LEVEL_OTHER) == IGNORED)
 		return;
 
 	if (do_hook(PONG_LIST, "%s %s %s", from, pong_server, pong_message))
@@ -569,52 +512,13 @@ static void	p_error (const char *from, const char *comm, const char **ArgList)
 		say("%s %s", from, the_error);
 }
 
-#if 0
-static void	add_user_who (int refnum, const char *from, const char *comm, const char **ArgList)
-{
-	const char 	*channel, *user, *host, *server, *nick;
-	size_t	size;
-	char 	*userhost;
-
-	if (!(channel = ArgList[0]))
-		{ rfc1459_odd(from, "*", ArgList); return; }
-	if (!(user = ArgList[1]))
-		{ rfc1459_odd(from, "*", ArgList); return; }
-	if (!(host = ArgList[2]))
-		{ rfc1459_odd(from, "*", ArgList); return; }
-	if (!(server = ArgList[3]))
-		{ rfc1459_odd(from, "*", ArgList); return; }
-	if (!(nick = ArgList[4]))
-		{ rfc1459_odd(from, "*", ArgList); return; }
-
-	/* Obviously this is safe. */
-	size = strlen(user) + strlen(host) + 2;
-	userhost = alloca(size);
-	snprintf(userhost, size, "%s@%s", user, host);
-	add_userhost_to_channel(channel, nick, refnum, userhost);
-}
-
-static void	add_user_end (int refnum, const char *from, const char *comm, const char **ArgList)
-{
-	char *	copy;
-	char *	channel;
-
-	if (!ArgList[0])
-		{ rfc1459_odd(from, "*", ArgList); return; }
-
-	copy = LOCAL_COPY(ArgList[0]);
-	channel = next_arg(copy, &copy);
-	channel_not_waiting(channel, refnum);
-}
-#endif
-
 static void	p_channel (const char *from, const char *comm, const char **ArgList)
 {
 	const char	*channel;
 	char 	*c;
 	int 	op = 0, vo = 0, ha = 0;
 	char 	extra[20];
-	char	*high;
+	int	l;
 
 	/* We cannot join channel 0 */
 	if (!(channel = ArgList[0]))
@@ -636,15 +540,8 @@ static void	p_channel (const char *from, const char *comm, const char **ArgList)
 
 	if (is_me(from_server, from))
 	{
-#if 0
-		char *copy = LOCAL_COPY(channel);
-#endif
-
 		add_channel(channel, from_server);
 		send_to_server("MODE %s", channel);
-#if 0
-		whobase(from_server, copy, add_user_who, add_user_end);
-#endif
 	}
 	else
 	{
@@ -652,19 +549,11 @@ static void	p_channel (const char *from, const char *comm, const char **ArgList)
 		add_userhost_to_channel(channel, from, from_server, FromUserHost);
 	}
 
-	switch (check_ignore_channel(from, FromUserHost, channel, IGNORE_JOINS))
-	{
-		case IGNORED:
-			return;
-		case HIGHLIGHTED:
-			high = highlight_char;
-			break;
-		default:
-			high = empty_string;
-			break;
-	}
+	if (check_ignore_channel(from, FromUserHost, 
+				channel, LEVEL_JOIN) == IGNORED)
+		return;
 
-	if (new_check_flooding(from, FromUserHost, channel, star, JOIN_FLOOD))
+	if (new_check_flooding(from, FromUserHost, channel, star, LEVEL_JOIN))
 		return;
 
 	set_server_joined_nick(from_server, from);
@@ -675,13 +564,13 @@ static void	p_channel (const char *from, const char *comm, const char **ArgList)
 	if (vo)
 		strlcat(extra, " (+v)", sizeof extra);
 
-	message_from(channel, LOG_CRAP);
+	l = message_from(channel, LEVEL_JOIN);
 	if (do_hook(JOIN_LIST, "%s %s %s %s", 
 			from, channel, FromUserHost, extra))
-		say("%s%s%s (%s) has joined channel %s%s%s%s", 
-			high, from, high, FromUserHost, 
-			high, check_channel_type(channel), high, extra);
-	message_from(NULL, LOG_CRAP);
+		say("%s (%s) has joined channel %s%s", 
+			from, FromUserHost, 
+			check_channel_type(channel), extra);
+	pop_message_from(l);
 
 	/*
 	 * The placement of this is purely ergonomic.  The user might
@@ -700,37 +589,28 @@ static void	p_channel (const char *from, const char *comm, const char **ArgList)
 static void 	p_invite (const char *from, const char *comm, const char **ArgList)
 {
 	const char	*invitee, *invited_to;
-	char	*high;
+	int	l;
 
 	if (!(invitee = ArgList[0]))
 		{ rfc1459_odd(from, comm, ArgList); return; }
 	if (!(invited_to = ArgList[1]))
 		{ rfc1459_odd(from, comm, ArgList); return; }
 
-	switch (check_ignore_channel(from, FromUserHost, 
-					invited_to, IGNORE_INVITES))
-	{
-		case IGNORED:
-			return;
-		case HIGHLIGHTED:
-			high = highlight_char;
-			break;
-		default:
-			high = empty_string;
-			break;
-	}
+	if (check_ignore_channel(from, FromUserHost, 
+				invited_to, LEVEL_INVITE) == IGNORED)
+		return;
 
-	if (check_flooding(from, FromUserHost, INVITE_FLOOD, invited_to))
+	if (check_flooding(from, FromUserHost, LEVEL_INVITE, invited_to))
 		return;
 
 	set_server_invite_channel(from_server, invited_to);
 	set_server_recv_nick(from_server, from);
 
-	message_from(from, LOG_CRAP);
+	l = message_from(from, LEVEL_INVITE);
 	if (do_hook(INVITE_LIST, "%s %s %s", from, invited_to, FromUserHost))
-		say("%s%s (%s)%s invites you to channel %s", high,
-			from, FromUserHost, high, invited_to);
-	message_from(NULL, LOG_CRAP);
+		say("%s (%s) invites you to channel %s", 
+			from, FromUserHost, invited_to);
+	pop_message_from(l);
 }
 
 /* 
@@ -790,10 +670,7 @@ static void	p_kill (const char *from, const char *comm, const char **ArgList)
 
 
 	/*
-	 * We've been killed.  Save our channels, close up shop on 
-	 * this server.  We need to call window_check_servers before
-	 * we do any output, in case we want the output to go anywhere
-	 * meaningful.
+	 * We've been killed.  Bummer for us.
 	 */
 	if (!reason) { reason = "No Reason Given"; }
 
@@ -807,29 +684,9 @@ static void	p_kill (const char *from, const char *comm, const char **ArgList)
 	/* 
 	 * If we are a bot, and /on disconnect didnt hook, 
 	 * then we arent going anywhere.  We might as well quit.
-	 * Also quit if #define QUIT_ON_OPERATOR_KILL
 	 */
-#ifndef QUIT_ON_OPERATOR_KILL
 	if (background && !hooked)
-#endif
-	{
-		say("Too bad, you lose.");
 		irc_exit(1, NULL);
-	}
-
-	if (!background && get_int_var(AUTO_RECONNECT_VAR))
-	{
-		char *	sc = new_malloc(16);
-
-		snprintf(sc, 15, "%d", from_server);
-		add_timer(0, empty_string, 
-			  get_int_var(AUTO_RECONNECT_DELAY_VAR), 1,
-			  auto_reconnect_callback,
-			  sc, NULL, current_window->refnum);
-	}
-
-	/* Suppress auto-reconnect (to wait for above timer) */
-	server_reconnects_to(from_server, NOSERV);
 }
 
 static void	p_ping (const char *from, const char *comm, const char **ArgList)
@@ -869,8 +726,8 @@ static void	p_nick (const char *from, const char *comm, const char **ArgList)
 	int		been_hooked = 0,
 			its_me = 0;
 	const char	*chan;
-	char		*high;
 	int		ignored = 0;
+	int		l;
 
 	if (!(new_nick = ArgList[0]))
 		{ rfc1459_odd(from, comm, ArgList); return; }
@@ -882,47 +739,40 @@ static void	p_nick (const char *from, const char *comm, const char **ArgList)
 	{
 		its_me = 1;
 		accept_server_nickname(from_server, new_nick);
-		set_server_nickname_pending(from_server, 0);
 	}
 
-	switch (check_ignore(from, FromUserHost, IGNORE_NICKS))
-	{
-		case IGNORED:
-			goto do_rename;
-		case HIGHLIGHTED:
-			high = highlight_char;
-			break;
-		default:
-			high = empty_string;
-			break;
-	}
+	if (check_ignore(from, FromUserHost, LEVEL_NICK) == IGNORED)
+		goto do_rename;
 
-	if (check_flooding(from, FromUserHost, NICK_FLOOD, new_nick))
+	if (check_flooding(from, FromUserHost, LEVEL_NICK, new_nick))
 		goto do_rename;
 
 	for (chan = walk_channels(1, from); chan; chan = walk_channels(0, from))
 	{
 		if (check_ignore_channel(from, FromUserHost, chan, 
-						IGNORE_NICKS) == IGNORED)
+						LEVEL_NICK) == IGNORED)
 		{
 			ignored = 1;
 			continue;
 		}
 
-		message_from(chan, LOG_CRAP);
+		l = message_from(chan, LEVEL_NICK);
 		if (!do_hook(CHANNEL_NICK_LIST, "%s %s %s", chan, from, new_nick))
 			been_hooked = 1;
+		pop_message_from(l);
 	}
 
 	if (!been_hooked && !ignored)
 	{
 		if (its_me)
-			message_from(NULL, LOG_CRAP);
+			l = message_from(NULL, LEVEL_NICK);
 		else
-			message_from(what_channel(from), LOG_CRAP);
+			l = message_from(what_channel(from), LEVEL_NICK);
 
 		if (do_hook(NICKNAME_LIST, "%s %s", from, new_nick))
 			say("%s is now known as %s", from, new_nick);
+
+		pop_message_from(l);
 	}
 
 do_rename:
@@ -936,6 +786,7 @@ static void	p_mode (const char *from, const char *comm, const char **ArgList)
 	const char	*target, *changes;
 	const char	*m_target;
 	const char	*type;
+	int		l;
 
 	PasteArgs(ArgList, 1);
 	if (!(target = ArgList[0]))
@@ -958,20 +809,24 @@ static void	p_mode (const char *from, const char *comm, const char **ArgList)
 		type = "for user";
 	}
 
-	if (check_ignore_channel(from, FromUserHost, target, IGNORE_CRAP) 
-					!= IGNORED)
-	{
-		message_from(m_target, LOG_CRAP);
-		if (do_hook(MODE_LIST, "%s %s %s", from, target, changes))
-		    say("Mode change \"%s\" %s %s by %s",
-					changes, type, target, from);
-		message_from(NULL, LOG_CRAP);
-	}
+	if (check_ignore_channel(from, FromUserHost, 
+					target, LEVEL_MODE) == IGNORED)
+		goto do_update_mode;
 
+	if (new_check_flooding(from, FromUserHost, target, changes, LEVEL_MODE))
+		goto do_update_mode;
+
+	l = message_from(m_target, LEVEL_MODE);
+	if (do_hook(MODE_LIST, "%s %s %s", from, target, changes))
+	    say("Mode change \"%s\" %s %s by %s",
+					changes, type, target, from);
+	pop_message_from(l);
+
+do_update_mode:
 	if (is_channel(target))
 		update_channel_mode(target, changes);
 	else
-		update_user_mode(changes);
+		update_user_mode(from_server, changes);
 }
 
 static void strip_modes (const char *from, const char *channel, const char *line)
@@ -981,6 +836,7 @@ static void strip_modes (const char *from, const char *channel, const char *line
 	char	mag = '+'; /* XXXX Bogus */
         char    *copy = (char *) 0;
 	char	*free_copy;
+	int	l;
 
 	free_copy = LOCAL_COPY(line);
 	copy = free_copy;
@@ -988,7 +844,7 @@ static void strip_modes (const char *from, const char *channel, const char *line
 
 	if (is_channel(channel))
 	{
-	    message_from(channel, LOG_CRAP);
+	    l = message_from(channel, LEVEL_MODE);
 
 	    for (pointer = mode; *pointer; pointer++)
 	    {
@@ -1023,12 +879,12 @@ static void strip_modes (const char *from, const char *channel, const char *line
 				from,channel,mag,c);
 	    }
 
-	    message_from(NULL, LOG_CRAP);
+	    pop_message_from(l);
 	}
 
 	else /* User mode */
 	{
-	    message_from(NULL, LOG_CRAP);
+	    l = message_from(NULL, LEVEL_MODE);
 
 	    for (pointer = mode; *pointer; pointer++)
 	    {
@@ -1038,20 +894,6 @@ static void strip_modes (const char *from, const char *channel, const char *line
 		{
 			case '+' :
 			case '-' : mag = c; break;
-#if 0	/* Not implemented yet */
-			case 's' :
-			{
-				if (umode_s_takes_arg(from_server))
-				{
-					do_hook(MODE_STRIPPED_LIST, 
-						"%s %s %c%c %s", 
-						from, channel, mag, c, 
-						next_arg(copy, &copy));
-					break;
-				}
-				/* ELSE FALLTHROUGH */
-			}
-#endif
 			default  : 
 				do_hook(MODE_STRIPPED_LIST, 
 					"%s %s %c%c", 
@@ -1060,7 +902,7 @@ static void strip_modes (const char *from, const char *channel, const char *line
 		}
 	    }
 
-	    message_from(NULL, LOG_CRAP);
+	    pop_message_from(l);
 	}
 
 }
@@ -1068,6 +910,7 @@ static void strip_modes (const char *from, const char *channel, const char *line
 static void	p_kick (const char *from, const char *comm, const char **ArgList)
 {
 	const char	*channel, *victim, *comment;
+	int	l;
 
 	if (!(channel = ArgList[0]))
 		{ rfc1459_odd(from, comm, ArgList); return; }
@@ -1081,7 +924,7 @@ static void	p_kick (const char *from, const char *comm, const char **ArgList)
 	 */
 	if (is_me(-1, victim))
 	{
-		Window *win, *old_tw, *old_cw;
+		Window *win, *old_cw;
 
 		/*
 		 * Uh-oh.  If win is null we have a problem.
@@ -1104,29 +947,12 @@ static void	p_kick (const char *from, const char *comm, const char **ArgList)
 		    return;
 		}
 
-		if (get_int_var(AUTO_REJOIN_VAR))
-		{
-			const char *key = get_channel_key(channel, from_server);
-			if (!key)
-				key = empty_string;
-
-			add_timer(0, empty_string, 
-				  get_int_var(AUTO_REJOIN_DELAY_VAR), 1,
-				  auto_rejoin_callback, 
-				  malloc_sprintf(NULL, "%s %d %d %s", channel, 
-						from_server, 
-						win->refnum,
-						key),
-				  NULL, win->refnum);
-		}
 		remove_channel(channel, from_server);
 		update_all_status();
 
-		old_tw = to_window;
 		old_cw = current_window;
 		current_window = win;
-		to_window = win;
-		message_from(channel, LOG_CRAP);
+		l = message_setall(win->refnum, channel, LEVEL_KICK);
 
 		if (do_hook(KICK_LIST, "%s %s %s %s", victim, from, 
 					check_channel_type(channel), comment))
@@ -1134,27 +960,32 @@ static void	p_kick (const char *from, const char *comm, const char **ArgList)
 					check_channel_type(channel), from, 
 					comment);
 
-		message_from(NULL, LOG_CRAP);
-		to_window = old_tw;
+		pop_message_from(l);
 		current_window = old_cw;
 		return;
 	}
 
 	if (check_ignore_channel(from, FromUserHost, 
-				 channel, IGNORE_CRAP) != IGNORED)
-	{
-	    if (check_ignore_channel(victim, fetch_userhost(from_server, victim), channel, IGNORE_CRAP) != IGNORED)
-	    {
-		message_from(channel, LOG_CRAP);
-		if (do_hook(KICK_LIST, "%s %s %s %s", 
-				victim, from, channel, comment))
-			say("%s has been kicked off channel %s by %s (%s)", 
-				victim, check_channel_type(channel), 
-				from, comment);
-		message_from(NULL, LOG_CRAP);
-	    }
-	}
+				channel, LEVEL_KICK) == IGNORED)
+		goto do_remove_nick;
 
+	if (check_ignore_channel(victim, fetch_userhost(from_server, NULL, 
+							victim), 
+					channel, LEVEL_KICK) == IGNORED)
+		goto do_remove_nick;
+
+
+	if (new_check_flooding(from, FromUserHost, channel, victim, LEVEL_KICK))
+		goto do_remove_nick;
+
+	l = message_from(channel, LEVEL_KICK);
+	if (do_hook(KICK_LIST, "%s %s %s %s", 
+			victim, from, channel, comment))
+		say("%s has been kicked off channel %s by %s (%s)", 
+			victim, check_channel_type(channel), from, comment);
+	pop_message_from(l);
+
+do_remove_nick:
 	/*
 	 * The placement of this is purely ergonomic.  When someone is
 	 * kicked, the user may want to know what their userhost was so 
@@ -1170,6 +1001,7 @@ static void	p_kick (const char *from, const char *comm, const char **ArgList)
 static void	p_part (const char *from, const char *comm, const char **ArgList)
 {
 	const char	*channel, *reason;
+	int	l;
 
 	PasteArgs(ArgList, 1);
 	if (!(channel = ArgList[0]))
@@ -1177,24 +1009,26 @@ static void	p_part (const char *from, const char *comm, const char **ArgList)
 	if (!(reason = ArgList[1])) { }
 
 	if ((check_ignore_channel(from, FromUserHost, 
-					channel, IGNORE_PARTS) != IGNORED))
+				channel, LEVEL_PART) != IGNORED)
+		&& !new_check_flooding(from, FromUserHost, channel,
+			reason ? reason : star, LEVEL_PART))
 	{
-		message_from(channel, LOG_CRAP);
+		l = message_from(channel, LEVEL_PART);
 		if (reason)		/* Dalnet part messages */
 		{
-			if (do_hook(LEAVE_LIST, "%s %s %s %s", 
+			if (do_hook(PART_LIST, "%s %s %s %s", 
 				from, channel, FromUserHost, reason))
 			    say("%s has left channel %s because (%s)", 
 				from, check_channel_type(channel), reason);
 		}
 		else
 		{
-			if (do_hook(LEAVE_LIST, "%s %s %s", 
+			if (do_hook(PART_LIST, "%s %s %s", 
 				from, channel, FromUserHost))
 			    say("%s has left channel %s", 
 				from, check_channel_type(channel));
 		}
-		message_from(NULL, LOG_CRAP);
+		pop_message_from(l);
 	}
 
 	if (is_me(from_server, from))
@@ -1234,6 +1068,228 @@ static void	p_rpong (const char *from, const char *comm, const char **ArgList)
 		from, target_server, millisecs, delay);
 }
 
+/* 
+ * This is a special subset of server (OPER) notice.
+ */
+static int 	p_killmsg (const char *from, const char *cline)
+{
+	char *poor_sap;
+	char *bastard;
+	const char *path_to_bastard;
+	char *reason;
+	char *line;
+
+	line = LOCAL_COPY(cline);
+	poor_sap = next_arg(line, &line);
+
+	/* Dalnet kill BBC and doesnt append the period */
+	if (!end_strcmp(poor_sap, ".", 1))
+		chop(poor_sap, 1);
+
+	/* dalnet kill BBC and doesnt use "From", but "from" */
+	if (my_strnicmp(line, "From ", 5))
+	{
+		yell("Attempted to parse an ill-formed KILL request [%s %s]",
+			poor_sap, line);
+		return 0;
+	}
+	line += 5;
+	bastard = next_arg(line, &line);
+
+	/* Hybrid BBC and doesn't include the kill-path. */
+	/* Fend off future BBC kills */
+	if (my_strnicmp(line, "Path: ", 6))
+	{
+		path_to_bastard = "*";
+		reason = line;		/* Hope for the best */
+	}
+	else
+	{
+		line += 6;
+		path_to_bastard = next_arg(line, &line);
+		reason = line;
+	}
+
+	return !do_hook(KILL_LIST, "%s %s %s %s %s", from, poor_sap, bastard,
+					path_to_bastard, reason);
+}
+
+
+/*
+ * This is a special subset of NOTICEs, that were sent from the server
+ * we are connected to (not a remote server), to us (not to a channel).
+ */
+static 	void 	p_snotice (const char *from, const char *to, const char *line)
+{
+	const char *	f;
+	int	l;
+	int	retval;
+
+	f = from;
+	if (!f || !*f)
+		if (!(f = get_server_itsname(from_server)))
+			f = get_server_name(from_server);
+
+	/* OPERator Notices */
+	if (!strncmp(line, "*** Notice -- ", 13))
+	{
+		if (!strncmp(line + 14, "Received KILL message for ", 26))
+			if (p_killmsg(f, line + 40))
+				return;
+
+		l = message_from(to, LEVEL_OPNOTE);
+		retval = do_hook(OPER_NOTICE_LIST, "%s %s", f, line + 14);
+		pop_message_from(l);
+		if (!retval)
+			return;
+	}
+
+	l = message_from(to, LEVEL_SNOTE);
+
+	/* Check to see if the notice already has its own header... */
+	if (do_hook(GENERAL_NOTICE_LIST, "%s %s %s", f, to, line))
+	{
+	    if (*line == '*' || *line == '#')
+	    {
+		if (do_hook(SERVER_NOTICE_LIST, "%s %s", f, line))
+			put_it("%s", line);
+	    }
+	    else
+		if (do_hook(SERVER_NOTICE_LIST, "%s *** %s", f, line))
+			say("%s", line);
+	}
+
+	pop_message_from(l);
+}
+
+/*
+ * The main handler for those wacky NOTICE commands...
+ * This is as much like p_privmsg as i can get away with.
+ */
+static void 	p_notice (const char *from, const char *comm, const char **ArgList)
+{
+	const char 	*target, *message;
+	int		hook_type;
+	const char *	flood_channel = NULL;
+	int		l;
+
+	PasteArgs(ArgList, 1);
+	if (!(target = ArgList[0]))
+		{ rfc1459_odd(from, comm, ArgList); return; }
+	if (!(message = ArgList[1]))
+		{ rfc1459_odd(from, comm, ArgList); return; }
+
+	set_server_doing_notice(from_server, 1);
+	sed = 0;
+
+	/* Do normal /CTCP reply handling */
+	/* XXX -- Casting "message" to (char *) is cheating. */
+	message = do_notice_ctcp(from, target, (char *)
+#ifdef HAVE_INTPTR_T
+							(intptr_t)
+#endif
+								message);
+	if (!*message) {
+		new_check_flooding(from, FromUserHost, flood_channel, 
+					message, LEVEL_NOTICE);
+		set_server_doing_notice(from_server, 0);
+		return;
+	}
+
+	/* Check to see if it is a "Server Notice" */
+	if ((!from || !*from) || !strcmp(get_server_itsname(from_server), from))
+	{
+		p_snotice(from, target, message);
+		set_server_doing_notice(from_server, 0);
+		return;
+	}
+	/* For pesky prefix-less NOTICEs substitute the server's name */
+	if (!from || !*from)
+		from = get_server_name(from_server);
+
+	/*
+	 * Note that NOTICEs from servers are not "server notices" unless
+	 * the target is not a channel (ie, it is sent to us).  Any notice
+	 * that is sent to a channel is a normal NOTICE, notwithstanding
+	 * _who_ sent it.
+	 */
+	if (is_channel(target) && im_on_channel(target, from_server))
+	{
+		flood_channel = target;
+		hook_type = PUBLIC_NOTICE_LIST;
+	}
+	else if (!is_me(from_server, target))
+	{
+		flood_channel = NULL;
+		hook_type = NOTICE_LIST;
+	}
+	else
+	{
+		flood_channel = NULL;
+		hook_type = NOTICE_LIST;
+		target = from;
+	}
+
+	/* Check for /ignore's */
+	if (check_ignore_channel(from, FromUserHost, 
+				target, LEVEL_NOTICE) == IGNORED)
+	{
+		set_server_doing_notice(from_server, 0);
+		return;
+	}
+
+	/* Let the user know if it is an encrypted notice */
+	/* Note that this is always hooked, even during a flood */
+	if (sed)
+	{
+		int	do_return = 1;
+
+		sed = 0;
+		l = message_from(target, LEVEL_NOTICE);
+
+		if (do_hook(ENCRYPTED_NOTICE_LIST, "%s %s %s", 
+				from, target, message))
+			do_return = 0;
+
+		pop_message_from(l);
+
+		if (do_return) {
+			set_server_doing_notice(from_server, 0);
+			return;
+		}
+	}
+
+	if (new_check_flooding(from, FromUserHost, flood_channel, 
+					message, LEVEL_NOTICE)) {
+		set_server_doing_notice(from_server, 0);
+		return;
+	}
+
+
+	/* Go ahead and throw it to the user */
+	l = message_from(target, LEVEL_NOTICE);
+
+	if (do_hook(GENERAL_NOTICE_LIST, "%s %s %s", from, target, message))
+	{
+	    if (hook_type == NOTICE_LIST)
+	    {
+		if (do_hook(hook_type, "%s %s", from, message))
+			put_it("-%s- %s", from, message);
+	    }
+	    else
+	    {
+		if (do_hook(hook_type, "%s %s %s", from, target, message))
+			put_it("-%s:%s- %s", from, target, message);
+	    }
+	}
+
+	/* Clean up and go home. */
+	pop_message_from(l);
+	set_server_doing_notice(from_server, 0);
+
+	/* Alas, this is not protected by protocol enforcement. :( */
+	notify_mark(from_server, from, 1, 0);
+}
 
 void	rfc1459_odd (const char *from, const char *comm, const char **ArgList)
 {
@@ -1253,50 +1309,50 @@ void	rfc1459_odd (const char *from, const char *comm, const char **ArgList)
 }
 
 protocol_command rfc1459[] = {
-{	"ADMIN",	NULL,		NULL,		0		},
-{	"AWAY",		NULL,		NULL,		0		},
-{ 	"CONNECT",	NULL,		NULL,		0		},
-{	"ERROR",	p_error,	NULL,		0		},
-{	"ERROR:",	p_error,	NULL,		0		},
-{	"INFO",		NULL,		NULL,		0		},
-{	"INVITE",	p_invite,	NULL,		0		},
-{	"ISON",		NULL,		NULL,		PROTO_NOQUOTE	},
-{	"JOIN",		p_channel,	NULL,		0		},
-{	"KICK",		p_kick,		NULL,		0		},
-{	"KILL",		p_kill,		NULL,		0		},
-{	"LINKS",	NULL,		NULL,		0		},
-{	"LIST",		NULL,		NULL,		0		},
-{	"MODE",		p_mode,		NULL,		0		},
-{	"NAMES",	NULL,		NULL,		0		},
-{	"NICK",		p_nick,		NULL,		PROTO_NOQUOTE	},
-{	"NOTICE",	p_notice,	NULL,		0		},
-{	"OPER",		NULL,		NULL,		0		},
-{	"PART",		p_part,		NULL,		0		},
-{	"PASS",		NULL,		NULL,		0 		},
-{	"PING",		p_ping,		NULL,		0		},
-{	"PONG",		p_pong,		NULL,		0		},
-{	"PRIVMSG",	p_privmsg,	NULL,		0		},
-{	"QUIT",		p_quit,		NULL,		PROTO_DEPREC	},
-{	"REHASH",	NULL,		NULL,		0		},
-{	"RESTART",	NULL,		NULL,		0		},
-{	"RPONG",	p_rpong,	NULL,		0		},
-{	"SERVER",	NULL,		NULL,		PROTO_NOQUOTE	},
-{	"SILENCE",	p_silence,	NULL,		0		},
-{	"SQUIT",	NULL,		NULL,		0		},
-{	"STATS",	NULL,		NULL,		0		},
-{	"SUMMON",	NULL,		NULL,		0		},
-{	"TIME",		NULL,		NULL,		0		},
-{	"TOPIC",	p_topic,	NULL,		0		},
-{	"TRACE",	NULL,		NULL,		0		},
-{	"USER",		NULL,		NULL,		0		},
-{	"USERHOST",	NULL,		NULL,		PROTO_NOQUOTE	},
-{	"USERS",	NULL,		NULL,		0		},
-{	"VERSION",	NULL,		NULL,		0		},
-{	"WALLOPS",	p_wallops,	NULL,		0		},
-{	"WHO",		NULL,		NULL,		PROTO_DEPREC	},
-{	"WHOIS",	NULL,		NULL,		0		},
-{	"WHOWAS",	NULL,		NULL,		0		},
-{	NULL,		NULL,		NULL,		0		}
+{	"ADMIN",	NULL,		0		},
+{	"AWAY",		NULL,		0		},
+{ 	"CONNECT",	NULL,		0		},
+{	"ERROR",	p_error,	0		},
+{	"ERROR:",	p_error,	0		},
+{	"INVITE",	p_invite,	0		},
+{	"INFO",		NULL,		0		},
+{	"ISON",		NULL,		PROTO_QUOTEBAD	},
+{	"JOIN",		p_channel,	0		},
+{	"KICK",		p_kick,		0		},
+{	"KILL",		p_kill,		0		},
+{	"LINKS",	NULL,		0		},
+{	"LIST",		NULL,		0		},
+{	"MODE",		p_mode,		0		},
+{	"NAMES",	NULL,		0		},
+{	"NICK",		p_nick,		PROTO_QUOTEBAD	},
+{	"NOTICE",	p_notice,	0		},
+{	"OPER",		NULL,		0		},
+{	"PART",		p_part,		0		},
+{	"PASS",		NULL,		0 		},
+{	"PING",		p_ping,		0		},
+{	"PONG",		p_pong,		0		},
+{	"PRIVMSG",	p_privmsg,	0		},
+{	"QUIT",		p_quit,		PROTO_QUOTEBAD	},
+{	"REHASH",	NULL,		0		},
+{	"RESTART",	NULL,		0		},
+{	"RPONG",	p_rpong,	0		},
+{	"SERVER",	NULL,		PROTO_QUOTEBAD	},
+{	"SILENCE",	p_silence,	0		},
+{	"SQUIT",	NULL,		0		},
+{	"STATS",	NULL,		0		},
+{	"SUMMON",	NULL,		0		},
+{	"TIME",		NULL,		0		},
+{	"TRACE",	NULL,		0		},
+{	"TOPIC",	p_topic,	0		},
+{	"USER",		NULL,		0		},
+{	"USERHOST",	NULL,		PROTO_QUOTEBAD	},
+{	"USERS",	NULL,		0		},
+{	"VERSION",	NULL,		0		},
+{	"WALLOPS",	p_wallops,	0		},
+{	"WHO",		NULL,		PROTO_QUOTEBAD	},
+{	"WHOIS",	NULL,		0		},
+{	"WHOWAS",	NULL,		0		},
+{	NULL,		NULL,		0		}
 };
 #define NUMBER_OF_COMMANDS (sizeof(rfc1459) / sizeof(protocol_command)) - 2;
 int 	num_protocol_cmds = -1;
@@ -1307,15 +1363,12 @@ int 	num_protocol_cmds = -1;
  */
 void 	parse_server (const char *orig_line, size_t orig_line_size)
 {
-	char	*from;
+	const char	*from;
 	const char	*comm;
 	const char	**ArgList;
 	const char	*TrueArgs[MAXPARA + 2];	/* Include space for command */
-	protocol_command *retval;
 	int	loc;
-	int	cnt;
 	char	*line;
-	size_t	size;
 
 	if (num_protocol_cmds == -1)
 		num_protocol_cmds = NUMBER_OF_COMMANDS;
@@ -1331,14 +1384,15 @@ void 	parse_server (const char *orig_line, size_t orig_line_size)
 	else if (!do_hook(RAW_IRC_LIST, "* %s", orig_line))
 		return;
 
-	size = (orig_line_size + 1) * 11;
-	line = alloca(size + 1);
-	strlcpy(line, orig_line, orig_line_size);
 	if (inbound_line_mangler)
 	{
-	    if (mangle_line(line, inbound_line_mangler, orig_line_size) > orig_line_size)
-		yell("mangle_line truncated its result.  Ack.");
+	    char *s;
+	    s = new_normalize_string(orig_line, 1, inbound_line_mangler);
+	    line = LOCAL_COPY(s);
+	    new_free(&s);
 	}
+	else
+	    line = LOCAL_COPY(orig_line);
 
 	ArgList = TrueArgs;
 	BreakArgs(line, &from, ArgList);
@@ -1349,22 +1403,16 @@ void 	parse_server (const char *orig_line, size_t orig_line_size)
 		return;		/* Serious protocol violation -- ByeBye */
 	}
 
-	/* 
-	 * I reformatted these in may '96 by using the output of /stats m
-	 * from a few busy servers.  They are arranged so that the most 
-	 * common types are high on the list (to save the average number
-	 * of compares.)  I will be doing more testing in the future on
-	 * a live client to see if this is a reasonable order.
-	 */
+	/* Some day all this needs to be replaced with an alist. */
 	if (is_number(comm))
 		numbered_command(from, comm, ArgList);
 	else
 	{
-		retval = (protocol_command *)find_fixed_array_item(
-			(void *)rfc1459, sizeof(protocol_command), 
-			num_protocol_cmds + 1, comm, &cnt, &loc);
+		for (loc = 0; rfc1459[loc].command; loc++)
+			if (!strcmp(rfc1459[loc].command, comm))
+				break;
 
-		if (cnt < 0 && rfc1459[loc].inbound_handler)
+		if (rfc1459[loc].command && rfc1459[loc].inbound_handler)
 			rfc1459[loc].inbound_handler(from, comm, ArgList);
 		else
 			rfc1459_odd(from, comm, ArgList);
